# -------------------------------------------------------------------------
# extract
#
# This utility reads lsb bits according to the no. of bits in field <bits>
# As a result the input <var> value is changed, as no. of bits were removed
# (shifted ro right)
# --------------------------------------------------------------------------
proc extract {var bits} {
    upvar $var v

    set result [expr $v & ((1 << $bits) - 1)]
    set v [expr $v >> $bits]
    return $result
}



# -----------------------------------------------------------------
# igu_dump_fifo
#
# Dump up to 8 IGU fifo entries
# Dump the first 8 IGU errors or
# the last 8 IGU commands if IGU_REG_COMMAND_DEBUG is set.
#
# Parameters:        None
# Return value:      None
# -----------------------------------------------------------------
proc  igu_dump_fifo {} {    
    
    # Parse the RDF file    
    parse_rdf
    
    if {[reg_read -n IGU_REG_ERROR_HANDLING_DATA_VALID] == 0} {
        error "IGU fifo is empty"
    }

    if {[set parseCommands [reg_read -n IGU_REG_COMMAND_DEBUG]]} {
        puts "Last IGU Commands: "
    } else {
        puts "First IGU Errors:"
    }
    
    # Read IGU fifo until it is empty
    while {[reg_read -n IGU_REG_ERROR_HANDLING_DATA_VALID]} {
        scan [reg_offset -n IGU_REG_ERROR_HANDLING_MEMORY] "0x%x-0x%x" lsb_addr msb_addr
        
        # Read WB register
        set val_l [reg read $lsb_addr]
        set val_m [reg read $msb_addr]
        
        # Parse data:
        #--------------------------------
        # parse cmd_fid bits [8:0]
        # if bit 8 is set - PF, else VF.
        # bits [7:0] - FID
        set cmd_fid [extract val_l 9]
        if {$cmd_fid & 0x100} {
            set cmd_fid [format "PF %d" [expr {$cmd_fid & 0xFF}]]
        } else {
            set cmd_fid [format "VFID 0x%x" $cmd_fid]
        }
        
        # parse source bits [12:9]
        # values 0-7 PXP, 8-CAU, 9-ATTN, 10-GRC
        set cmd_source [extract val_l 4]
        if {$cmd_source < 7} {
            set err_src "PXP"
        } elseif {$cmd_source == 8} {
            set err_src "CAU"
        } elseif {$cmd_source == 9} {
            set err_src "ATTN"
        } elseif {$cmd_source == 10} {
            set err_src "GRC"
        } else {
            set err_src "Invalid"
        }        
       
        # parse error type bits [16:13]
        # 0 - no error,
        # 1 - length error, 
        # 2 - function disabled,
        # 3 - VF sent command to attnetion address, 
        # 4 - host sent prod update command, 
        # 5 - read of during interrupt register while in MIMD mode, 
        # 6 - access to reserved address, 
        # 7 - producer update command to attention index, 
        # 9 - SB index not valid, 
        # 10 - SB relative index and FID not found, 
        # 11 - FID not match, 
        # 12 - command with error flag aserted (PCI error or CAU discard)
        # 13 - VF sent cleanup and RF cleanup is disabled, 
        # 14 - cleanup command on type bigger than 4).

        set error_type_d [lindex {"no error" "length error"  "function disabled" 
                "VF command to attention address" "host producer update command"
                "read during interrupt register while in MIMD mode"
                "access to reserved address" "producer update to attention index" 
                <invalid> "SB index not valid" "SB relative index and FID not found"                 
                "FID not match" "command with error flag asserted"
                "VF sent cleanup but RF cleanup disabled"
                "cleanup on type > 4"} [extract val_l 4]]
         
        # Get command address bits [31:17]    
        set cmd_addr [format "0x%x" [extract val_l 15]]
        
        # GetCommand write or read. bit 32: 0 - read, 1 - write.
        set cmd_wr_or_rd [format "0x%x" [extract val_l 1]]
        
        # Get write data bits [64:33] (MSB bits)
        set wr_data [format "0x%x" $val_m]
        
        puts "addr=$cmd_addr cmd_wr_or_rd=$cmd_wr_or_rd data=$wr_data err_src=$err_src fid=$cmd_fid error=$error_type_d"
    }
}