###########################################################################
#
# Copyright(c) 2014-2017 Cavium, Inc., all rights reserved
# Proprietary and Confidential Information.
#
# Name:        l2pkt.tcl
#
###########################################################################


# Specify any dependency on other packages
package require Tcl

# Specify the package this module provides
package provide _l2pkt_tests 1.0

namespace eval _l2pkt_tests {
    
    # Export the package
    namespace export l2pkt_tests   
}

# Give a short-name for the regression test.
interp alias {} l2test {}  "_l2pkt_tests::l2test_main"


#Globals:

#Registers information in 4 arrays
global regsName
global regsOffset
global org_val1
global org_val2
global org_dntfwd0_reg
global org_dntfwd1_reg

# -------------------------------------------------------
# The description of the test usage
# ------------------------------------------------------
proc _l2pkt_tests::l2test_usage {} {
    puts "l2test - Send and receive L2 packets"
    puts "Usage: l2test"    
    puts "Parameters: "
    puts "            -n <number of packets>"
    puts "            -lb <int/ext> set internal or external loopback"
    puts "            -speed <10g/40g> sets speed 10G or 40G"
    puts "            -amp power down amplitude"
    puts "            -p2p set port to port test"
    puts "            -di <0|1> data check off (0) or on (1)"
}

# -------------------------------------------------------
# The test main proc
# -------------------------------------------------------
proc _l2pkt_tests::l2test_main { args } { 
    
    global rdfFile
    global regsName
    global regsOffset
    global org_val1
    global org_val2
    global org_dntfwd0_reg
    global org_dntfwd1_reg

    #Parse the appropriate RDF file 
    parse_rdf

    # Data integrity is set by default
    set no_pkts 0
    set loop_mode "ext"
    set phy_type "eagle"
    set force_speed 0
    set port_swap 0
    set returnVal "failed"
    set amplitude_chk 0
    set port_to_port 0
    set peer_device 0
    set data_check 1

    
    # Parse the input arguments
    for {set i 0} {$i < [llength $args]} {incr i} {
        switch -- [lindex $args $i] {    
            "-n"        { incr i; set no_pkts [lindex $args $i]}
            "-lb"       { incr i; set loop_mode [lindex $args $i]}
            "-force_speed"    {incr i; set force_speed [lindex $args $i]}
            "-swap"     {incr i; set port_swap [lindex $args $i]}
	    "-amp"	{incr i; set amplitude_chk 1}
            "-p2p"	{incr i; set port_to_port 1}
            "-di"	{incr i; set data_check [lindex $args $i]}
            "-help"     {incr i; l2test_usage; return}	
            default     { l2test_usage ; return -1}
        }
    }
    
    #If user did not set the number of packets to test, exit with error.
    if {$no_pkts == 0} {
        puts "Number of packets was not set !!\n"
        l2test_usage ; return -1
    }
   
    puts "\nl2test - Send and receive L2 packets"

    if {$port_to_port} {
        set cur_dev $::current(DEV)
        set pci_func $::current(PCI_FUNC)
        set total_dev $::current(TOTAL_DEV)
	
	if ($::sys(IS_UEFI)) {
	    puts "l2test - port to port mode is not supported on UEFI (INTA mode)"
	    set returnVal "failed: Not supported"
	    return $returnVal
	}

        if {[expr {$pci_func % 2}] == 0} {
            set peer_device [expr {$cur_dev+1}]
        } else {
            set peer_device [expr {$cur_dev-1}]
        }
        puts "l2test - Port to port mode (peer device $peer_device)"
    }
    
    # Initialization 
    puts "Initialize $loop_mode loopback"
    if {[l2test_init $loop_mode $amplitude_chk $peer_device $force_speed] != 0} {
        set returnVal "failed: Initialization failed"
    } else {
        if {$::current(DRV_STATE) != "LINK_UP"} {
            set returnVal "failed: LINK DOWN"
        } else {
            # Test procedures/commands
            set returnVal [l2test_test $no_pkts $peer_device $data_check]  
        }
    }

    # TBD: Verify idleChk passes after this test 
    #      Currently idleChk is not clean  
  
    if {$returnVal != "passed"} {
        puts "***********************************************************"
        puts "**             TRAFFIC TEST FAILED !!!                     "
        puts "***********************************************************"
    } else {
        puts "***********************************************************"
        puts "**             TRAFFIC TEST PASSED                         "
        puts "***********************************************************"
    }
    
    # Cleanup before exiting test
    if {[l2test_cleanup $amplitude_chk $peer_device] != 0} {
        puts "Cleanup after test failed !!"
        set returnVal "failed: Cleanup failed"
    } 

    # Test should return passed/failed
    return $returnVal
}


# -------------------------------------------------------
# The test init proc
# -------------------------------------------------------
proc _l2pkt_tests::l2test_init {{loop_mode} {amplitude_chk 0} {peer_device 0} {force_speed 0}} {
    
    global org_val1
    global org_val2
    global org_dntfwd0_reg
    global org_dntfwd1_reg

    # Set loopback type
    if {$loop_mode == "int"} {
        puts "Set internal loopback"
        if {[is_emul]} {
            #Use NIG loopback on emulation
            reg_write -n NIG_REG_ECO_RESERVED -v 3
        } else {
            driver set_loopback INT_PHY
        }
    } else {
        driver set_loopback NONE
    }
    
    #If user requested -force_speed use the requested speed
    if {$force_speed != 0} {
         if {($force_speed != "1G") && ($force_speed != "10G") &&
                ($force_speed != "25G") && ($force_speed != "40G") && ($force_speed != "50G")} {
                puts "Unknown speed requested $force_speed (Support 1G/10G/25G/40G/50G)!"
        } else {
            puts "Force link speed $force_speed"
            driver set_speed $force_speed
        }
    } else {

        set speed [nvm cfg 29=]
        set capabilities [nvm cfg 27=]
        
        #If Autoneg configured, force link speed to maximum speed
        if {$speed == 0} {
            if {[expr {$capabilities & 0x40}] == 0x40} {
                puts "Force link speed 100G"
                driver set_speed 100G
            } elseif {[expr {$capabilities & 0x20}] == 0x20} {
                puts "Force link speed 50G"
                driver set_speed 50G
            } elseif {[expr {$capabilities & 0x10}] == 0x10} {
                puts "Force link speed 40G"
                driver set_speed 40G
            } elseif {[expr {$capabilities & 0x8}] == 0x8} {
                puts "Force link speed 25G"
                driver set_speed 25G
            } elseif {[expr {$capabilities & 0x2}] == 0x2} {
                puts "Force link speed 10G"
                driver set_speed 10G
            } elseif {[expr {$capabilities & 0x1}] == 0x1} {
                puts "Force link speed 1G"
                driver set_speed 1G
            } else {
                puts "Unknown link capabilities"
            }
        }
    }
    
    # Disable LLDP packets to avoid D1 internal loopback test failure
    # Use rule to filter packets by ethertype.
    puts "Filter LLDP"
    set org_dntfwd0_reg [reg_read -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7]
    reg_write -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7 -v 0x21

    if {$::current(HWFN) == 1} {
        set cur_dev $::current(DEV)
        device $cur_dev-0
        reg_write -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7 -v 0x21
        device $cur_dev-1
    }

    # clear regFifo before driver load
    dbgRegFifo

    # load driver at the begining of the test
    puts "loading driver ..."
    if {[driver load] != 0} {
        puts "Failed to load driver !!"
        return 1
    }
    delay ms 100
    
    if {[is_emul]} {
        #Use NIG loopback on emulation
        reg_write -n NIG_REG_ECO_RESERVED -v 3
    }

    # Disable DCBX to avoid packet drops due to congestion with DCBX packets
    set dcbx_mode [nvm cfg 26=]
    if {$dcbx_mode != 0} {
        puts "Disable DCBX"
        driver disable_dcbx
    }

    # Reset NCSI block to reset
    # MISCS_REG_RESET_PL_HV [5] rst_ncsi; 
    puts "Reset NCSI"
    set offset [reg_offset -n "MISCS_REG_RESET_PL_HV"]
    set startOffset [lindex [split $offset "-"] 0]   
    reg write [expr ($startOffset+8)] 0x20
    reg write [expr ($startOffset+4)] 0x20
    
    if {$peer_device != 0} {
        set primary_dev $::current(DEV)
        device $peer_device
	
        #Disable DCBX packets to avoid D3 est failure
        set org_dntfwd1_reg [reg_read -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7]
        reg_write -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7 -v 0x21
	
        #If user requested -force_speed use the requested speed
        if {$force_speed != 0} {
            if {($force_speed != "1G") && ($force_speed != "10G") &&
                ($force_speed != "25G") && ($force_speed != "40G") && ($force_speed != "50G")} {
                puts "Unknown speed requested $force_speed (Support 1G/10G/25G/40G/50G)!"
            } else {
                puts "Force link speed $force_speed"
                driver set_speed $force_speed
            }
        } else {
	
        	set speed [nvm cfg 29=]
        	set capabilities [nvm cfg 27=]
            
        	#If Autoneg configured, force link speed to maximum speed
        	if {$speed == 0} {
        	    if {[expr {$capabilities & 0x40}] == 0x40} {
        		puts "Force link speed 100G on peer device"
        	        driver set_speed 100G
        	    } elseif {[expr {$capabilities & 0x20}] == 0x20} {
        		puts "Force link speed 50G on peer device"
        	        driver set_speed 50G
        	    } elseif {[expr {$capabilities & 0x10}] == 0x10} {
        		puts "Force link speed 40G on peer device"
        	        driver set_speed 40G
        	    } elseif {[expr {$capabilities & 0x8}] == 0x8} {
        		puts "Force link speed 25G on peer device"
        	        driver set_speed 25G
        	    } elseif {[expr {$capabilities & 0x2}] == 0x2} {
        		puts "Force link speed 10G on peer device"
        	        driver set_speed 10G
        	    } elseif {[expr {$capabilities & 0x1}] == 0x1} {
        		puts "Force link speed 1G on peer device"
        	        driver set_speed 1G
        	    } else {
        	        puts "Unknown link capabilities"
        	    }
        	}
        }
        
        #clear regFifo before driver load
        dbgRegFifo
	
        puts "loading driver on peer $peer_device"
        if {[driver load] != 0} {
            puts "Failed to load driver on peer !!"
            return 1
        }
        device $primary_dev
        puts "link state [driver link_state]"
    }

    # CQ104325 - Ignore packets are already inside the RX pipe
    if {$::current(DRV_STATE) == "LINK_UP"} {
        set no_pkts [l2pkt receive 0 0]
        if {$no_pkts != 0} {
            puts "NOTE: Ignore $no_pkts packets that already in RX pipe."
        }
    }
       
    if {$amplitude_chk} {
	# Find the PHY type on the current device
	set info [phy info]
	set port [expr {$::current(DEV)-1}]    
	set pattern "*Port $port is in Eagle*"
	if {[string match -nocase $pattern $info]} {
	    set phy_type "eagle"
	    set port 1
	} else {
	    set phy_type "falcon"
	    set port 0
	}
	
	set pattern "*4x10G*"
	if {[string match -nocase $pattern $info]} {
	    set lane [expr {$::current(DEV)-1}]	
	} else {
	    set pattern "*2x50G*"
	    if {[string match -nocase $pattern $info]} {
		#In 2x50G lane 4,5 makes MFW set all lanes of current port
		if {$::current(DEV) == 1} {
		    set lane 4
		} else {
		    set lane 5
		}
	    } else {
		#In 2x40G, 1x100G lane 6 make MFW set all lanes of current port
		set lane 6
	    }        
	}
	puts "Amplitude check on phy_type $phy_type"
	if {$phy_type == "eagle"} {
	    #set 0xd0a2 [3:0] to 0x0
	    set org_val1 [phy raw $port $lane 0x800d0a2]
            set val1 [expr {$org_val1 & 0xfffffff0} | 0xc]
	    phy raw $port $lane 0x800d0a2 $val1
	    puts "PHY write port $port lane $lane 0x800d0a2 [format "0x%x" $val1]"
	    
	    #set 0xd111 [6:0] to 0x00
	    set org_val2 [phy raw $port $lane 0x800d111] 
	    phy raw $port $lane 0x800d111 [expr {$org_val2 & 0xffffff80}]
	    puts "PHY write port $port lane $lane 0x800d111 [format "0x%x" [expr {$org_val2 & 0xffffff80}]]"
	} else {
	    #set 0xd095 [6:0] to 0x00
	    set org_val1 [phy raw $port $lane 0x800d095]
	    phy raw $port $lane 0x800d095 [expr {$org_val1 & 0xffffff80}]
	    puts "PHY write port $port lane $lane 0x800d095 [format "0x%x" [expr {$org_val1 & 0xffffff80}]]"
	    
	    #set 0xd0d2 [3:0] to 0x0
	    set org_val2 [phy raw $port $lane 0x800d0d2]
	    set val2 [expr {$org_val2 & 0xfffffff0} | 0xc]
	    phy raw $port $lane 0x800d0d2 $val2
	    puts "PHY write port $port lane $lane 0x800d0d2 [format "0x%x" $val2]"
	}
    } 
    return 0     
}


# -------------------------------------------------------
# The test cleanup commands
# -------------------------------------------------------
proc _l2pkt_tests::l2test_cleanup {{amplitude_chk 0} {peer_device 0}} {
    
    global org_val1
    global org_val2
    global org_dntfwd0_reg
    global org_dntfwd1_reg
    
    # Cleanup commands before exiting 
    # Such as: Unset variables, driver unload    
            
    # unload the driver if running
    puts "unloading driver ..."
    if {[driver unload] != 0} {
        puts "Failed to unload driver !!"
        return 1
    }
    
    driver set_loopback NONE

    if {$peer_device != 0} {
        set primary_dev $::current(DEV)
        device $peer_device
        puts "unloading driver on peer device $peer_device ..."
        driver unload
        #Restore DCBX
        reg_write -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7 -v $org_dntfwd1_reg
        device $primary_dev
    }
        
    if {$amplitude_chk} {
	# Find the PHY type on the current device
	set info [phy info]    
	set pattern "*is in Eagle*"
        if {[string match -nocase $pattern $info]} {
            set phy_type "eagle"
	    set port 1
	} else {
	    set phy_type "falcon"
	    set port 0
	}
    
	set pattern "*4x10G*"
	if {[string match -nocase $pattern $info]} {
	    set lane [expr {$::current(DEV)-1}]	
	} else {
	    set pattern "*2x50G*"
	    if {[string match -nocase $pattern $info]} {
	        #In 2x50G lane 4,5 makes MFW set all lanes of current port
	        if {$::current(DEV) == 1} {
		    set lane 4
		} else {
		    set lane 5
		}
	    } else {
	        #In 2x40G, 1x100G lane 6 make MFW set all lanes of current port
	        set lane 6
	    }        
	}
	puts "Amplitude check on phy_type $phy_type"
	if {$phy_type == "eagle"} {	   	    
	    phy raw $port $lane 0x800d0a2 $org_val1
	    puts "PHY write port $port lane $lane 0x800d0a2 [format "0x%x" $org_val1]"	    
	    phy raw $port $lane 0x800d111 $org_val2
	    puts "PHY write port $port lane $lane 0x800d111 [format "0x%x" $org_val2]"
	} else {	    
	    phy raw $port $lane 0x800d095 $org_val1
	    puts "PHY write port $port lane $lane 0x800d095 [format "0x%x" $org_val1]"
	    phy raw $port $lane 0x800d0d2 $org_val2
	    puts "PHY write port $port lane $lane 0x800d0d2 [format "0x%x" $org_val2]"
	}
    }      
    
    #Restore LLDP
    reg_write -n NIG_REG_RX_LLH_L2FILT_BRB_DNTFWD_RULE7 -v $org_dntfwd0_reg
    
    # Restore reset values (including NCSI, enable DCBX)
    if {[is_emul]} {
        #Don't reset mcp on emulation
        reset_chip 1
        delay ms 3000
    } else {
        reset_chip 
    }
                    
    # Unset variables, arrays, specs before exiting test
    
    # Free allocated memories
    return 0
}

# -------------------------------------------------------
# The test commands
# -------------------------------------------------------
proc _l2pkt_tests::l2test_test { {no_pkts} {peer_device 0} {is_data_check}} {
    
    set is_receive 1
   
    if {$peer_device == 0} {
        if {[l2pkt send $no_pkts $is_receive $is_data_check] != 0} {
            set returnVal "failed :l2pkt send failed. [qstat]"
        } else {
            set returnVal "passed"
        }
    } else {
        set is_receive 0
       
        # Send packets on primary device
        set primary_dev $::current(DEV)
        puts "Sending $no_pkts packets on device $primary_dev"
        l2pkt send $no_pkts $is_receive $is_data_check
       
        #receive packets on peer device
        device $peer_device
        puts "Receiving packets on device $peer_device ..."
        set recv_pkts [l2pkt receive $no_pkts $is_data_check]
        puts "Received $recv_pkts packets successfully."
        if { $recv_pkts != $no_pkts} {
            set returnVal "failed"
            return $returnVal
        } 

        # Now send packets on peer device
        puts "Sending $no_pkts packets on device $peer_device"
        l2pkt send $no_pkts $is_receive $is_data_check

        #receive packets on primary device
        device $primary_dev
        puts "Receiving packets on device $primary_dev ..."
        set recv_pkts [l2pkt receive $no_pkts $is_data_check]
        puts "Received $recv_pkts packets successfully."
        if { $recv_pkts != $no_pkts} {
            set returnVal "failed"
        } else { 
            set returnVal "passed"  
        }
    }
    return $returnVal
}





