##############################################################################
# Remote debugger for E4 MCP
# Implementation of the GDB RSP protocol for E4 MCP
#
# mcp_dbg can work in two modes: "online" and "snapshot".
#
# "online" mode allows debugging a live MCP through GRC:
# Usage:
#	debug_mcp
#
# Then connect to port 2159 using GDB remote connection 
# gdb> target remote host:2159
#
# "snapshot" mode allows debugging offline MCP, accroding to a GRC dump.
# Usage:
#	debug_mcp_snapshot <mcp.txt>
#	or
#	parse_mcp_snapshot <mcp.txt> [<MFW image>]
#
# debug_mcp_snapshot allows remote debugging in the same manner as 
# debug_mcp (target remote host:2159). You can inspect the call stack
# and variables, but cannot set breakpoints run, change memory, etc.
#
# parse_mcp_snapshot runs gdb automatically behind the scenes and prints
# the stack trace and some variables.
#
# <mcp.txt> is the parsed MCP memories and registers, created by parseDump 
# (or parseDumpAuto). parseDump parses the binary results of grcDump.
#
# <MFW image> is an optional parameter, which specifies the path to the MCP 
# linked image (for example, the path to b900mim). 
# If it is not given, an image is searched on a file server according to 
# MCP version.
#
# For snapshot mode, the following environment variables should be defined 
# in order to allow searching for MFW image on a file server and running
# gdb behind the scenes:
#
# MCP_VER_ROOT: Root path for MFW versions on the file server
# MCP_IMAGE_PATH: Path to the MFW image within specific version
# MCP_GDB_PATH: Path to the gdb executable
# MCP_GDB_DATADIR: Path to gdb Data Dir (contains Python code for gdb)
#
#
##############################################################################


namespace eval rsp {

########################################################################
# Constants
########################################################################

# Constants from reg_addr.h

	set MCP_REG_CPU_MODE_SOFT_HALT                                                               (0x1<<10)
	set MCP_REG_CPU_MODE_SOFT_HALT_SHIFT                                                         10
	
	set MCP_REG_CPU_MODE_STEP_ENA                                                                (0x1<<1)
	set MCP_REG_CPU_MODE_STEP_ENA_SHIFT                                                          1
	
	set MCP_REG_CPU_STATE_BREAKPOINT                                                             (0x1<<0)
	set MCP_REG_CPU_STATE_BREAKPOINT_SHIFT                                                       0
	
	set MCP_REG_CPU_STATE_BAD_INST_HALTED                                                        (0x1<<2)
	set MCP_REG_CPU_STATE_BAD_INST_HALTED_SHIFT                                                  2	

# Constants from misc_bits.h
	
	set MISCS_REG_RESET_PL_UA_SET__FIELDOFFSET			4
	set MISCS_REG_RESET_PL_UA_CLEAR__FIELDOFFSET			8
	
	set MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_REG_HARD_CORE		(0x1<<1)
	set MISCS_REG_RESET_PL_UA_RST_MCP_N_HARD_CORE_RST_B		(0x1<<2)
	set MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CPU		(0x1<<3)
	set MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CORE		(0x1<<4)

# Other constants
	
	set RSP_PORT		2159					  ;# The TCP port where we will listen for RSP connections (Registered GDB Remote Debug Port)
	set MAX_RSP_PORT	[expr $RSP_PORT + 100]			  ;# If TCP port is not available, try other ports up to MAX_RSP_PORT
	
	set MAX_PACKET_SIZE	16384				          ;# Max size of a single RSP packet. 16384 is the maximum supported by GDB
	set MEM_MASK 		0x08000000			          ;# the bit we need to clear when looking at memory addresses
	set STATIC_MEM_SIZE 	[expr 160*1024]			          ;# data until this offset will not be paged and can be easily accessed
	set PIM_MEM_MASK	[expr $MEM_MASK + $STATIC_MEM_SIZE]       ;# PIM area
	set RO_PIM_MEM_SIZE	[expr 16*4096]                            ;# PIM size
	set RW_MEM_SIZE 	$STATIC_MEM_SIZE		          ;# size of the memory GDB will be allowed to access
	set DISABLED_BREAKPOINT	0x1				          ;# value for HW breakpoint which disables the breakpoint (see MCP_REG_CPU_HW_BREAKPOINT description)
	set NOP 		0x0				          ;# NOP instruction	
	set BREAK		0x0005000D			          ;# BREAK 5 instruction		
	set NUM_OF_GPRS		32				          ;# Total number of MIPS GPRs
	set PC_REG_INDEX	37				          ;# Index of PC register
	
	;# Memory map response for qXfer:memory-map:read
	;# PIM areas are marked as read-only in order to avoid SW breakpoints there.	
	
	set MEMORY_MAP "<?xml version=\"1.0\"?>
		<!DOCTYPE memory-map
			  PUBLIC \"+//IDN gnu.org//DTD GDB Memory Map V1.0//EN\"
				 \"http://sourceware.org/gdb/gdb-memory-map.dtd\">
		<memory-map>
			<memory type=\"ram\" start=\"$MEM_MASK\" length=\"$RW_MEM_SIZE\"/>			
			<memory type=\"rom\" start=\"$PIM_MEM_MASK\" length=\"$RO_PIM_MEM_SIZE\"/>			
		</memory-map>	
	"
	
# Constants for obtaining the correct MFW image and parsing snapshot
	
	set MCP_REG_SCRATCH	0xe20000	
	
# Snapshot GDB commands:
	
	proc set_snp_gdb_cmd {port} {
		global rsp::SNP_GDB_CMD
		set SNP_GDB_CMD "
			set pagination off
			set print array on
			set print array-indexes on
			set print elements unlimited
			set print pretty on
			target remote localhost:$port
			backtrace full
			p/x g_spad
			detach
			quit
			"
	}
	
	set_snp_gdb_cmd $RSP_PORT


########################################################################
# Globals
########################################################################

	variable stop_debugger_event 0		;# RSP socket waits on this event in order to enter event loop and serve requests
	variable show_rsp_info 1		;# Whether to display rsp info and errors
	
	variable connected 0
	variable chan 0
	variable server_socket 0
	
	variable state "control" 		;# control, data, checksum
	variable checksum {} 			;# list of two hex chars
	variable packet {} 			;# list of incoming packet chars from gdb (only data)
	variable _start "" 			;# The start address of the .text section
	variable sw_breakpoints  		;# key is breakpoint address, value is the previous data on that address. Address resultion is 32bit (dword)
	set sw_breakpoints [dict create]
	
	# Snapshot
		
	set snp_scratchpad {}			;# snapshot of the scratchpad memory. List of 32bit values.
	set snp_regfile {}			;# snapshot of the CPU register file memory. List of 32bit values.
	set snp_regs  [dict create]		;# snapshot of MCP GRC registers. Key is register name, value is 32bit.
	set batch 0
	
########################################################################
# General functions
########################################################################
	
	proc is_snapshot {} {	
		variable snp_regs
		return [expr [dict size $snp_regs] != 0]
	}

	proc reset_state {} {
		variable state
		variable checksum
		variable packet
		
		set state "control"
		set checksum {}
		set packet {}
	}

	proc info {s} {
		variable show_rsp_info
		
		if {$show_rsp_info} {
			puts "GDB RSP: $s"
		}
	}

	proc error {s} {		
		info "ERROR: $s"
		close_connection
	}
	
########################################################################
# String and Arithmetic functions
########################################################################

	# Given a list of numbers $l, convert them to a string where each number is 
	# represented by word_size hex digits
	proc format_hex {l {word_size 4}} {
		set l1 {}
		foreach n $l {
			lappend l1 [format "%.[expr $word_size*2]X" $n]
		}
		return [join $l1 ""]
	}
	
	# Given a string where each number is represented by 8 hex digits
	# convert it to a list of 32bit numbers
	proc unformat_hex {hex {word_size 4}} {
		set result {}
		while {[string length $hex] > 0} {
			set value [string range $hex 0 [expr $word_size*2-1]]
			set hex [string replace $hex 0 [expr $word_size*2-1]]
			if {[string length $value] != [expr $word_size*2]} {error "Data not aligned to word size of $word_size bytes: $value"}
			lappend result [scan $value %[expr $word_size*2]x]
		}
		return $result
	}

	# Calculates the RSP checksum. data is a list of characters
	proc calc_checksum {data} {
	    set sum 0
	    for {set k 0} {$k < [llength $data]} {incr k} {
		binary scan [lindex $data $k] c dd
		set sum [expr $dd + $sum]
	    }
	    set sum [expr $sum % 256]
	    return [format %.2x $sum]
	}
	
	proc hex {value} {
		return [format 0x%08X [expr $value]]
	}
	
	proc hex_list {l} {
		set hex_l {}
		foreach i $l {
			lappend hex_l [hex $i]
		}
		return $hex_l
	}
	
	proc string_to_hex {str} {
		binary scan $str H* hex
		return $hex
	}

        # Converts value to a list of bytes. word_size is the number of bytes that should be returned.
        proc byte_list {value word_size} {
                set l {}
                for {set i 0} {$i < $word_size} {incr i} {                    
                    lappend l [expr $value & 0xFF]
                    set value [expr $value >> 8]
                }
                return [lreverse $l]
        }
	

########################################################################
# MCP functions that communicate through GRC
########################################################################
		
	# reg read
	proc rr {name} {
		variable snp_regs
		
		if {[is_snapshot]} {
			return [dict get $snp_regs $name]
		}
		
		return [expr [reg_read -n $name]]
	}

	# reg write where value can be any number
	proc rw {name value} {		
		if {[is_snapshot]} {
			error "Cannot write register $name on snapshot mode"
			return
		}
		
		reg_write -n $name -v [hex $value]		
	}

	# reg read field. Assume vars are defined as a mask, and also with a _SHIFT suffix
	proc rrf {regName fieldName} {
		set fieldRegName [join [list {rsp::} $fieldName] {}]
		set fieldShiftName [join [list {rsp::} $fieldName {_SHIFT}] {}]
		return [expr ([rr $regName] & [set $fieldRegName]) >> [set $fieldShiftName]]
	}

	# reg write field
	proc rwf {regName fieldName value} {
		set fieldRegName [join [list {rsp::} $fieldName] {}]
		set fieldShiftName [join [list {rsp::} $fieldName {_SHIFT}] {}]
		rw $regName [expr ([rr $regName] & (~[set $fieldRegName])) | (($value << [set $fieldShiftName]) & [set $fieldRegName])]
	}
	
	# halts or resumes MCP according to halt_state
	proc halt_mcp {halt_state} {
		global batch
		
		if {!$batch} {
			rwf MCP_REG_CPU_MODE MCP_REG_CPU_MODE_SOFT_HALT $halt_state
		}
	}
		
	# checks whether MCP was halted
	proc is_mcp_halted {} {
		return [rrf MCP_REG_CPU_MODE MCP_REG_CPU_MODE_SOFT_HALT] 
	}
	
	# given memory name $m, returns a list of values in this memory.
	# optionally, offset and size in the memory can be specified. offset/size in byte resolution
	proc read_grc_memory {m {offset 0} {size 0} } {				
		if {[is_snapshot]} {
			error "Cannot read grc memory on snapshot mode"
			return
		}
		
		set result {}
		set range [split [reg_offset -n $m] -] ;# range is {start-addr end-addr} of memory
		set start [expr [lindex $range 0] + $offset]
		
		if {$size == 0} {
			set end [lindex $range 1]
		} else {
			set end [expr $start + $size - 1]
		}
		info "read_grc_memory $m: [hex $start] - [hex $end]"
		for {set i $start} {$i <= $end} {incr i 4} {
			lappend result [expr [reg read $i]]
		}
		return $result
	}
	
	# given memory name $m, and $offset and $data (list of 32bit values)
	# writes $data into memory $m starting from $offset 
	proc write_grc_memory {m offset data} {
		if {[is_snapshot]} {
			error "Cannot write grc memory on snapshot mode"
			return
		}
		
		set result {}
		set range [split [reg_offset -n $m] -] ;# range is {start-addr end-addr} of memory
		set start [expr [lindex $range 0] + $offset]
		set end [expr $start + ([llength $data]*4) - 1]
		
		info "write_grc_memory $m: [hex $start] - [hex $end]"
		
		set dataIndex 0
		for {set i $start} {$i <= $end} {incr i 4; incr dataIndex} {			
			reg write $i [lindex $data $dataIndex]
		}
	}
	
	# reads program counter from GRC
	proc read_pc {} {
		return [rr MCP_REG_CPU_PROGRAM_COUNTER]
	}
	
	# write program counter
	proc write_pc {value} {
		info "Writing Program Counter (PC): [hex $value]"
		
		# Writing pc is only valid when MCP is halted. Halt it if needed.
		
		set need_to_halt [expr ![is_mcp_halted]]
		if {$need_to_halt} {halt_mcp 1}
		rw MCP_REG_CPU_PROGRAM_COUNTER $value
		if {$need_to_halt} {halt_mcp 0}	
	}
	
	# reads current instruction on PC, from GRC
	proc read_inst {} {
		return [rr MCP_REG_CPU_INSTRUCTION]
	}

	proc write_inst {data} {
		info "Writing current instruction: [hex $data] instead of [hex [read_inst]] (pc=[hex [read_pc]])"
		rw MCP_REG_CPU_INSTRUCTION $data
	}
		
	# breakpoint is either a HW breakpoint (CPU_STATE_BREAKPOINT), or a "break" instruction (BAD_INST_HALTED...)
	proc is_breakpoint_hit {} {
		return [expr [rrf MCP_REG_CPU_STATE MCP_REG_CPU_STATE_BREAKPOINT] || [rrf MCP_REG_CPU_STATE MCP_REG_CPU_STATE_BAD_INST_HALTED]]
	}
	
	# halt and clear breakpoint indications (both SW and HW breakpoints)
	proc handle_breakpoint {} {
		variable sw_breakpoints
		
		# halt mcp
		halt_mcp 1
		
		# write-clear registers is set with rw and not rwf!
		
		# if "break" instruction encountered, this is either a bad instruction or a breakpoint
		# - if it's a bad instruction - need to change it to nop before continuing
		# - if it's a breakpoint - need to restore the original instruction before continuing
		if {[rrf MCP_REG_CPU_STATE MCP_REG_CPU_STATE_BAD_INST_HALTED]} {			
			set pc [read_pc]
			if {[dict exists $sw_breakpoints $pc]} {
				write_inst [dict get $sw_breakpoints $pc]
			} else  {
				write_inst $rsp::NOP
			}
			
			rw MCP_REG_CPU_STATE $rsp::MCP_REG_CPU_STATE_BAD_INST_HALTED
		}		
		
		# clear the breakpoint indication
		rw MCP_REG_CPU_STATE $rsp::MCP_REG_CPU_STATE_BREAKPOINT
	}
	
	# read 32 gprs from GRC
	proc read_gprs {} {
		variable snp_regfile
		
		if {[is_snapshot]} {
			return $snp_regfile
		}
		
		# Reading gps is only valid when MCP is halted. Halt it if needed.
		
		set need_to_halt [expr ![is_mcp_halted]]
		if {$need_to_halt} {halt_mcp 1}
		set result [read_grc_memory MCP_REG_CPU_REG_FILE]
		if {$need_to_halt} {halt_mcp 0}
				
		return $result
	}
	
	# write a single MIPS gpr. 
	# regValue is the new GPR value, regIndex is the index of the gpr to write to.
	proc write_gpr {regValue regIndex} {
	
		info "Write GPR register $regIndex: [hex $regValue]"
		
		# Writing gps is only valid when MCP is halted. Halt it if needed.
		
		set need_to_halt [expr ![is_mcp_halted]]
		if {$need_to_halt} {halt_mcp 1}
		write_grc_memory MCP_REG_CPU_REG_FILE [expr $regIndex*4] [list $regValue]
		if {$need_to_halt} {halt_mcp 0}		
	}
	
        proc is_access_aligned {addr {size 0}} {
                return [expr (($addr & 3) == 0) && (($size & 3) == 0)]
        }
        
	proc normalize_address {addr size} {	
		
                if {![is_access_aligned $addr $size]} {
			error "Address [hex $addr] and Size [hex $size] must both be aligned to 4"
                }
	
		set offset [expr $addr - $rsp::MEM_MASK]		
		if {$offset + $size <= $rsp::STATIC_MEM_SIZE} {
			return $offset
		} else {		
			info "Cannot read/write from memory address [hex $offset]"
			return -1
		}		
	}
	
	proc read_mem {offset size} {	
		variable snp_scratchpad
		
		info "read_mem [hex $offset] [hex $size]"
		set normalized_offset [normalize_address $offset $size]
		if {$normalized_offset >= 0} {
			if {[is_snapshot]} {
				return [lrange $snp_scratchpad [expr $normalized_offset/4] [expr ($normalized_offset+$size)/4 - 1]]
			} else {
				return [read_grc_memory MCP_REG_SCRATCH $normalized_offset $size]
			}
		}
		
		info "Cannot read from memory address [hex $offset]"
		
		return {}
	}
        
        # returns list of bytes from any offset/size, even unaligned.
        proc read_unaligned_mem {offset size} {
                set ALIGNMENT 4
                set aligned_offset [expr ($offset / $ALIGNMENT)*$ALIGNMENT]
                set aligned_offset_delta [expr $offset - $aligned_offset]
                set aligned_size [expr (($aligned_offset_delta + $size + $ALIGNMENT - 1)/$ALIGNMENT)*$ALIGNMENT]
                set aligned_data [read_mem $aligned_offset $aligned_size] 
                set aligned_byte_data {}
                foreach v $aligned_data {
                        lappend aligned_byte_data {*}[byte_list $v $ALIGNMENT]
                }
                return [lrange $aligned_byte_data $aligned_offset_delta [expr $size + $aligned_offset_delta -1]]
        }
	
	proc write_mem {offset data} {
		info "write_mem [hex $offset]: [format_hex $data]"
		set size [expr [llength data] * 4]
		
		if {$offset == [read_pc] && $size == 4} {
			write_inst [lindex $data 0]
		}

		set normalized_offset [normalize_address $offset $size]
		if {$normalized_offset >= 0} {
			write_grc_memory MCP_REG_SCRATCH $normalized_offset $data
			return 1
		}
		
		info "Cannot write to memory address [hex $offset]"
		
		return 0
	}
	
	proc single_step {} {
		handle_breakpoint ;# if a breakpoing indication is set, clear it
		rwf MCP_REG_CPU_MODE MCP_REG_CPU_MODE_STEP_ENA 1
	}
	
	proc is_breakpoint_active {} {
		return [expr [rr MCP_REG_CPU_HW_BREAKPOINT] != $rsp::DISABLED_BREAKPOINT]
	}
	
	proc insert_hw_breakpoint {addr} {		
		if {![is_access_aligned $addr]} {error "Breakpoint addr must be aligned to 4"}
		if {[rr MCP_REG_CPU_HW_BREAKPOINT] != $rsp::DISABLED_BREAKPOINT} {error "Breakpoint already set"}
		rw MCP_REG_CPU_HW_BREAKPOINT $addr
	}
	
	proc remove_hw_breakpoint { {addr -1} } {
		if {$addr != -1 && [rr MCP_REG_CPU_HW_BREAKPOINT] != $addr} {error "Cannot clear breakpoint from address $addr"}
		rw MCP_REG_CPU_HW_BREAKPOINT $rsp::DISABLED_BREAKPOINT
	}
	
	# Replace instructions with BREAK instructions, and record the previous instructions
	proc insert_sw_breakpoint {addr {len 4} } {
		variable sw_breakpoints
		
		if {![is_access_aligned $addr $len]} {error "Breakpoint addr and length must be aligned to 4"}
		
		set dword_len [expr $len/4]
		set new_mem [lrepeat $dword_len $rsp::BREAK]
		set saved_mem [read_mem $addr $len]		
		write_mem $addr $new_mem
				
		for {set i 0} {$i < $dword_len} {incr i} {
			dict set sw_breakpoints [expr $addr+($i*4)] [lindex $saved_mem $i]
		}
				
		return 1
	}
	
	# Restore previous instructions before inserting the breakpoint
	proc remove_sw_breakpoint { addr {len 4} } {
		variable sw_breakpoints
		
		if {![is_access_aligned $addr $len]} {error "Breakpoint addr and length must be aligned to 4"}
		
		set saved_mem {}
		set dword_len [expr $len/4]		
		for {set i 0} {$i < $dword_len} {incr i} {
			set breakpoint_addr [expr $addr+($i*4)]
			lappend saved_mem [dict get $sw_breakpoints $breakpoint_addr]
			dict unset sw_breakpoints $breakpoint_addr
		}
		write_mem $addr $saved_mem
		
		return 1
	}
		
	proc remove_all_sw_breakpoints {} {
		variable sw_breakpoints
		set breakpoints_removed 0
		
		dict for {addr inst} $sw_breakpoints {
			if {[remove_sw_breakpoint $addr]} {incr breakpoints_removed}
		}
		
		return [expr $breakpoints_removed == [dict size $sw_breakpoints]]
	}
	
	proc reset_mcp {} {
		variable sw_breakpoints		
		
		if {[is_snapshot]} {
			error "Cannot reset mcp on snapshot mode"
			return
		}
		
		set MISCS_REG_RESET_PL_UA		[lindex [split [reg_offset -n MISCS_REG_RESET_PL_UA] -] 0]
		set MISCS_REG_RESET_PL_UA_SET		[expr $MISCS_REG_RESET_PL_UA + $rsp::MISCS_REG_RESET_PL_UA_SET__FIELDOFFSET]
		set MISCS_REG_RESET_PL_UA_CLEAR		[expr $MISCS_REG_RESET_PL_UA + $rsp::MISCS_REG_RESET_PL_UA_CLEAR__FIELDOFFSET]	
		
		reg write $MISCS_REG_RESET_PL_UA_CLEAR [expr	\
				  ($rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_REG_HARD_CORE |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_HARD_CORE_RST_B |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CPU |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CORE)]
					  			
		after 5	;# Sleep 5 ms
			
		reg write $MISCS_REG_RESET_PL_UA_SET [expr	\
				  ($rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_REG_HARD_CORE |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_HARD_CORE_RST_B |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CPU |\
					  $rsp::MISCS_REG_RESET_PL_UA_RST_MCP_N_RESET_CMN_CORE)]

		after 500 ;# Wait for MCP to be up 500msec
		
		set sw_breakpoints [dict create] ;# MCP was reset, clear all recorded breakpoints
	}
	
########################################################################
# RSP functions
########################################################################
	
	proc close_connection {} {
		variable chan
		variable connected
		variable stop_debugger_event
		global batch
		
		if {$connected} {
			close $chan
			set connected 0
			set $chan 0			
		}
		
		# In case of offline debugging, don't wait for another gdb connection
		
		if {[is_snapshot] || $batch} {
			incr stop_debugger_event
		} else {
			remove_all_sw_breakpoints
			remove_hw_breakpoint
		}
		
		reset_state		
	}
	
	# When GDB continues, wait until a breakpoint hit by polling.
	proc poll {time} {
		variable connected
		
		if {!$connected} {return}
		
		if {[is_breakpoint_hit] || [is_mcp_halted]} {
			info "Breakpoint hit at [hex [read_pc]]"
			handle_breakpoint
			send_packet "S05"
		} else {
			info "Waiting for Breakpoint $time seconds..."			
			incr time
			after 1000 "rsp::poll $time"
		}
	}
	
	# returns a list of "all MIPS registers", for use in regs packet
	# Today we only support GPRs and PC. All other regs are set to 0
	proc read_all_regs {} {
		
		set l {}
		lappend l {*}[read_gprs]	 ;# gprs
		lappend l 0 ;# status
		lappend l 0 0;# lo/hi
		lappend l 0 ;# badvaddr
		lappend l 0 ;# cause
		lappend l [read_pc] ;# pc. This added in index of PC_REG_INDEX 
		lappend l {*}[lrepeat 32 0] ;# 32 floating-point registers
		lappend l 0 0;# fcsr/fir
		lappend l 0 ;# restart
		
#		lappend l 0 0;# fp
#		lappend l {*}[lrepeat 16 0] ;# 16 cp0_index
				
		return $l
	}
	
	# Write a single MIPS register
	proc write_reg {regIndex regValue} {
	
		if {$regIndex < $rsp::NUM_OF_GPRS} {			
			write_gpr $regValue $regIndex
		} elseif {$regIndex == $rsp::PC_REG_INDEX} {
			write_pc $regValue			
		} else {
			info "Writing to register $regIndex is not supported"
			return 0
		}
		
		return 1
	}

	# sends a packet back to gdb. Calculates CRC
	proc send_packet {outPacket} {
		variable chan
		
		info "-> $outPacket"
		set outChecksum [calc_checksum [split $outPacket {}]]		
		puts -nonewline $chan "\$$outPacket#$outChecksum"		
	}
		
	# Handle a received packet. Respond to GDB requests
	proc receive_packet {} {
		variable packet
		variable _start	
		
		set data [join $packet ""]
		info "<- $data"
		
		switch -regexp -matchvar matches -- $data {
						
			"^g"					{send_packet [format_hex [read_all_regs]]}			
			"^qSupported.*" 			{send_packet "PacketSize=[format_hex $rsp::MAX_PACKET_SIZE];qXfer:memory-map:read+"}
			"^qOffsets" 				{send_packet "Text=0;Data=0;Bss=0"}
			"^qAttached" 				{send_packet "1"}
			"^qfThreadInfo"				{send_packet "l"}
			"^qSymbol::"				{send_packet "qSymbol:[string_to_hex _start]"}
			"^qSymbol:([0-9a-f]+):([0-9a-f]+)"	{set _start 0x[lindex $matches 1]; send_packet "OK"} 
			"^qXfer:memory-map:read::0,([0-9a-f]+)"	{send_packet "l[set rsp::MEMORY_MAP]"}
			
			"^[?].*" {			
				# TODO: calculate response instead of hard coded S05 ?
				send_packet "S05"
			}
			
			"^p([0-9a-f]+)" {
				set regIndex 0x[lindex $matches 1]
				set regValue [lindex [read_all_regs] $regIndex]
				info "Read register $regIndex: [hex $regValue]"
				send_packet [format_32_bit [list $regValue]]			
			}
			
			"^P([0-9a-f]+)=([0-9a-f]+)" {
				set regIndex 0x[lindex $matches 1]
				set regValue 0x[lindex $matches 2]
				if {[write_reg $regIndex $regValue]} {
					send_packet "OK"
				} else {
					send_packet "E01"
				}
			}
			
			"^m([0-9a-f]+),([0-9a-f]+)"		{
                                set offset 0x[lindex $matches 1]
                                set size 0x[lindex $matches 2]
                                if {[is_access_aligned $offset $size]} {
                                        set result [read_mem $offset $size]
                                        set alignment 4
                                } else {
                                        set result [read_unaligned_mem $offset $size]
                                        set alignment 1
                                }

				if {[llength $result] > 0} {
					send_packet [format_hex $result $alignment]
				} else {
					send_packet "E01"
				}
			}
			
			"^M([0-9a-f]+),([0-9a-f]+):([0-9a-f]+)"	{
				set data [unformat_hex [lindex $matches 3]]
				set len 0x[lindex $matches 2]
				if {$len != [llength $data]*4} {
					error "Data does not match length: [lindex $matches 3]"
				} else {
					if {[write_mem 0x[lindex $matches 1] $data]} {
						send_packet "OK"
					} else {
						send_packet "E01"
					}
				}
			}
			
			"^Z0,([0-9a-f]+),([0-9a-f]+)" {
				set address 0x[lindex $matches 1]
				set len 0x[lindex $matches 2]				
				if [insert_sw_breakpoint $address $len] {
					info "SW breakpoint inserted on $address $len"
					info "Breakpoints (addr/value) = [hex_list $rsp::sw_breakpoints]"
					send_packet "OK"
				} else {
					send_packet "E01"
				}
			}
			
			"^z0,([0-9a-f]+),([0-9a-f]+)" {
				set address 0x[lindex $matches 1]
				set len 0x[lindex $matches 2]				
				if [remove_sw_breakpoint $address $len] {
					info "SW breakpoint removed on $address $len"
					info "Breakpoints (addr/value) = [hex_list $rsp::sw_breakpoints]"
					send_packet "OK"
				} else {
					send_packet "E01"
				}
			}
			
			"^Z1,([0-9a-f]+),4" {
				set address 0x[lindex $matches 1]
				if {[is_breakpoint_active]} {
					info "HW Breakpoint not set on $address, because another HW breakpoint is already active"
					send_packet "E01"
				} else	{					
					insert_hw_breakpoint $address
					info "HW Breakpoint set on $address"
					send_packet "OK"
				}
			}
			
			"^z1,([0-9a-f]+),4" {
				set address 0x[lindex $matches 1]
				info "HW Breakpoint cleared on $address"
				remove_hw_breakpoint $address
				send_packet "OK"
			}
			
			"^s$" {	
				info "Single Step"
				single_step
				send_packet "S05"
			}
			
			"^c$" {			
				info "Continue..."
				halt_mcp 0
				poll 1
			}
			
			"^D$" {
				info "Detach requested. Closing connection."
				send_packet "OK"
				close_connection
			}
			
			"^k$" {
				info "Kill target requested. Resetting MCP!"
				reset_mcp
				close_connection
			}
			
			default {
				# This option is not recognized. Return an empty result.
				send_packet ""
			}
		}
	}
	
	# Handle communications: handle state machine to receivce packets CRC, acks etc'. 
	# Call receive_packet after packet received and CRC calculated
	proc handle_comm {} {
		variable chan	
		variable state
		variable checksum
		variable packet
		
		set ch  [read $chan 1]
		
		if {[eof $chan]} {
			info "Connection closed."
			close_connection
			
		} elseif {![fblocked $chan]} {
			# puts -nonewline $ch	
			
			# Some characters are valid in all states ($#).
			# TODO: support escapes
				
			switch $ch {
				"\$" {set state "data"}
				"\#" {set state "checksum"}
				default {					
					switch $state {
						"control" {
						
							# Acks are ignored since we assume that connection is reliable (TCP)
							
							switch $ch {
								"+" {} # This is an ack to our packet. Ignore it
								"-" {error "Got NACK from GDB"}									
								default {
									set ch_value [scan $ch %c]
									
									# Handle interrupt from GDB (char 3)
									
									if {$ch_value==3} {
										info "Interrupt received from GDB!"
                                                                                while 1 {
                                                                                    halt_mcp 1									
                                                                                    if {[is_mcp_halted]} break
                                                                                }
										info "MCP halted."
									} else {
										error "Unexpected control character '$ch' ($ch_value)"
									}
								}
							}
						}
						
						"data" {lappend packet $ch;}
						
						"checksum" {
							lappend checksum $ch					
							if {[llength $checksum] == 2} {
								set calculatedChecksum [calc_checksum $packet]
								set receivedChecksum [join $checksum ""]
								if {[string compare $calculatedChecksum $receivedChecksum] !=0} {
									error "Checksum mismatch: <[hex $calculatedChecksum]> != <[hex $receivedChecksum]>"
								}
								
								# send an ACK to GDB
								puts -nonewline $chan "+"
								
								# Checksum is OK. Handle the packet
								receive_packet
								
								# clear the ground for a new packet
								reset_state
							}
						}
						
						default { error "Unknown state $state";}
					}
				}
			}			
		}
	}
	
########################################################################
# Socket functions
########################################################################

	# Accept new TCP connection from GDB
	proc accept {newChan addr port} {
		variable connected
		variable chan
		
		if {$connected} {
			close $chan ;# only one simultaneous connection supported
		} else {
			set connected 1
			set chan $newChan
			
			if {![is_snapshot]} {
				handle_breakpoint
				remove_hw_breakpoint
			}
			
			fconfigure $chan -buffering none -blocking 0 -translation binary
			fileevent $chan readable \
				[list rsp::handle_comm]          ;# set up to handle for incoming data
		}
	 }                                       
	 
	 # Start the server
	proc start {} {		
		variable server_socket		
		global rsp::RSP_PORT
		global rsp::MAX_RSP_PORT
		
		# Search for available port. See: http://wiki.tcl.tk/%092230
		while {$RSP_PORT <= $MAX_RSP_PORT} {			
			if { [catch {socket -server rsp::accept $RSP_PORT} server_socket] } {         ;# Try to create a server socket			
				info "TCP Port $RSP_PORT is not available."
			} else {				
				break;
			}
			incr RSP_PORT
		}
		
		# -sockname returns address, hostname, and port number
		set RSP_PORT [lindex [fconfigure $server_socket -sockname] 2]		
		rsp::set_snp_gdb_cmd $RSP_PORT
	}
	
	# Stop the server
	proc stop {} {
		variable server_socket
		global batch
		
		close $server_socket		
	}

########################################################################
# MCP snapshot functions
########################################################################

	# Clear MCP snapshot 
	proc unload_snapshot {} {
		variable snp_scratchpad
		variable snp_regfile
		variable snp_regs
		
		set snp_scratchpad {}
		set snp_regfile {}
		set snp_regs [dict create]
	}
	
	# Read an MCP snapshot file (a result of GRC-dump)
	proc load_snapshot {filename} {
		variable snp_scratchpad
		variable snp_regfile
		variable snp_regs
		
		unload_snapshot
				
		set mem_name "" ;# name of current memory variable. If empty, line represents registers and not memory
		
		set fp [open $filename r]
		while { [gets $fp line] >= 0 } {			
			switch -regexp -matchvar matches -- $line {						
			
				"^DATA MCP_REG_SCRATCH"	 	{set mem_name "snp_scratchpad"}
				"^DATA MCP_REG_CPU_REG_FILE" 	{set mem_name "snp_regfile"}				
				"^[[:space:]]*$" 		{set mem_name ""}
					
				default {					
					if {[string length $mem_name] > 0} {
						lappend $mem_name [expr 0x[set line]]						
					} else {
						set reg_tokens 	[regexp -all -inline {\S+} $line]
						set reg_name 	[lindex $reg_tokens 0]
						set reg_address [lindex $reg_tokens 1]
						set reg_size 	[lindex $reg_tokens 2]
						set reg_access	[lindex $reg_tokens 3]
						set reg_value	[lindex $reg_tokens 4]
						
						dict set snp_regs $reg_name $reg_value
					}
				}
			}
		}
		close $fp
	}
	
########################################################################
# Helper functions for running gdb
########################################################################

	proc calcualte_mfw_ver {} {
		
		# Read the MCP version
		
		set misc_reg_shared_mem_addr [rr MISC_REG_SHARED_MEM_ADDR]
		if {[string length $misc_reg_shared_mem_addr] == 0} {return "<Invalid MISC_REG_SHARED_MEM_ADDR>"}
		set public_base [expr ($misc_reg_shared_mem_addr | $rsp::MEM_MASK | $rsp::MCP_REG_SCRATCH) - $rsp::MCP_REG_SCRATCH]		
		set glob_offsize [read_mem [expr $public_base + 4 + 4*2] 4] 									;# u32 glob_offsize = reg_rd(public_base +  4 + (sizeof(offsize_t) * (PUBLIC_GLOBAL)));
		if {[string length $glob_offsize] == 0} {return "<Invalid GLOB_OFFSIZE>"}
		set mfw_ver [read_mem [expr $rsp::MEM_MASK | (($glob_offsize & 0xFFFF)*4) + 24] 4]						;# u32 pub_glob_addr = SECTION_OFFSET(glob_offsize) + GRCBASE_MCP; return pub_glob_addr + 24;
		if {[string length $mfw_ver] == 0} {return "<Invalid MFW_VER>"}
		
		# Build MCP version string
				
		for {set shift 0} {$shift < 32} {incr shift 8} {
			set mfw_ver_element "[expr ($mfw_ver>>$shift) & 0xFF]"
			if {$shift == 0} {
				set mfw_ver_str $mfw_ver_element
			} else {
				set mfw_ver_str "$mfw_ver_element.$mfw_ver_str"
			}
		}
		
		return $mfw_ver_str
	}

}


########################################################################
# User API functions
########################################################################

proc kill_process {pid} {	
	variable tcl_platform
	
	if {$tcl_platform(platform) == "windows"} {
		exec taskkill /F /PID $pid
	} else {
		exec kill $pid 
	}
}

# Debug MCP online using gdb remote debugging
proc debug_mcp { {run {""}} } {
	variable rsp::stop_debugger_event
	global batch
	
	if {[string length $run]} {
		set batch 1
	} else {
		set batch 0
	}
	
        set mfw_ver [rsp::calcualte_mfw_ver]
	rsp::start
	rsp::info ""
	rsp::info "******************************************************"
	rsp::info "* Started MCP debugger stub on port $rsp::RSP_PORT"
        rsp::info "* MFW version: $mfw_ver"
	rsp::info "******************************************************"
	rsp::info ""
	uplevel eval $run 				;# If script is given in $run, run it now after server is up
	vwait rsp::stop_debugger_event                  ;# Enter the event loop	 
	rsp::stop
}

# Debug MCP snapshot (offline) using gdb remote debugging
proc debug_mcp_snapshot { snapshot } {
	
	rsp::load_snapshot $snapshot
	debug_mcp 
	rsp::unload_snapshot
}

# Parse MCP snapshot (run gdb behind the scenes and print stack trace and some variables)
proc parse_mcp_snapshot { {snapshot ""} {mfw_image ""}  {outfile_name ""}} {
	variable rsp::show_rsp_info
	
	#If snapshot is provided, load_snapshot.
	#Otherwise, read registers
	if {[string length $snapshot]} {
		rsp::load_snapshot $snapshot	
	}
	
	# Disable info prints by the RSP server
	set prev_show_rsp_info $rsp::show_rsp_info
	set rsp::show_rsp_info 0
	
	if { [rsp::rr MCP_REG_CACHE_PAGING_ENABLE] == 0 } {
		puts stderr "ERROR: Cannot parse MCP scratchpad! Probably a recovery MFW is running now."
		exit 1
	}


        # If mfw_image is not given, calculate if from MCP version and some env vars
	if {[string length $mfw_image] == 0} {
		set mfw_ver [rsp::calcualte_mfw_ver]
		scan $mfw_ver %d.%d.%d.%d major minor build drop
		if {$drop == 0} {
			# Official MFW version
			set mfw_image "$::env(MCP_VER_ROOT)/$mfw_ver/$::env(MCP_IMAGE_PATH)"
		} else {
			# Bootleg MFW version
			set mfw_image "$::env(MCP_VER_ROOT)/bootlegs/$mfw_ver/$::env(MCP_IMAGE_PATH)"
		}
	}	        

	set gdb_pid 0
	
        # Start the RSP server
	if { [catch {
		debug_mcp {
		
			# Convert GDB commands to command line options using the "-ex" option
			set gdb_cmd_list [split $rsp::SNP_GDB_CMD "\n"]
			set gdb_cmd_options_list {}
			foreach cmd $gdb_cmd_list {
			    set tcmd [string trim $cmd]
			    if {[string length $tcmd] > 0} {
				lappend gdb_cmd_options_list "-ex"
				lappend gdb_cmd_options_list "$tcmd"
			    }
			}
			
			if {[string length $outfile_name] > 0} {				
				lappend gdb_cmd_options_list ">"
				lappend gdb_cmd_options_list "$outfile_name"
			}

			# Execute GDB on the background with all the needed command line options
			set gdb_pid [exec "$::env(MCP_GDB_PATH)" -q -nx $mfw_image --data-directory "$::env(MCP_GDB_DATADIR)" {*}$gdb_cmd_options_list &]
			
		}} debug_mcp_result] } {		
			puts "ERROR: Failed running debug_mcp: $debug_mcp_result\n\nFull info: $::errorInfo"
			if {$gdb_pid > 0} {
				kill_process $gdb_pid
			}
	}

        # Restore prints and unload
	set rsp::show_rsp_info $prev_show_rsp_info
	
	if {[string length $snapshot]} {
		rsp::unload_snapshot
	}
}

if {[info exists ::env(TCL_CMD)]} {
	eval $::env(TCL_CMD)
}
