###########################################################################
#
# Copyright(c) 2014-2017 Cavium, Inc., all rights reserved
# Proprietary and Confidential Information.
#
# Module Description:
# This script tests all the EVERES T memories, based on an RDF file
# which contains a list of registers and memories.
# The script tests several data patterns inorder to verify no bit
# is stuck at either zero or one.
#
# This script uses the regsUtils.tcl script to parse the RDF into
# global arrays: regsOffset regsType regsSize and regsResetVal
# 
###########################################################################

# Specify any dependency on other packages
package require Tcl

# Specify the package this module provides
package provide _memtest 1.0

namespace eval _memtest {
    
    # Export the package
    namespace export memtest    
   
    variable testResult    
   
    set testResult ""
}

# Give a short-name for the regression test.
interp alias {} memtest {}  "_memtest::memtest_main"

#Globals:

#Blocks to skip
global skipBlocks { DBG MCP }

#Registers information in 4 arrays
global regsName
global regsOffset
global regsSize
global regsType
global regsResetVal
global regsProtectRead
global regsProtectWrite
global BB_A0_ExceptionList
global blocksTable

# -----------------------------------------------------------------
# This is the main procedure of the memtest.
# The proc parses the input options, builds the blocks offset table and
# Calls the memory_test procedure which parses the RDF file and
# performs the test for each entry in the memory register.
# -----------------------------------------------------------------
proc _memtest::memtest_main {{args}} {
    #use the global variables   
    global skipBlocks       
    global isFastMode
    global BB_A0_ExceptionList
    global blocksTable
    
    variable testResult
    
    # skip the MCP blocks for now, they are not validated yet.
    set skipBlocks    { MCP }
    set testBlock     ""
    set endBlock      ""
    set returnVal     0
    set skipCommonReg 0   
    set extended  0    
    set testResult {}   
    
    # Parse the options.
    for {set i 0} {$i < [llength $args]} {incr i} {
        switch -- [lindex $args $i] {
            "-f"         { incr i; set rdfFile [lindex $args $i] }
            "-skip"      { incr i; lappend skipBlocks [lindex $args $i] }
            "-b"         { incr i; set testBlock [lindex $args $i] }
            "-e"         { incr i; set endBlock [lindex $args $i] }
            "-skipCommon" { set skipCommonReg 1 }            
            "-extended"   { set extended 1}
            "-help"      { mem_help; return 1 }
            default      { mem_help; return 1 }
        }
    }

    #Build excption list (if not already built)
    if {![info exists BB_A0_ExceptionList]} {
        puts "Build exception list"
        build_exception_list 
    }
    
    # parse the rdf file into regsOffset array
    parse_rdf -skip $skipBlocks
    
    if {$extended} {
        # reset the chip and halt the MCP before testing
        # only in extended mode        
        prep_for_test
    }

    #Clear regFifo results before test
    dbgRegFifo
    
    # run the regtest_test on all the registers in regsOffset
    set returnVal [memory_test $testBlock $skipCommonReg $extended]
   
    # Verify there are no GRC errors
    set regFifoResults [dbgRegFifo]
    
    if {![string match "*fifo contained 0 elements" $regFifoResults]} {
        # mark returnVal as failed
        if {$returnVal == "passed"} {set returnVal "failed"}
        lappend returnVal "ERROR: There are grc errors !\n $regFifoResults"
        puts "ERROR: There are grc errors !\n $regFifoResults"
    }

    #Verify there are no PARITY error
    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS"
    append mask_reg "_REG_PRTY_MASK"

    set offset [reg_offset -n $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n  $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal "ERROR: $prty_reg returned $val (mask $mask)!"
            puts "ERROR: $prty_reg returned $val (mask $mask)!"
        }
    }

    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS_H_0"
    append mask_reg "_REG_PRTY_MASK_H_0"
    set offset [reg_offset -n  $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n  $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal "ERROR: $prty_reg returned $val (mask $mask) !"
            puts "ERROR: $prty_reg returned $val (mask $mask) !"
	}
    }

    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS_H_1"
    append mask_reg "_REG_PRTY_MASK_H_1"
    set offset [reg_offset -n $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal "ERROR: $prty_reg returned $val (mask $mask)!"
            puts "ERROR: $prty_reg returned $val (mask $mask)!"
        }
    }
           
    if {$returnVal == "passed"} {
        puts "***********************************************************"
        puts "**             $testBlock MEMORY TEST PASSED               "
        puts "***********************************************************"        
    } else {
        puts "***********************************************************"
        puts "**             $testBlock MEMORY TEST FAILED !!!!          "
        puts "***********************************************************"        
    } 
    
    if {$extended} {
        # Restore reset values and run MCP
        # only in extended mode
        cleanup_after_test
    }
    
    return $returnVal
}

# -----------------------------------------------------------------
# Print march_1s / march_0s result
# Set testResult variables
# -----------------------------------------------------------------
proc _memtest::print_res {{res}} {
    variable testResult    
    
    if {$res == "Success"} {
        puts "\nmtest passed."
    } else {
        puts "\nmtest failed !! $res"
        lappend testResult $res        
    }    
}

proc _memtest::rw_test {{name} {offset} {value} {dataWidth} } {

    puts "R/W test for memory $name address [format "0x%x" $offset] size $dataWidth value [format "0x%x" $value]"
    
    # Save the number of words (32 bit) of the register
    set noWords [expr ($dataWidth/32)]
    if { [expr ($dataWidth % 32) != 0] } {
        incr noWords
    }

    # Save the register's original value and mask in arrays
    # (Incase of WB each dword of the register's original value is saved in an array orgVal).
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]
        set orgVal($i,value) [reg read $wordOffset]
                     
        # Create the mask according to the register's dataWidth
        if {$i < [expr ($noWords-1)]} {
            set mask($i,value) 0xFFFFFFFF            
        } else  {
            set mask($i,value) [get_mask $dataWidth]             
        }
    }

    # Write the requested value
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]
        if {$i < [expr ($noWords-1)]} {
            set writtenVal($i,value) $value
            reg write $wordOffset $value            
        } else  {
            # write to the remaining part of the register
            # only to the valid bits (according to mask)
            set writtenWord [expr {$value & $mask($i,value)}]            
            set writtenVal($i,value) $writtenWord
            reg write $wordOffset $writtenWord            
        }
    }

    # Read and compare to expected value
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr {$offset + $i*4}]
        set readVal($i,value) [reg read $wordOffset]
        set readWord   [expr {$readVal($i,value) & $mask($i,value)}]
                        
        set writeMask  [write_mask -n $name]
        set writtenWord [expr {$writtenVal($i,value) & $writeMask}]
               
        if {$readWord != $writtenWord} {
            set res "\nERROR: Check memory $name address [format "0\x%x" $wordOffset] failed (wrote [format "0x%x" $writtenWord] read [format "0x%x" $readWord])"                    
            return $res                   
        } 
    }

    # Restore original value
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]        
        reg write  $wordOffset $orgVal($i,value)
    }

    set res "Success"
    return $res
}

# -----------------------------------------------------------------
# This proc parses each line in the rdf file and performs for each 
# memory the marching ones and marching zeros tests
# -----------------------------------------------------------------
proc _memtest::memory_test { {testBlock} {skipCommonReg} {extended}}  {
        global regsName
        global regsOffset
        global regsSize
        global regsType
        global regsResetVal
        global regsSplitVal
        global regsPhase        
        global regsProtectRead
        global regsProtectWrite        
        
        variable testResult
        variable isEmul
        variable isFpga
        
        # Initialize variables
        set line         0              
        set offset       0
        set dataWidth    32
        set noMemories   0
        set noEntries    0        
        set curFunc     $::current(FUNC)
        set isEmul 0
        set INIT_DELAY  240

        set isEmul [is_emul]
       
               
        # Go through each line in the regsOffset array        
        for {set idx 0 } {$idx < [llength $regsName]} {incr idx} {            
            set name [lindex $regsName $idx]

            # Skip lines which are not in the tested block 
            set tmpStr [format "%s_REG_*" $testBlock]
            if { ($testBlock != "") && (![string match "$tmpStr" $name]) } {
                continue
            }
            if { [string match "*_STATISTICS" $name] } {
                    set is_stat 1
            } else {
                    set is_stat 0
            }

            # Get the registers information: offset, dataWidth, accessType
            set addrRange [lindex $regsOffset $idx]
            set dataWidth [lindex $regsSize $idx]
            set accessType [lindex $regsType $idx]
            set resetVal [lindex $regsResetVal $idx]
            set splitVal [lindex $regsSplitVal $idx]
            set protectRead [lindex $regsProtectRead $idx]
            set protectWrite [lindex $regsProtectWrite $idx]            
            
            # Skip unused memories and exceptional memories            
            if { [skip_register $name] } {
                continue
            }

            # Skip the lines which are not memory, meaning their offset
            # information is not a range startOffset-endOffset
            if { ([string first "-" $addrRange] == -1) } {
                continue
            }
            
            # Skip memories which are read-only or write-only
            if {($accessType == "WR") || ($accessType == "W") || ($accessType == "R") ||
                ($accessType == "WB_R") || ($accessType == "WB_W") || ($accessType == "ST")} {
                continue
            }
	    
	    set mode [nvm secure_mode]

            # Skip memories with protectWrite HV, UA
            if {(($mode == "Secure Mode") || ($mode == "Fully Secure Mode")) && ($protectWrite != "PDA")} {
                #puts "Skip memory $name cause of protect write."
                continue
            }
            
            # Check if need to test this memory register
            set b_run_test [check_split_reg $skipCommonReg $splitVal $curFunc]            
            
            # Perform reset test to the register
            if {$b_run_test} {
                # For WB regs, calc the reg reserved width (the upper 2^x)
                if {$dataWidth > 32} {                
                    set power [expr (log($dataWidth)/log(2))]
                    #round up the power value
                    if { $power != [expr (int($power))]} {
                        set power [expr (int($power))]
                        incr power
                    }
                    set width [expr (int(pow(2,$power)))]
                    set noWords [expr ($width/32)]
                    set noBytes [expr {4*$noWords}]
                } else  {
                    set noWords 1
                    set noBytes 4
                }                   
            
                # Initialize the memory with all zeros
                #puts "Debug: memory_test initialize $name $addrRange"
                set memory_start_address [get_start_offset $addrRange]
		set memory_end_address [get_end_offset $addrRange]
               
                # Count the number of memories tested    
                incr noMemories 
                
                # Special handling for *CFC_REG_CID_RAM in emulation
                if { [is_emul] } {
                    if {([string match "CCFC_REG_CID_CAM" $name] || 
                         [string match "TCFC_REG_CID_CAM" $name]) } {
                         set memory_end_address [expr ($memory_start_address + 0x2F8)]
                         set noWords 192
                         puts "NOTE: $name size in emulation is 192 dwords only"
                    }

                    if {([string match "CCFC_REG_STRING_CAM" $name] ||
                         [string match "TCFC_REG_STRING_CAM" $name])} {

                         set memory_end_address [expr ($memory_start_address + 0x17c)]
                         set noWords 96
                         puts "NOTE: $name size in emulation is 96 dwords only"
                    }
                }
                
                if {!$extended} {
                    set res [rw_test $name $memory_start_address 0x55555555 $dataWidth]
                    print_res $res 

                    set res [rw_test $name $memory_start_address 0xAAAAAAAA $dataWidth]
                    print_res $res
                    
                } else {                    
                    mtest init $memory_start_address $memory_end_address $noWords
                    if {$isEmul} {delay ms $INIT_DELAY}
                    
                    # Perform the marching ones test on the memory
                    puts "memory_test march_1s $name [format 0x%x $memory_start_address]-[format 0x%x $memory_end_address]"                    
                    set res [mtest march_1s $memory_start_address $memory_end_address $dataWidth $noWords $accessType]                   
                    print_res $res 
                    
                    # Perform the marching zeroes test on the memory
                    puts "memory_test march_0s $name [format 0x%x $memory_start_address]-[format 0x%x $memory_end_address]"                    
                    set res [mtest march_0s $memory_start_address $memory_end_address $dataWidth $noWords $accessType $is_stat]                    
                    print_res $res 
                }               
            }

            #By default (not extended mode), test only one memory per block
            if {!$extended && ($noMemories >= 1)} {
                break;
            }
        }        
               
        if {$testResult == "" } {set testResult "passed" }
        
        if { $noMemories != 0 } {
            puts "Tested $noMemories memories\n"
        } else {
            puts "Warning: no memories found\n"
            set testResult "passed"   
        }      
        
        return $testResult        
    }
    
# -----------------------------------------------------------------
# This proc prints out the usage information
# -----------------------------------------------------------------
proc _memtest::mem_help {}  {
    
    puts stdout "Usage:   memtest -b <BLOCK>              "
    puts stdout "optional:                                "
    puts stdout "          -extended: runs extended tests "
    puts stdout "          -f     <rdf file name>         "
    puts stdout "          -skip  <skip blocks>           "
    puts stdout "          -e     <end block>             "
    puts stdout "          -skipCommon                    "
    puts stdout "Note: if no rdf file is specified, the test  "
    puts stdout "will use the rdf under the current directory."
    return 1
}

# -----------------------------------------------------------------
# This proc resets the chip and makes the initializations required.
# -----------------------------------------------------------------
proc _memtest::prep_for_test {}  {       
    variable isEmul
    variable isFpga

    puts "Run _memtest::prep_for_test"
    set isEmul [is_emul]
    set isFpga [is_fpga]

    # Reset the chip
    if {!$isEmul && !$isFpga} {
        mcp halt 
    }
    reset_chip 1

    if {$isEmul} {
        delay ms 3000
    }

    # This initialization is done in dev_if_init
    #reg_write -n PGLUE_B_REG_START_INIT_PTT_GTT -v 1
    #reg_write -n PGLUE_B_REG_START_INIT_INB_INT_MEM -v 1
    
    #Dummy read required for emulation
    if {$isEmul} {
	delay ms 10
        reg read 0x8000	
    }
}

# -----------------------------------------------------------------
# This proc cleans up after tests
# -----------------------------------------------------------------
proc _memtest::cleanup_after_test {}  {  
  variable isEmul
  variable isFpga

  puts "Run _memtest::cleanup_after_test"

  # Avoid driver load failure cause of this parity after PSWHST memory test
  set val [reg_read -n PSWHST_REG_PRTY_STS_H_0]
  if {$val != 0} {
      reg_write -n PSWHST_REG_PRTY_STS_WR_H_0 -v $val
  }

  if {$isEmul} {
      reset_chip 1
      delay ms 3000
  } elseif {$isFpga} {
  	reset_chip 1
  } else {
      mcp run
      reset_chip
  }
}


# -----------------------------------------------------------------
# This proc parses the start offset out of an offset field
# (startOffset - endOffset)
# -----------------------------------------------------------------
proc _memtest::get_start_offset {{addrRange}}  {
    
    # Parse out the starting offset
    return [lindex [split $addrRange "-"] 0]
}

# -----------------------------------------------------------------
# This proc parses the start offset out of an offset field
# (startOffset - endOffset)
# -----------------------------------------------------------------
proc _memtest::get_end_offset {{addrRange}}  {
    
    # Parse out the ending offset
    return [lindex [split $addrRange "-"] 1]
}






