###########################################################################
#
# Copyright(c) 2014-2017 Cavium, Inc., all rights reserved
# Proprietary and Confidential Information.                                
#
###########################################################################

# Specify any dependency on other packages
package require Tcl

# Specify the package this module provides
package provide _pda_test 1.0

namespace eval _pda_test {
    
    # Export the package
    namespace export pda_test   
}

# Give a short-name for the regression test.
interp alias {} pda_test {}  "_pda_test::pda_test_main"
interp alias {} pda_win_test {}  "_pda_test::test_grc_window"
interp alias {} set_privilege_level {}  "_pda_test::set_privilege_level"

#Globals:

#Blocks table is an array of block name & block offset
global blocksTable

#Registers information from global arrays
global regsName
global regsOffset
global regsSize
global regsType
global regsProtectWrite
global regsResetVal

# Priviledge levels
global PDA_LEVEL
global HV_LEVEL
global UA_LEVEL


# -----------------------------------------------------------------
# This is the main procedure of the pda_test:
# - Runs the PDA test with three priviledge levels (PDA, HV, UA).
# - Runs each test on all blocks.
# - If a block is in the fullCheckBlocks list then all registers of
#   this block are tested. (full check).
# - Otherwise only one PDA register and all HV,UA registers are
#   tested. (fast check)
# -----------------------------------------------------------------
proc _pda_test::pda_test_main {{args}} {

    #Use global arrays
    global blocksTable
    global skipBlocks
    global regsName
    global regsOffset
    global regsSize
    global regsType    
    global regsProtectWrite
    global regsWriteMask
    global regsResetVal
    
    # Privilege levels
    global PDA_LEVEL
    global HV_LEVEL
    global UA_LEVEL
    set PDA_LEVEL 1
    set HV_LEVEL  2
    set UA_LEVEL  3
    
    # On these blocks to all registers should be tested
    # TBD - add CNIG, MCP to this list, currently some registers fail.
    set fullCheckBlocks { MISC MISCS NIG PSWRQ PSWWR PGLUE_B OPTE ATC BMB DBU IPC GRC}
        
    set returnVal     0
       
    # Parse the RDF file    
    parse_rdf
    
    # Build the list of registers to skip
    build_exception_list  
                 
    # Parse the options.
    for {set i 0} {$i < [llength $args]} {incr i} {
        switch -- [lindex $args $i] {                                               
            "-help"      { help; return 1 }
            default      { help; return 1 }
        }
    }
    
    # Build the blocks table (block name, block offset)
    # from the rdf file.
    build_blocks_table    
        
    # Reset the chip and halt the MCP before testing    
    prep_for_test     
        
    # Run PDA test with three different privilege levels
    for {set level $PDA_LEVEL} {$level <= $UA_LEVEL} {incr level} {
        puts "Testing PDA level $level"
        
        #Set privilege level according to the tested level
        set_privilege_level $level       
                
        # Un parse the RDF file, test will parse only the registers
        # of the tested block
        un_parse_rdf
        
        foreach blockName [lsort [array names blocksTable]] {
            
            # Skip blocks
            set skipBlocks { MCP GRC}
            if { [lsearch -exact $skipBlocks $blockName] != -1 } {
                continue
            }
            
            # Parse the registers of this block from the RDF file into
            # regsOffset array           
            parse_rdf -b $blockName
            
            # Check if block is in the list of blocks to be fully tested.
            if  { [lsearch -exact $fullCheckBlocks $blockName] != -1 } {
                puts "Running PDA full check on block: $blockName"
                set isFullCheck 1
            } else {
                puts "Running PDA fast check on block: $blockName"
                set isFullCheck 0
            }
            
            # Run the privilege test on this block
            if {[test_privilege_level $isFullCheck $level] != 0 } {
                #set returnVal -1
                return -1
            }
            
            un_parse_rdf             
        }
        
        # Parse the RDF file    
        parse_rdf
    }
    
    # Run the test_grc_window
    if {[test_grc_window] != 0 } {
        return -1
    }
    
    if {$returnVal != 0} {                
        puts "***********************************************************"
        puts "**             PDA TEST FAILED !!!!                        "
        puts "***********************************************************"
              
    } else {
        puts "***********************************************************"
        puts "**             PDA TEST PASSED                             "
        puts "***********************************************************"
    }
              
    # Restore reset values
    reset_chip          
    return $returnVal
}

# -----------------------------------------------------------------
# This proc prints out the usage information
# -----------------------------------------------------------------
proc _pda_test::help {}  {
    
    puts stdout "Usage:   pda_test "
}

# -----------------------------------------------------------------
# Reset the chip and make the initializations required.
# -----------------------------------------------------------------
proc _pda_test::prep_for_test {}  {    
      
    # Reset the chip
    reset_chip 1
}

# -----------------------------------------------------------------
# Set privilege level according to the requested level
# -----------------------------------------------------------------
proc _pda_test::set_privilege_level {level}  {
   
    set PGLUE_B_REG_PF_BAR_PRIVILEGE 0x2aaea4
    
    if {[is_e4_BigBear]} {
        set PCICFG_GRC_ADDR 0x78
        set PCICFG_GRC_DATA 0x80
    } else {
        set PCICFG_GRC_ADDR 0xe0
        set PCICFG_GRC_DATA 0xe4
    }

    pcicfg write $PCICFG_GRC_ADDR $PGLUE_B_REG_PF_BAR_PRIVILEGE
    pcicfg write $PCICFG_GRC_DATA $level
            
    set val [reg read $PGLUE_B_REG_PF_BAR_PRIVILEGE]
}

# -----------------------------------------------------------------
# -----------------------------------------------------------------
proc _pda_test::test_privilege_level {{isFullCheck} {pdLevel}} {        
    global regsName
    global regsOffset
    global regsSize
    global regsType       
    global regsProtectWrite
    global regsWriteMask
    global regsResetVal
    
    global PDA_LEVEL
    global HV_LEVEL
    global UA_LEVEL

    # Initialize variables       
    set testResult   ""
    set offset       0
    set dataWidth    32
    set rc 0
    
    # Flag is 1 if at least one PDA register was already tested
    set pdaRegTested 0
                                
    # Go through each register in this block:
    # Note: Incase of fast check (!isFullCheck) check only one register
    # with PDA level and all registers with HV, UA level.
    for {set idx 0 } {$idx < [llength $regsName]} {incr idx} {
           
        # Check if need to skip this register cause its in the exception list
        set name [lindex $regsName $idx]                           
        if { [skip_register $name]} {
            continue
        }
        
        # Do not test Write-Only or Read-Only registers
        set accessType [lindex $regsType $idx]        
        if {($accessType == "WR") || ($accessType == "W") ||
            ($accessType == "R") || ($accessType == "RC") || ($accessType == "WB_R")} {
            continue
        }
                        
        # The write protection level is higher than read protection level
        # so we use this higher protection level for the check.
        set regProtectLevel [lindex $regsProtectWrite $idx]
        switch $regProtectLevel { 
	    UA
            { set regProtectLevel $UA_LEVEL }
            HV
            { set regProtectLevel $HV_LEVEL }
            PDA
            { set regProtectLevel $PDA_LEVEL }
        }
        
        # Test this register if in FullCheck mode or
        # register protect level > PDA or at least one PDA register check. 
        if {$isFullCheck ||
            ($regProtectLevel == $UA_LEVEL) || ($regProtectLevel == $UA_LEVEL) ||
            (($regProtectLevel == $PDA_LEVEL) && !$pdaRegTested)} {               
                                
            # Skip memories (offset field is a range startOffset - endOffset)            
            set offset [lindex $regsOffset $idx]
            if { ([string first "-" $offset] != -1) } {
                continue           
            }
            
            set dataWidth [lindex $regsSize $idx]
            set writeMask [lindex $regsWriteMask $idx]
            set resetVal [lindex $regsResetVal $idx]
                
            # Attempt to write to this register            
            set testResult [reg_access_check $offset $dataWidth $writeMask $resetVal]
            
            # If pdLevel is higher than register protection level,
            # the reg_access_check should pass. Otherwise return failure.
            if {($pdLevel >= $regProtectLevel) && ($testResult != "")} {
                puts "Tesing $name failed !! pdLevel ($pdLevel) is higher than register protection level ($regProtectLevel) \
                      register access should have passed, but got $testResult"
                #set rc -1
                return -1               
            }
            
            # If  pdLevel is lower than register protection level,
            # the reg_access_check should fail. Otherwise return failure.
            if {($pdLevel < $regProtectLevel) && ($testResult == "")} {
                puts "Tesing $name failed !! pdLevel ($pdLevel) is lower than register protection level ($regProtectLevel) \
                      register access should have failed !"
                #set rc -1
                return -1
            }
            
            # Mark that one pda register was tested, no need to check more
            # pda registers in non-full mode.
            if {$regProtectLevel == $PDA_LEVEL} {                
                set pdaRegTested 1
            }
        } 
    }        
        
    return $rc  
}


# -----------------------------------------------------------------
# -----------------------------------------------------------------
proc _pda_test::reg_access_check {{offset} {dataWidth} {writeMask} {resetVal}} { 
    
    set testResult ""
    
    set noWords [expr ($dataWidth/32)]
    if { [expr ($dataWidth % 32) != 0] } {
        incr noWords
    }                                                                                                                       
                        
    # For WB regs, calc the reg reserved width (the upper 2^x)
    if {$dataWidth > 32} {
        set power [expr (log($dataWidth)/log(2))]
        #round up the power value
        if { $power != [expr (int($power))]} {
            set power [expr (int($power))]
            incr power
        }
        set width [expr (int(pow(2,$power)))]
        set noWords [expr ($width/32)]
    } else  {
        set noWords 1
    }
       
    # Save the register's original value and mask in arrays
    # incase of WB (wide bus) each part (32bit) of the register's original
    # value is saved in an array orgValue.
    # Otherwise (dataWidth<32bit) original value is in 1'st element of the array.
    if {$noWords > 1} {
        for {set i 0} {$i < $noWords} {incr i} {
             set wordOffset [expr ($offset + $i*4)]
             set orgValue($i,value) [reg read $wordOffset]
                         
             # Create the mask according to the register's dataWidth
             if {$i < [expr ($noWords-1)]} {
                set mask($i,value) 0xFFFFFFFF          
             } else  {
                 # The last part of the register's mask is according to remaining
                 # data width e.g 100 bit register, mask is for remaining 4 bits (0x1111)
                 set mask($i,value) [get_mask $dataWidth]            
             }         
        }
    } else {
        set mask(0,value) $writeMask
        set orgValue(0,value) [reg read $offset]
    }
    
    # Set the test value to 0xffffffff or 0x55aa55aa according to resetVal
    if {$resetVal != "X"} {
        if {[expr {$resetVal & $writeMask}] == [expr {0xfffffff & $writeMask}]} {        
            set val 0x55aa55aa
        } else {
            set val 0xffffffff
        }
    } else {
        set val 0xffffffff
    }
           
    # Write to the register words the value 0xffffffff
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]
        if {$i < [expr ($noWords-1)]} {           
            set writtenValue($i,value) $val
            reg write $wordOffset $val           
        } else  {
            # write to the remaining part of the register
            # only to the valid bits (according to mask)
                       
            set writtenWord [expr {$val & $mask($i,value)}]            
            set writtenValue($i,value) $writtenWord
            reg write $wordOffset $writtenWord            
        }
    }

    # Read the register's value
    # Incase of WB go through all its words (32bit each)
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr {$offset + $i*4}]
        set readValue($i,value) [reg read $wordOffset]
        
        set readWord     [expr {$readValue($i,value) & $mask($i,value)}]
        set writtenWord  $writtenValue($i,value)
                   
        # In case of access failure, exit with error
	if {$readWord == "bad0_bad0"} {
		set testResult "bad0_bad0"
		break
	}
        
        # read-write register read value should be as written value
        # Ignore the reseved bits according to register's write_mask
        if {$readWord != [expr {$writtenWord & $writeMask}]} {
            set testResult "Read value [format "0x%x" $readWord] is different \
                            than expected [format "0x%x" $writtenWord]"                        
            break
        }
    }
    
    # Restore the original value of the register's words
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]        
        reg write  $wordOffset $orgValue($i,value)
    }    
    
    return $testResult
}

proc _pda_test::test_grc_window {} {
    
    #Use global arrays
    global blocksTable
    global skipBlocks
    global regsName
    global regsOffset
    global regsSize
    global regsType    
    global regsProtectWrite
    global regsWriteMask
    global regsResetVal
    
    # Privilege levels
    global PDA_LEVEL
    global HV_LEVEL
    global UA_LEVEL
    set PDA_LEVEL 1
    set HV_LEVEL  2
    set UA_LEVEL  3
    
    set INT_MEM_OFFSET 0x20000
    set int_mem_size   0x2
    set mcp_scratch_mem_size 0x2
    
    set val1 0x55aa55aa
    set val2 0xffffffff
    
    parse_rdf
    
    reset_chip

    if {[is_e4_BigBear]} {
        set PCICFG_GRC_ADDR 0x78
        set PCICFG_GRC_DATA 0x80
    } else {
        set PCICFG_GRC_ADDR 0xe0
        set PCICFG_GRC_DATA 0xe4
    }
    
    #Set privilege level to PDA
    set_privilege_level $UA_LEVEL
    
    puts "halt mcp"
    set val [reg_read -n MCP_REG_CPU_MODE]
    set val [expr {$val | (1 << 10)}]
    reg_write -n MCP_REG_CPU_MODE -v $val
    reg_write -n MCP_REG_CPU_STATE -v 0xffffffff
        
    #Set privilege level to PDA
    set_privilege_level $PDA_LEVEL      
    
    #Write to TSTORM internal memory two values and make sure
    #read value is as written value
    puts "Write values to TSEM internal memory"
    set offset [reg_offset -n TSEM_REG_FAST_MEMORY] 
    set int_mem_offset [expr {$offset + $INT_MEM_OFFSET}]
    
    reg write $int_mem_offset $val1
    set res [reg read $int_mem_offset]        
    if {$res != $val1} {
        puts "ERROR: Read value from TSEM internal memory $res is different than expected $val1"
        return -1
    }
    
    reg write $int_mem_offset $val2
    set res [reg read $int_mem_offset]        
    if {$res != $val2} {
        puts "ERROR: Read value from TSEM internal memory $res is different than expected $val2"
        return -1
    }
          
    puts "Read values from TSEM internal memory as expected."
    set no_win_reg_offset [reg_offset -n GRC_REG_NUMBER_VALID_OVERRIDE_WINDOW]
    pcicfg write $PCICFG_GRC_ADDR $no_win_reg_offset
    set no_windows [pcicfg read $PCICFG_GRC_DATA]     
    puts "Current no. of windows is $no_windows"
    if {$no_windows == 20} {
        puts "ERROR: All GRC windows are in-use !"
        return -1
    }

    # Calulate the offset of GRC_REG_PROTECTION_OVERRIDE_WINDOW
    # at position {no_windows+1}.
    # each window is of 2 registers (64 bit)
    set win_reg_offset [get_start_offset [reg_offset -n GRC_REG_PROTECTION_OVERRIDE_WINDOW]]
    
    set offset [expr {$win_reg_offset + $no_windows * 8}]
    puts "DEBUG: window register offset [format 0x%x $offset]"
    
    # Write to the #<no_windows+1> offset the following data:
    # wr_protect_level (54:52) = 6 for HV| rd_protect_level (51:49) = 6 for HV |
    # wr_access (48) | rd_access (47) |
    # int_mem_size (46:23) | int_mem_offset (22:0)
    set int_mem_msb [expr {$int_mem_offset >> 2} & 0x7fffff]    
    puts "DEBUG: int_mem_msb [format 0x%x $int_mem_msb]"
    set data [expr {$int_mem_msb + ($int_mem_size << 23) + \
              (1 << 47) + (1 << 48) + (6 << 49) + (6 << 52)}]
        
    set data_low [expr {$data & 0xffffffff}]
    set data_high [expr {($data & 0xffffffff00000000) >> 32}]
    
    #Indirect write to GRC_REG_PROTECTION_OVERRIDE_WINDOW
    puts "Indirect write to GRC_REG_PROTECTION_OVERRIDE_WINDOW [format 0x%x $data_high] [format 0x%x $data_low]"
    pcicfg write $PCICFG_GRC_ADDR $offset
    pcicfg write $PCICFG_GRC_DATA $data_low
    
    pcicfg write $PCICFG_GRC_ADDR [expr {$offset+4}]
    pcicfg write $PCICFG_GRC_DATA $data_high           
      
    # Update the no. of windows register
    puts "Indirect write to GRC_REG_NUMBER_VALID_OVERRIDE_WINDOW"    
    pcicfg write $PCICFG_GRC_ADDR $no_win_reg_offset
    pcicfg write $PCICFG_GRC_DATA [expr {$no_windows+1}]

    # Now access to TSEM internal memory should fail
    reg write $int_mem_offset $val1
    set res [reg read $int_mem_offset]        
    if {$res == $val1} {
        puts "ERROR: Access to TSEM internal memory should have failed"
        return -1
    }
    puts "DEBUG: Access to TSEM internal memory denied as expected."    
        
        
    set mcp_scratch_offset [reg_offset -n MCP_REG_SCRATCH]
    set mcp_scratch_start_offset [get_start_offset $mcp_scratch_offset]
    
    puts "Write values to MCP_REG_SCRATCH [format "0x%x" $mcp_scratch_start_offset] [format "0x%x" $val1]"
    reg write $mcp_scratch_start_offset $val1
    set res [reg read $mcp_scratch_start_offset]        
    if {$res == $val1} {
        puts "ERROR: Access to MCP scratchpad should have failed"
        return -1
    }
    
    pcicfg write $PCICFG_GRC_ADDR $no_win_reg_offset
    set no_windows [pcicfg read $PCICFG_GRC_DATA]     
    puts "Current no. of windows is $no_windows"
    if {$no_windows == 20} {
        puts "ERROR: All GRC windows are in-use !"
        return -1
    }    
    set offset [expr {$win_reg_offset + $no_windows * 8}]
    puts "DEBUG: window register offset [format 0x%x $offset]"
    
    # Write to the #<no_windows+1> offset the following data:
    # wr_protect_level (54:52) = 5 for PDA| rd_protect_level (51:49) = 5 for PDA|
    # wr_access (48) | rd_access (47) |
    # int_mem_size (46:23) | int_mem_offset (22:0)
    set mcp_scratch_mem_msb [expr {$mcp_scratch_start_offset >> 2} & 0x7fffff]    
    puts "DEBUG: mcp_scratch_mem_msb [format 0x%x $mcp_scratch_mem_msb]"
    set data [expr {$mcp_scratch_mem_msb + ($mcp_scratch_mem_size << 23) + \
              (1 << 47) + (1 << 48) + (5 << 49) + (5 << 52)}]
        
    set data_low [expr {$data & 0xffffffff}]
    set data_high [expr {($data & 0xffffffff00000000) >> 32}]
    
    #Indirect write to GRC_REG_PROTECTION_OVERRIDE_WINDOW
    puts "Indirect write to GRC_REG_PROTECTION_OVERRIDE_WINDOW [format 0x%x $data_high] [format 0x%x $data_low]"
    pcicfg write $PCICFG_GRC_ADDR $offset
    pcicfg write $PCICFG_GRC_DATA $data_low
    
    pcicfg write $PCICFG_GRC_ADDR [expr {$offset+4}]
    pcicfg write $PCICFG_GRC_DATA $data_high           
      
    # Update the no. of windows register
    puts "Indirect write to GRC_REG_NUMBER_VALID_OVERRIDE_WINDOW"    
    pcicfg write $PCICFG_GRC_ADDR $no_win_reg_offset
    pcicfg write $PCICFG_GRC_DATA [expr {$no_windows+1}]   
        
    # Now access to MCP scratch memory should pass
    puts "Write values to MCP_REG_SCRATCH [format "0x%x" $mcp_scratch_start_offset] [format "0x%x" $val1]"
    reg write $mcp_scratch_start_offset $val1
    set res [reg read $mcp_scratch_start_offset]        
    if {$res != $val1} {
        puts "ERROR: Access to MCP scratch memory should have passed"
        return -1
    }
    puts "GRC protection window test passed !"
    return 0
    
}
