################################################################################
#
# Copyright(c) 2014-2019 Marvell Technology Group Ltd., all rights reserved
# Proprietary and Confidential Information.
#
# Module Description:
# Display real-time Storm CPU usage information:
# Cpu usage counters, BRB drops, pause packets.
#
# Version number: 1.2.1 (correct version number is defined @ g_profilerVersion)
# Last update 01/09/19 dbalandin
################################################################################

#globals
global g_profileDevModeOn
set g_profileDevModeOn 0

global g_profilerVersion
set g_profilerVersion "1.2.1"

global g_scriptOwner
set g_scriptOwner "dbalandin"

################################################################################
# DEVELOPER FUNCTIONS ##########################################################
################################################################################
#re-source files (only in ediag)
proc prefresh {} {
    set files {"scripts/profiler.tcl" "scripts/regutils.tcl" "scripts/utils.tcl"}

    #if in diag
    if {[isDiag] == 1} {
        foreach file $files {
            #make sure files exist
            if { [ file exists $file ]} {
                source $file
            } else {
                puts "File $file does not exist, skipping."
            }
        }
    }
}

proc devMsg {msg {useColor 0} {color none}} {
    global g_profileDevModeOn
    if {!$g_profileDevModeOn} {return}
    if {$useColor} {
        cPuts $msg $color
    } else {
        puts $msg
    }
}

################################################################################
# PROFILER FUNCTIONS ###########################################################
################################################################################
#check if we're in diag
proc isDiag {} {
    if {[info exist ::current(BUS)]} {
            set bus [expr $::current(BUS)]
            if {$bus != ""} {
                return 1
            }
    }

    return 0
}

# Start of run splash screen
proc profileGreetUser {} {
    global g_profilerVersion
    global g_profileDevModeOn
    global g_arrChipTypesReverse
    global g_chipModeName
    global g_chipType
    global g_frequency
    global g_scriptOwner
    global g_storms
    global g_iterationTime
    global rdfFile

    puts     "********************************************************************************"
    if {$g_profileDevModeOn} {
        puts "*** PROFILER DEVELOPER MODE - For official version please contact $g_scriptOwner ***"
    }
    set defaultValueColor magenta
    puts "** Profiler version : [cStr $g_profilerVersion $defaultValueColor]"
    puts "** Chip type        : [cStr $g_arrChipTypesReverse($g_chipType) $defaultValueColor]"
    puts "** Chip mode        : [cStr $g_chipModeName $defaultValueColor]"
    puts "** Selected storms  : [cStr [dict keys $g_storms] $defaultValueColor]"
    puts "** Iteration time   : [cStr "$g_iterationTime (ms)" $defaultValueColor]"

    set freqMhz [expr double($g_frequency)/1000000]
    if {$freqMhz < 1} {
        #Emulation frequency is usually below 1MHz, requiring different format resolution
        set mhzFormat "%1.4f"
    } else  {
        set mhzFormat "%4.0f"
    }

    set rdfFullPath $rdfFile
    catch {set rdfFullPath [exec realpath $rdfFile]}
    set frequencyString [format "${mhzFormat}MHz (${g_frequency}Hz)" $freqMhz]
    puts "** Chip frequency   : [cStr $frequencyString $defaultValueColor]"
    puts         "** RDF file         : [cStr $rdfFullPath $defaultValueColor]"
    puts         "********************************************************************************"
}

# profiler main function
proc profile {args} {
    prefresh
    global g_profileDevModeOn
    global rdfFile
    global g_iterationTime
    global g_pciRegs
    global g_regsText
    global g_rxHistRegs
    global g_txHistRegs
    global g_pxpHistRegs
    global g_extHistRegs
    global g_frequency
    global g_chipType
    global g_arrChipTypes
    global g_emuClockRatio
    global g_bucketType
    global g_decimalHistVals
    global g_activeStorms
    global g_newline
    global g_stormUtilColorScheme
    global g_noColor

    #Parse input arguments ans set variables appropriatley
    set parsedArgs [parseArgs $args];

    #Chip frequency - used for script calculations
    set g_frequency       [lindex $parsedArgs  0]
    #number of iterations - can be 10000000 if '-forever'
    set iterations        [lindex $parsedArgs  1]
    #mode 1 '-f'
    set isFile            [lindex $parsedArgs  2]
    #outpuf logfile, fileName specified in 1st mode: '-f <fileName>'
    set logname           [lindex $parsedArgs  3]
    #mode 2 '-h' (histogram) or mode 3 '-v' (histogram + csv)
    set isHist            [lindex $parsedArgs  4]
    #histogram filename, 2nd arg used in '-h <histname>'
    set histname          [lindex $parsedArgs  5]
    #number of samples per iteration, set in '-s <samples>'
    set samples           [lindex $parsedArgs  6]
    #mode 3 '-v' (histogram + csv)
    set isDump            [lindex $parsedArgs  7]
    #run on 2 engines if on '-d'
    set isDevices         [lindex $parsedArgs  8]
    #device0
    set dev0              [lindex $parsedArgs  9]
    #device1 when isDevices(2 engines) is on
    set dev1              [lindex $parsedArgs 10]
    #user defined Arrowhead '-ah'
    set usrSetAH          [lindex $parsedArgs 11]
    #user defined BlackBear '-bb'
    set usrSetBB          [lindex $parsedArgs 12]
    #user defined Everest5 '-e5'
    set usrSetE5          [lindex $parsedArgs 13]
    #mode 4 '-sc' (stall count)
    set isStallCount      [lindex $parsedArgs 14]
    #for usage with stall-count mode(4) - display realtime CPU utilization
    set displayUtil       [lindex $parsedArgs 15]
    #mode 5 '-ps' (PCIe core Statistics)
    set isPciStats        [lindex $parsedArgs 16]
    #Keep results from all iterations in the terminal (instead of overriding)
    set keepIters         [lindex $parsedArgs 17]
    #If external register list was given
    set isExt             [lindex $parsedArgs 18]
    #name of external register list file if one was given
    set externalFileName  [lindex $parsedArgs 19]
    #Clock ratio for emulations
    set g_emuClockRatio   [lindex $parsedArgs 20]
    #Type of histogram bucketing algorithm (simple or weighted)
    set g_bucketType      [lindex $parsedArgs 21]
    #If set to 1, default histogram format for values is decimal instead of hex
    set g_decimalHistVals [lindex $parsedArgs 22]
    #Storms to analyze, default is TMUXYP
    set g_activeStorms    [lindex $parsedArgs 23]
    #If true - storm utilization will be printed in new lines
    set g_newline         [lindex $parsedArgs 24]
    #if true, storm utilization will be printed as a table instead of a line
    set isVert            [lindex $parsedArgs 25]
    #if true, strips all colors
    set g_noColor         [lindex $parsedArgs 26]

    #In case of stallCount mode with '-u' flag (display storm util) - the line
    #deletion is handled by the stallCount proc
    set g_newline [expr $g_newline | $isStallCount]

    #Filename for CSV dump, relative for mode 3
    #(sample regs and save histogram + CSV dump)
    set dumpFileId ""

    #init global variables based on Ah/BB/E5 configuration
    initGlobals $usrSetAH $usrSetBB $usrSetE5
    # Parse the RDF file
    parse_rdf
    #splash message
    profileGreetUser

    #Add external registers
    set g_extHistRegs [parseExternalInput $externalFileName $isExt]

    if {$isFile} {
        set fileId [open $logname "w"]
    } else {
        set fileId -1
    }

    if {$isDevices == 1} {
        puts "Profiler running on 2 engines!"
        dev $dev0
        set currentDev $dev0
    } else {
        set currentDev 0
    }

    if {$g_chipType == $g_arrChipTypes(E5)} {
        set stormUtilHeadLine "\nDisplaying: Storm [cStr %busy $g_stormUtilColorScheme(busy)] [cStr %stall $g_stormUtilColorScheme(stall)] [cStr %sleep $g_stormUtilColorScheme(sleep)] [cStr %inactive $g_stormUtilColorScheme(inactive)] [cStr %passiveBuffSwitch $g_stormUtilColorScheme(passiveBufferSwitch)]; BRB drops; RX pause packets; PCI header credits; PCI data credits; Iteration#"
    } else {
        set stormUtilHeadLine "\nDisplaying: Storm [cStr %busy $g_stormUtilColorScheme(busy)] [cStr %stall $g_stormUtilColorScheme(stall)] [cStr %sleep $g_stormUtilColorScheme(sleep)] [cStr %inactive $g_stormUtilColorScheme(inactive)]; BRB drops; RX pause packets; PCI header credits; PCI data credits; Iteration#"
    }

    #mode 1
    if {$isHist == 0 && $isStallCount == 0 && $isPciStats == 0} {
        puts $stormUtilHeadLine

    #modes 2/3
    } elseif {$isHist == 1} {
        initHists
    }

    #2nd dump file for mode3 if there are 2 engines:
    set dumpFileId2 0

    #Mode 3
    if {$isDump == 1} {
        set dumpFileName $histname
        if {$isDevices == 0} {
            append dumpFileName ".csv"
        } else {
            append dumpFileName "$dev0.csv"
        }
        puts "Dumping to: $dumpFileName"
        set dumpFileId [open $dumpFileName "w"]
        set regsToHist [concat $g_rxHistRegs "," $g_txHistRegs "," $g_pxpHistRegs "," $g_extHistRegs]
        set regsList [string map {\n , \t "" " " ""} $regsToHist]
        puts $dumpFileId [string map [array get g_regsText] $regsList]

        #2 engines
        if {$isDevices == 1} {
            set dumpFileName $histname
            append dumpFileName "$dev1.csv"
            puts "Dumping to: $dumpFileName"
            set dumpFileId2 [open $dumpFileName "w"]
            puts $dumpFileId2 [string map [array get g_regsText] $regsList]
        }
    }

    if {$isHist == 1} {
        puts "Sampling registers (Could take a few seconds for progress to start):"
    }

    set numLines 0
    for {set i 0} {$i < $iterations} {incr i} {
        #modes 2/3
        if {$isHist == 1} {
            hist $samples $isDump $dumpFileId $isDevices $dev0 $dev1 $dumpFileId2 $i $iterations
            flush stdout
            continue
        }
        #mode 5
        if {$isPciStats == 1} {
            pcieStats $g_iterationTime $isFile $fileId $i $keepIters
            flush stdout
            continue
        }
        #mode 4
        if {$isStallCount == 1} {
            puts ""
            # perform stall count
            if { $i == 0 } {
                puts "Analyzing stall count"
            }
            stallCount $g_iterationTime $isFile $fileId $i $displayUtil $keepIters $iterations $numLines
            # if don't need to display storm CPU utilization, continue to the next round of the loop
            if {$displayUtil == 0} {
                continue
            }
            puts $stormUtilHeadLine
            if {$isFile} {
                puts $fileId $stormUtilHeadLine
            }
        }

        #modes 1 or 4(if on)
        # display storm CPU utilization
        set numLines [stormUtilization $g_iterationTime $currentDev $isFile $fileId $isDevices $dev0 $dev1 $i $isVert]
    }
    if {[info exists numLines]} {
        puts "\r"
        for {set i 0} {$i < $numLines} {incr i} {
            puts -nonewline "\n"
        }
    } else {
        puts "\r\n\r\n"
    }

    #mode 1
    if {$isFile} {
        close $fileId
    }

    #mode 3
    if {$isDump} {
        close $dumpFileId
        if {$isDevices == 1} {
            close $dumpFileId2
        }
    }

    #modes 2 or 3
    if {$isHist == 1} {
        #inverse engine 0 registers
        inverseRegisters 0
        if {$isDevices} {
            #inverse engine 1 registers
            inverseRegisters 1
        }
        printHistsToFile $histname $isDevices $dev0 $dev1
    }
}

# Help info on "profile" command
proc profile_help {} {
    set help_info     "****************************************************\n"
    append help_info  "Mode 1 - display real time Storm CPU utilization\n"
    append help_info  "Mode 2 - sample registers and save a histogram to file\n"
    append help_info  "Mode 3 - sample registers and save both a histogram and full CSV dump to files\n"
    append help_info  "Mode 4 - display real time Storm Stall Count\n"
    append help_info  "Mode 5 - display PCIe core statistics (BB only)\n"
    append help_info  "****************************************************\n"
    append help_info  "Mode 1 usage: profile \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-d <dev0> <dev1>\] \[-ah | -bb\] \[-storms <storm letters> | -tx | -rx\]\n"
    append help_info  "Mode 2 usage: profile -h <filename> \[-forever | -i <iterations>\] \[-s <samples>\] \[-d <dev0> <dev1>\] \[-ah | -bb | -e5\] \[-ext <external filename>\] \[-dec\] \[-sbucket\]\n"
    append help_info  "Mode 3 usage: profile -v <filename> \[-forever | -i <iterations>\] \[-s <samples>\] \[-d <dev0> <dev1>\] \[-ah | -bb | -e5\] \[-ext <external filename>\] \[-dec\] \[-sbucket\]\n"
    append help_info  "Mode 4 usage: profile -sc \[-u\] \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-k\] \[-d <dev0> <dev1>\] \[-ah | -bb\] \[-storms <storm letters> | -tx | -rx\]\n"
    append help_info  "Mode 5 usage: profile -ps \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-k\] \[-d <dev0> <dev1>\]\n"
    append help_info  "****************************************************\n"
    append help_info  "Notes: \n"
    append help_info  "\t-forever runs in a loop without stopping (will never exit to user prompt)\n\n"
    append help_info  "\t-freq <frequency> update actual chips frequency for script calculations \n\n"
    append help_info  "\t-f <fileName>  writes results into file (first mode only)\n\n"
    append help_info  "\t-h <fileName>  creates histograms and writes into file\n\n"
    append help_info  "\t-s <samples> specifies number of samples per iteration (default is 1000 samples)\n\n"
    append help_info  "\t-d <dev0> <dev1> specifies the device for engine 0 and 1 for joint profiling\n\n"
    append help_info  "\t-i <iterations> specifies number of iteration to display profiling data (default is 10 iteration)\n\n"
    append help_info  "\t-u in case of stall count mode, also display real time Storm CPU utilization\n\n"
    append help_info  "\t-k keep results from all iterations in the terminal (without this flag each iteration will override the last one in the terminal)\n\n"
    append help_info  "\t-ah specifies chip type is Arrowhead\n\n"
    append help_info  "\t-bb specifies chip type is BigBear\n\n"
    append help_info  "\t-e5 specifies chip type is E5\n\n"
    append help_info  "\t-dec show histogram values in decimal instead of hexadecimal\n\n"
    append help_info  "\t-sbucket use simple bucket modes in histogram presentation\n\n"
    append help_info  "\t-ext Include additional registers from an external file to show in histogram mode\n\n"
    append help_info  "\t-r Emulation frquency ratio modifier - fastest clock speed divided by actual speed (emulation GUI clocks tab)\n\n"
    append help_info  "\t-storms Choose specific storms and display order to analyze utilization/stall-count.\n\n"
    append help_info  "\t-tx Choose Tx storms and display order to analyze utilization/stall-count.\n\n"
    append help_info  "\t-rx Choose Rx storms and display order to analyze utilization/stall-count.\n\n"
    append help_info  "\t-nl Prints storm utilization in newlines, instead of a running line.\n\n"
    append help_info  "\t-vert Prints storm utilization as a table instead of a line.\n\n"
    append help_info  "\t-nc Strip all colors.\n\n"
    append help_info  "\t-help show this text\n\n"
    append help_info  "Example: profile -h histOut.txt -i 5"
    add_help profile $help_info
}

#error proc after bad argument while parsing
proc argError {arg} {
    error "ERROR Invalid argument: $arg \nFirst mode usage: profile \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-d <dev0> <dev1>\]\nSecond mode usage: profile -h <filename> \[-forever | -i <iterations>\] \[-s <samples>\] \[-d <dev0> <dev1>\]\nThird mode usage: profile -v <filename> \[-forever | -i <iterations>\] \[-s <samples>\] \[-d <dev0> <dev1>\]\nForth mode usage: profile -sc \[-u\] \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-k\] \[-d <dev0> <dev1>\] \[-ah | -bb\]\nFifth mode usage: profile -ps \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\] \[-k\] \[-d <dev0> <dev1>\] (BB only)\n"
}

#parse arguments for running the script, according to usage.
proc parseArgs {arguments} {
    global g_profileDevModeOn
    global g_userSetFreq
    set g_userSetFreq 0
    set bucketType Weighted
    set frequency -1
    set isFile 0;
    set iterations 10
    if {$g_profileDevModeOn} {
        set iterations 1
    }
    set samples 1000
    set input_pending 0
    set doubleInput 0
    set args_counter 0
    set logname ""
    set isHist 0
    set histname ""
    set isIteration 0
    set isSample 0
    set isfrequency 0
    set isDump 0
    set isDevices 0
    set isExt 0
    set isRatio 0
    set emuClockRatio 0
    set externalFileName ""
    set dev0 0
    set dev1 1
    set usrSetAH 0
    set usrSetBB 0
    set usrSetE5 0
    set isStallCount 0
    set displayUtil 0
    set isPciStats 0
    set keepIters 0
    set decimalHistVals 0
    set rxStorms TMU
    set txStorms XYP
    set selectedStorms $rxStorms$txStorms
    set newline 0
    set isVert 0
    set noColor 0

    global g_iterationTime
    set g_iterationTime   1000 ;#The refresh time of the main loop in ms, only in case we don't calculate histogram
    if {[is_emul] && [is_e5]} {
        set g_iterationTime [expr $g_iterationTime*10];
    }

    foreach arg $arguments {
        if {$input_pending == 0} {
            switch -- $arg {
                "-freq"     {
                                if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" frequency] != 1 || $isfrequency == 1 } {argError $arg}; #checking input is integer
                                set isfrequency 1
                                set input_pending 1; #skip next argument
                                set g_userSetFreq 1; #indicate that frquency was manually set
                            }
                "-f"        {
                                if {$isFile == 1} {argError $arg}
                                set isFile 1
                                set newline 1 ;#When printing to file, we want the storm utilization to be presented in new lines
                                set noColor 1 ;#When printing to file, we don't want colors syntax (unreadable).
                                set logname [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-h"        {
                                if {$isHist == 1} {argError $arg}
                                set isHist 1
                                set histname [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-v"        {
                                if {$isHist == 1} {argError $arg}
                                set isHist 1
                                set isDump 1
                                set histname [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-s"        {
                                if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" samples] != 1 || $isSample == 1} {argError $arg}
                                set isSample 1
                                set input_pending 1
                            }
                "-i"        {
                                if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" iterations] != 1 || $isIteration == 1} {argError $arg}
                                set isIteration 1
                                set input_pending 1
                            }
                "-d"        {
                                if {$isDevices == 1} {argError $arg}
                                set isDevices 1
                                set dev0 [lindex $arguments [expr $args_counter + 1]]
                                set dev1 [lindex $arguments [expr $args_counter + 2]]
                                set input_pending 1
                                set doubleInput 1
                            }
                "-ah"       {   set usrSetAH 1 }
                "-bb"       {   set usrSetBB 1 }
                "-e5"       {   set usrSetE5 1 }
                "-forever"  {   set iterations 10000000 }
                "-sc"       {   set isStallCount 1 }
                "-ps"       {   set isPciStats 1 }
                "-k"        {   set keepIters 1 }
                "-u"        {
                                if {$isStallCount == 0} {argError $arg}
                                set displayUtil 1
                            }
                "-dec"      {
                                #set decimal hist mode to 1
                                set decimalHistVals 1
                            }
                "-sbucket"  {
                                #use simple bucket system
                                set bucketType Simple
                            }
                "-ext"      {
                                #use external register list
                                if {$isExt == 1} {argError $arg}
                                set isExt 1
                                set externalFileName [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-r"        {
                                ;#For emulations - enter custom clock ratio
                                ;#ratio is calculated by reading the clock tab in the controller emulation GUI
                                ;#Pick the fastest clock (2000_mhz usually) and divide the frequency by the real frequency.
                                if {$isRatio == 1} {argError $arg}
                                set isRatio 1
                                set emuClockRatio [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-storms"   {
                                ;#Select specific storms only
                                set selectedStorms [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-it"   {
                                ;#Select iteration time
                                set g_iterationTime [lindex $arguments [expr $args_counter + 1]]
                                set input_pending 1
                            }
                "-tx"       {
                                set selectedStorms $txStorms
                            }
                "-rx"       {
                                set selectedStorms $rxStorms
                            }
                "-nl"       {
                                set newline 1
                            }
                "-vert"     {
                                set isVert 1
                            }
                "-nc"       {
                                set noColor 1
                            }
                "-help"     {
                                help profile
                                error ""
                            }
                default     { argError $arg }
            }

        } else {
            if {$doubleInput == 0} {
                set input_pending 0
            }
            set doubleInput 0
        }

        incr args_counter
    }

    if {$input_pending == 1} {
        argError [lindex $arguments [expr $args_counter - 1]]
    }

    # only one mode of operation
    if {[expr $isHist == 1 && $isFile == 1]} {
        argError "-f and also -h | -v"
    }
    if {$isStallCount == 1 && $isHist == 1} {
        argError "-sc and also -h | -v"
    }
    if {$isStallCount == 1 && $isPciStats == 1} {
        argError "-sc and also -ps"
    }

    # can't be both AH and BB
    if {$usrSetBB == 1 && $usrSetAH == 1} {
        argError "-bb and also -ah"
    }
    if {$isPciStats == 1 && $usrSetAH == 1} {
        argError "-ps mode only works for bb"
    }

    # -k option is only available for -sc and -ps modes
    if {$keepIters == 1 && $isPciStats == 0 && $isStallCount == 0} {
        argError "-k option is only available for -sc and -ps modes"
    }

    # Remove some visual features in case we're not in diag
    if { [isDiag] == 0 } {
        set noColor 1
        set newline 1
    }

    return [list $frequency $iterations $isFile $logname $isHist $histname $samples $isDump $isDevices $dev0 $dev1 $usrSetAH $usrSetBB $usrSetE5 $isStallCount $displayUtil $isPciStats $keepIters $isExt $externalFileName $emuClockRatio $bucketType $decimalHistVals $selectedStorms $newline $isVert $noColor]
}


proc clearDict {dictName} {
    upvar 1 $dictName dictionary
    if {[info exists dictionary]} {
        foreach element [dict keys $dictionary] {
            dict unset dictionary $element
        }
    }
}

proc getActiveStorms {} {
    global g_activeStorms
    global g_storms
    set legalStorms {T M U X Y P}

    clearDict g_storms

    set activeStorms [wordToUpperChars $g_activeStorms]
    foreach storm  $activeStorms {
        if {[isInList $storm $legalStorms]} {
            devMsg "Adding $storm"
            dict set g_storms $storm "${storm}SEM"
        } else {
            puts "Illegal storm requested:'$storm'. Ignored."
        }
    }
}

proc initGlobals {usrSetAH usrSetBB usrSetE5} {
    global g_offsetRegsArgs
    global g_regsToCount
    global g_regsText
    global g_regsFocCredit
    global g_inverseRegisters
    global g_storms
    global g_pxpHistRegs
    global g_rxHistRegs
    global g_txHistRegs
    global g_extHistRegs
    global g_pciRegs
    global g_pciSamples
    global g_pciSamplesInterval
    global g_numberOfHistBuckets
    global g_regsGroups
    global g_frequency
    global g_stallCountNames
    global g_pcieStatsRegs
    global g_sem_regs
    global g_prevLineLength
    global g_progressPrev
    global g_chipModeName
    global g_coreAString
    global g_coreBString
    global g_noCoreString
    global rdfFile

    global g_chipType
    global g_arrChipTypes
    global g_arrChipTypesReverse

    global g_chipMode
    global g_chipModeAsic
    global g_chipModeEmul
    global g_chipModeFpga

    global g_profileDevModeOn

    set g_progressPrev   0
    set g_prevLineLength 0

    set g_pciSamples        10 ;#PCI registers' samples per g_iterationTime
    set g_pciSamplesInterval 1 ;#PCI registers' samples interval in ms

    set g_numberOfHistBuckets 10
    #determine chip type
    get_chip_type
    #determine chip mode
    get_chip_mode

    # check if user specified chip type - override $g_chipType if required
    if {$usrSetAH == 1 || $usrSetBB == 1 || $usrSetE5 == 1} {
        # set chip type according to user (Only one of the user set chiptypes can be 1)
        set g_chipType [expr ($usrSetE5 * $g_arrChipTypes(E5)) + ($usrSetAH * $g_arrChipTypes(AH)) + ($usrSetBB * $g_arrChipTypes(BB)])
    }

    #if frequency wasn't defined manually - read chip frequency
    while {![info exists g_frequency] || $g_frequency == -1} {
        set g_frequency [getChipFrequency]
    }

    if {$g_profileDevModeOn} {


    }

    getActiveStorms

    set g_coreAString [cStr "core A  " green]
    set g_coreBString [cStr "core B  " blue]
    set g_noCoreString "\t    "

    clearDict g_sem_regs

    if {$g_chipType == $g_arrChipTypes(E5)} {

        dict set g_sem_regs $g_coreAString {_REG_FAST_MEMORY_STORM_ACTIVE_CYCLES_A _REG_FAST_MEMORY_STORM_STALL_CYCLES_A _REG_FAST_MEMORY_IDLE_SLEEPING_CYCLES_A _REG_FAST_MEMORY_IDLE_INACTIVE_CYCLES_A}
        dict set g_sem_regs $g_coreBString {_REG_FAST_MEMORY_STORM_ACTIVE_CYCLES_B _REG_FAST_MEMORY_STORM_STALL_CYCLES_B _REG_FAST_MEMORY_IDLE_SLEEPING_CYCLES_B _REG_FAST_MEMORY_IDLE_INACTIVE_CYCLES_B}
    } else {
        dict set g_sem_regs $g_noCoreString {_REG_FAST_MEMORY_STORM_ACTIVE_CYCLES _REG_FAST_MEMORY_STORM_STALL_CYCLES _REG_FAST_MEMORY_IDLE_SLEEPING_CYCLES _REG_FAST_MEMORY_IDLE_INACTIVE_CYCLES}
    }

    initRegsForProccessing

    global g_stormUtilColorScheme
    array unset g_stormUtilColorScheme *

    array set g_stormUtilColorScheme [sl {busy                green;
                                          stall               red;
                                          sleep               cyan;
                                          inactive            yellow;
                                          passiveBufferSwitch 94; #bright blue
    }]
}

proc initDumpFile {dumpFileName {isDevices 0} {dev 0}} {
    global g_regsText
    global g_rxHistRegs
    global g_txHistRegs
    global g_pxpHistRegs

    set regsToHist [concat $g_rxHistRegs "," $g_txHistRegs "," $g_pxpHistRegs "," $g_extHistRegs]
    set regsList [string map {\n , \t "" " " ""} $regsToHist]

    if {$isDevices} {
        append dumpFileName "$dev.csv"
    } else {
        append dumpFileName ".csv"
    }

    puts "Dumping to: $dumpFileName"
    set dumpFileId [open $dumpFileName "w"]
    puts $dumpFileId [string map [array get g_regsText] $regsList]
}

proc prepareLineFormat {busy stall sleep inactive passiveBuffSwitch} {
    global g_profileDevModeOn
    global g_chipType
    global g_arrChipTypes

    #set start floating point accuracy for displaying results
    #startAccuracy indiactes how many digits after the decimal point are presented by default
    #maxAccuracy indicates how many digits after the decimal point can be presented if the number is too small
    if {$g_profileDevModeOn} {
        set startAccuracy 0
        set maxAccuracy   0
    } else {
        set startAccuracy 0
        set maxAccuracy   2
    }
    set accuracyThreshold 1
    set formatList {}
    set defaultFrmt "%3.${startAccuracy}f"
    #set custom format for each field.
    set registerList {}
    if {$g_chipType == $g_arrChipTypes(E5)} {
        set registerList [list $busy $stall $sleep $inactive $passiveBuffSwitch]
    } else {
        set registerList [list $busy $stall $sleep $inactive]
    }
    foreach current $registerList {
        #if the current field is smaller than $accuracyThreshold but not zero, present it more accurately
        if {double($current) < $accuracyThreshold && double($current) != 0} {
            set j $startAccuracy
            set extraFrmt "%3.${j}f"
            #keep increasing format accuracy untill the presented field is no longer zero
            while { [format $extraFrmt $current] == 0 && $j < $maxAccuracy} {
                incr j
                set extraFrmt "%3.${j}f"
            }
            #if accuracy increase did not help - set back to minimal accuracy (to prevent showing 0.00...)
            if {[format $extraFrmt $current] == 0} {
                set extraFrmt $defaultFrmt
            }
            lappend formatList $extraFrmt
        } else  {
            lappend formatList $defaultFrmt
        }
    }
    return $formatList
}

proc getLineSpacesString {prevLineLenght newLineLength} {
    set diff [expr $prevLineLenght - $newLineLength]
    if {$diff <= 0} {
        return ""
    }
    return [string repeat " " $diff]
}

proc getCoreTableOrLine {formatList line storm busy stall sleep inactive passiveBuffSwitch isVert numLines} {
    global g_chipType
    global g_arrChipTypes
    global g_stormUtilColorScheme

    upvar 1 $numLines numLinesUpvared
    set extraSpace ""
    if {$isVert} {
        set extraSpace "          "
    }
    if {$g_chipType == $g_arrChipTypes(E5)} {
        set line [format "%s$extraSpace%s [cStr [lindex $formatList 0] $g_stormUtilColorScheme(busy)] [cStr [lindex $formatList 1] $g_stormUtilColorScheme(stall)] [cStr [lindex $formatList 2] $g_stormUtilColorScheme(sleep)] [cStr [lindex $formatList 3] $g_stormUtilColorScheme(inactive)] [cStr [lindex $formatList 4] $g_stormUtilColorScheme(passiveBufferSwitch)]| "\
                    $line $storm $busy                $stall              $sleep              $inactive               $passiveBuffSwitch]
    } else {
        set line [format "%s$extraSpace%s [cStr [lindex $formatList 0] $g_stormUtilColorScheme(busy)] [cStr [lindex $formatList 1] $g_stormUtilColorScheme(stall)] [cStr [lindex $formatList 2] $g_stormUtilColorScheme(sleep)] [cStr [lindex $formatList 3] $g_stormUtilColorScheme(inactive)]| "\
                    $line $storm $busy                   $stall                  $sleep                  $inactive]
    }

    if {$isVert} {
        set line "$line\n"
        incr numLinesUpvared
    }

    return $line
}

proc moveUpLines {numLines} {
        puts -nonewline "\033\[${numLines}A"
}

# display storm CPU utilization
proc stormUtilization {iterationTime currentDev isFile fileId isDevices dev0 dev1 i {isVert 0}} {
    global g_pciRegs
    global g_frequency
    global g_regsText
    global g_profileDevModeOn
    global g_prevLineLength
    global g_chipType
    global g_arrChipTypes
    global g_newline
    global g_stormUtilColorScheme

    readCpuData; #to clear the registers
    set timeDelta [time {
        after $iterationTime; # measure cycles in $iterationTime milliseconds
        set cpu [readCpuData]
    }]
    set timeDelta [lindex $timeDelta 0]; # time return a sentence: "<usec> microseconds per iteration"
    devMsg "\ntimeDelta is [expr double($timeDelta)/1000000.0] (sec); iteration time is [expr double($iterationTime)/1000] (sec); Diff is [expr double($timeDelta)/1000.0 - double($iterationTime)] (ms)"


    set linePrefix ""
    if {$isDevices == 0} {
        #Single engine
        set linePrefix ""
    } else {
        #2 engines
        set linePrefix "Device $currentDev: "
    }
    # calculate busy data
    set num_lines 0
    dict for {prefix cpu_data} $cpu {
        if {$isVert} {
            set line "$linePrefix$prefix\n"
        } else {
            set line "$linePrefix$prefix"
        }
        set num_lines [expr $num_lines + 1 + $isVert]
        if {![info exists g_frequency] || $g_frequency == ""} {
            puts "Freq is $g_frequency"
        }
        set totalCyclesCalculated [expr ( double($timeDelta) * double($g_frequency) )  /1000000.0 ]
        foreach {storm data} $cpu_data {
            set analyzedResults     [analyzeCpuUtil $data $timeDelta $g_frequency $storm]
            set busy                [lindex $analyzedResults 0]
            set stall               [lindex $analyzedResults 1]
            set sleep               [lindex $analyzedResults 2]
            set inactive            [lindex $analyzedResults 3]
            set passiveBuffSwitch   [lindex $analyzedResults 4]
            set formatList [prepareLineFormat $busy $stall $sleep $inactive $passiveBuffSwitch]
            set line [getCoreTableOrLine $formatList $line $storm $busy $stall $sleep $inactive $passiveBuffSwitch $isVert num_lines]
        }
        # Print Storm utilization line
        set newLineLength [string length $line]
        set spaces [getLineSpacesString $g_prevLineLength $newLineLength]
        puts -nonewline "\r$line $spaces\n"
        if {$isFile} {
            puts $fileId "$line $i\n"
        }
        set g_prevLineLength $newLineLength
    }
    set line "\t    "

    #Read registers' differences from initial value
    set countedRegs [countRegs]

    set line [format "%s %s" $line $countedRegs]

    foreach reg $g_pciRegs {
        set line  [format "%s %s %s;" $line $g_regsText($reg) [format 0x%x [getAveOfReg $reg]]]
    }
    # Print the regs line
    puts -nonewline "\r$line $i"

    if {$isFile} {
        puts $fileId "$line $i\n"
    }
    # Go back with the cursor to overwrite lines
    if {!$g_newline} {moveUpLines $num_lines};

    flush stdout

    if {$isDevices == 1} {
        if {$currentDev == $dev0} {
            dev $dev1
            set currentDev $dev1
            puts -nonewline "\n"; # after printing the first device set cursor at the next line (to print the second device)
        } else {
            dev $dev0
            set currentDev $dev0
            puts -nonewline "\x1b\[F"; # after printing the second device set cursor at the previous line (to overprint the first device)
        }
    }
    return $num_lines
}

#Return a list of storm and its registers values
proc readCpuData {} {
    global g_storms
    global g_sem_regs
    #iterate over cores
    dict for {prefix regs} $g_sem_regs {
        set result {}
        #iterate over each storm per-core (if relevant)
        dict for {storm sem} $g_storms {
            lappend result $storm
            #iterate over each register per-storm (per-core if relevant)
            foreach {reg_name} $regs {
                lappend data [getRegVal $sem$reg_name]
            }
            lappend result $data
            unset data
        }
        dict set res $prefix $result
    }
    return $res
}

#Return percentage of argument 'cycles' from all cycles in argument 'millisec' interval
proc calcUtilization  {cycles usec frequency} {
    set val [expr double($cycles) / double($usec) ]
    set val [expr 100.0 * double($val) ]
    set val [expr double($val) / double($frequency) ]
    set val [expr 1000000.0 * double($val)]

    return $val
}

proc analyzeCpuUtil {data timeDelta frequency storm} {
    global g_chipType
    global g_chipMode
    global g_chipModeAsic
    global g_arrChipTypes
    global g_iterationTime

    #Theoretical 100%
    set totalCyclesCalculated [expr ( double($timeDelta) * double($frequency) )  /1000000.0 ]
    set activeCycles   [lindex $data 0]
    set stallCycles    [lindex $data 1]
    set sleepCycles    [lindex $data 2]
    set inactiveCycles [lindex $data 3]

    #stall cycles are part of the active cycles, so no need to count them as towards the 100%
    set totalCyclesMeasured   [expr $activeCycles + $sleepCycles  + $inactiveCycles ]

    set result {}
    set totalCycles $totalCyclesCalculated
    #if E4 - use measurement sum as 100% reference
    if {$g_chipType == $g_arrChipTypes(AH) || $g_chipType == $g_arrChipTypes(BB)} {
        set totalCycles $totalCyclesMeasured
    }

    lappend result [expr 100.0 * $activeCycles   / $totalCycles]
    lappend result [expr 100.0 * $stallCycles    / $totalCycles]
    lappend result [expr 100.0 * $sleepCycles    / $totalCycles]
    lappend result [expr 100.0 * $inactiveCycles / $totalCycles]

    #if E5 - calculate cycles 'lost' to DRA
    if {$g_chipType == $g_arrChipTypes(E5) } {
        set passiveBuffSwitchCycles [expr $totalCyclesCalculated - $totalCyclesMeasured]
        set passiveBuffSwitchPercent [expr 100.0 * double($passiveBuffSwitchCycles) / double($totalCycles)]
        if {$passiveBuffSwitchPercent < 1 && $passiveBuffSwitchPercent >= 0} {
            #needs to be at least 1% to make sure it's not just due to inacruacy of measurement
            lappend result 0
        } elseif {$passiveBuffSwitchPercent >= 1} {
            lappend result $passiveBuffSwitchPercent
        } else {
            #negative number
            lappend result $passiveBuffSwitchPercent
            puts "\n\033\[31mERROR \033\[92m\033\[1m${storm}STORM\033\[0m:totalCyclesCalculated - totalCyclesMeasured = $totalCyclesCalculated - $totalCyclesMeasured = passiveBuffSwitchCycles=$passiveBuffSwitchCycles;"
        }
        lappend result $passiveBuffSwitchCycles ;#cycles
    }
    return $result
}

proc getAsicFreq {{clockType STORM}} {
    global g_chipMode
    global g_chipModeAsic

    if {![info exists g_chipMode]} {
        get_chip_mode
    }
    if {$g_chipMode != $g_chipModeAsic} {
        error "Not ASIC, can't read clock frequency"
    }

    set freq [expr [getAsicFrequency $clockType]/1000000]
    return "$clockType clock: $freq MHz"
}

proc setConstFreq {freq} {
    global g_constantFreq
    set g_constantFreq $freq
}

proc unsetConstFreq {} {
    global g_constantFreq
    unset g_constantFreq
}

proc getChipFrequency {} {
    global g_chipType
    global g_arrChipTypes
    global g_chipMode
    global g_chipModeEmul
    global g_emuClockRatio
    global g_userSetFreq
    global g_constantFreq

    if { [info exists g_constantFreq] && $g_constantFreq > 0} {
        set g_frequency $g_constantFreq
        return $g_frequency
    }

    if { ![info exists g_chipType]} {
        get_chip_type
    }

    if { ![info exists g_chipMode]} {
        get_chip_mode
    }

    set chipID [format %d [reg_read -n MISCS_REG_CHIP_NUM]]

    array set chipsFrequency {5673  1000000000
                              5674  1000000000
                              5675  1000000000
                              5676  1000000000
                              5677  1000000000
                              5678  1000000000
                              5679  1000000000
                              5680  1000000000
                              5681  1000000000
                              5682  1000000000
                              5683  1000000000
                              5684  1000000000
                              5685  1000000000
                              32880 1200000000
                              33008 1200000000};

    if {$g_chipType == $g_arrChipTypes(E5)} {
        set g_frequency 1700000000
    } elseif {[info exists chipsFrequency($chipID)]} {
        set g_frequency $chipsFrequency($chipID)
    } else {
        error "Could not determine storm frequency. To enter manual frequency use '-freq <frequency in Hz>' flag"
    }

    if {$g_chipMode == $g_chipModeEmul} {
        if { $g_emuClockRatio != 0} {
            set g_frequency [expr $g_frequency/$g_emuClockRatio]
        } elseif { [info exists g_userSetFreq] && $g_userSetFreq == 0} {
            error "ERROR: Emulation mode detected. User must provide custom frequency with '-freq' flag, and/or custom clock ratio with '-r' flag (run 'profile -help' for more details)"
        }

    }

    return $g_frequency
}

#Read chip clock frequency (default is STORM clock, c)
proc getAsicFrequency {{clockType STORM} {secondRun 0}} {
    global g_chipType
    global g_arrChipTypes

    if { ![info exists g_chipType]} {
        get_chip_type
    }

    array set legalClocks [sl {
        $g_arrChipTypes(AH) {STORM MAIN NWM}                          ;#E4 ArrowHead
        $g_arrChipTypes(BB) {STORM MAIN NW}                           ;#E4 BigBear
        $g_arrChipTypes(E5) {STORM MAIN MISC SVC SCLK_PEM PCE 25 125} ;#E5
    }]

    if { ![info exists legalClocks($g_chipType)] } {
        set legalClocksList {STORM MAIN}
    } else {
        set legalClocksList $legalClocks($g_chipType)
    }

    if {![isInList $clockType $legalClocksList]} {
        puts "Illegal clock name '$clockType'. Legal values are: [join $legalClocksList ","]"
        return -1
    }
    set freqCaptureReg IPC_REG_FREQ_CAPTURE
    #$clockType can be one of the following:
    #STORM,MAIN
    #E4-AH can also be:
    #NWM
    #E4-BB can also be:
    #NW
    #E5 can also be:
    #MISC,SVC,125,SCLK_PEM,25,PCE
    set freqValReg IPC_REG_FREQ_$clockType

    set validBitMask 0x00010000
    set valueBitMask 0x0000FFFF
    #read storm frequency register
    set freqVal [getRegVal $freqValReg]
    set preVal 0
    #if valid bit is on, save the frequency to subtract later on
    if {([expr $freqVal & $validBitMask]) && ($g_chipType != $g_arrChipTypes(E5))} {
        set preVal [expr $freqVal & $valueBitMask]
    }

    #capture
    reg_write -n $freqCaptureReg  -v 1

    set freqVal [getRegVal $freqValReg]
    set pollLimit 10
    set i 0
    #if freq still isn't valid, keep polling
    while {([expr $freqVal & $validBitMask] == 0) && ($i < $pollLimit)} {
        after 100
        incr i
        set freqVal [getRegVal $freqValReg]
    }

    #if valid bit is still invalid after polling, something's wrong - error.
    if {[expr $freqVal & $validBitMask] == 0 } {
        error "Failed polling chip frequency from register $freqValReg"
    }

    #subtract measured value from previous value
    set freqVal [expr ($freqVal & $valueBitMask) - $preVal]

    if {($freqVal < 0 || $freqVal eq "") && !$secondRun} {
        #sometime there's a wrap around, which requires an additional call  to the function
        return [getAsicFrequency $clockType 1]
    } else {
        return [expr $freqVal*100000]
    }
}

#Get a register's value, while considering its offset and bitmask as set in cfg file. Different flow for address regs and name regs.
proc getRegVal {reg} {
    global g_offsetRegsArgs
    global g_regsGroups

    set reg [stripFormatFromRegName $reg]

    if { ![info exists g_offsetRegsArgs($reg)]} {
        set bitMask 0xFFFFFFFF
        set offset 0
    } else {
        set bitMask [lindex $g_offsetRegsArgs($reg) 0]
        set offset [lindex $g_offsetRegsArgs($reg) 1]

        # if the register is split into fields, we first set the mask according to both the
        # register and the field, and then we get the value of the register (without the field)
        set reg [getRegNoField $reg]
    }
    if {[info exists g_regsGroups($reg)]} {
        set val 0
        foreach register $g_regsGroups($reg) {
            incr val [getRegVal $register]
        }
    } else {
        if {[string first "REG_ADDR" $reg] != -1} {
            set addr [lindex [split $reg "_"] 2]
            set val [expr [expr ([reg read $addr] & $bitMask)] >> $offset]
        } else {
            set splReg [split $reg "+"]
            # check if the register is of type <regName>[<index>]
            if {[llength $splReg] > 1} {
                # get the address range of the array
                set range [reg_offset -n [lindex $splReg 0]]
                set splRange [split $range "-"]

                # get the index
                set index [lindex $splReg 1]
                # check if index is out of range
                if {[expr ([lindex $splRange 1] - [lindex $splRange 0])/4] < $index} {
                    error [format "Index out of range %s: register's byte adress range is %s" $reg $range]
                }

                # get the address using the register's neame and the index
                set addr [expr [lindex $splRange 0] + (4 * $index)]
                # get the value of the register in that address
                set val [expr [expr ([reg read $addr] & $bitMask)] >> $offset]
            } else {
                # check if need to sum value of lo and hi
                if {[string first "LoAndHi#" $reg] != -1} {
                    # split reg to lo and hi
                    set regName [lindex [split $reg "#"] 1]
                    set valLo [getRegVal [format "%s%s" $regName "_LO"]]
                    set valHi [getRegVal [format "%s%s" $regName "_HI"]]

                    # get reg value
                    set val [expr ($valLo)+(($valHi)<<32)]
                } else {
                    set addr [ get_start_offset [reg_offset -n $reg] ]
                    set val [expr [expr ([reg read $addr] & $bitMask)] >> $offset]
                }
            }
        }
    }
    return $val
}

#Return a string for printing the registers we want to count
proc countRegs {} {
    global g_regsText
    global g_regsToCount

    set toPrint ""
    foreach reg $g_regsToCount {
        set reg [stripFormatFromRegName $reg]
        set regVal [getRegVal $reg];
         if { ![info exists g_regsText($reg)]} {
            set g_regsText($reg) "Register $reg has incremented by"
         }
        append toPrint "$g_regsText($reg) [format "%u" $regVal];"
    }

    return $toPrint
}

#Calculate a register's average.
proc getAveOfReg {reg} {
    global g_pciSamples
    global g_pciSamplesInterval

    set reg [stripFormatFromRegName $reg]

    set sum 0
    for {set i 0} {$i < $g_pciSamples} {incr i} {
        incr sum [getRegVal $reg]
        after $g_pciSamplesInterval
    }

    return [expr $sum / $g_pciSamples]
}

#Initialization of histogram data structure
proc initHists {} {
    global g_rxHistRegs
    global g_txHistRegs
    global g_pxpHistRegs
    global g_extHistRegs
    global g_histsInMem
    global g_histsInMem2
    array unset g_histsInMem *
    array unset g_histsInMem2 *

    set regsToHist [concat $g_rxHistRegs $g_txHistRegs $g_pxpHistRegs $g_extHistRegs]
    foreach reg $regsToHist {
            set g_histsInMem($reg) {}
            set g_histsInMem2($reg) {}
    }
}

#add a value to the register's histogram.
proc addValToHist {val regKey isDev1} {
    global g_histsInMem
    global g_histsInMem2

    array unset regHist *
    if {$isDev1 == 0} {
        array set regHist $g_histsInMem($regKey)
    } else {
        array set regHist $g_histsInMem2($regKey)
    }
    set keysList [array names regHist]; #list of the values that has been sampled
    set index [lsearch $keysList $val]
    if {$index == -1} {
        if {$isDev1 == 0} {
            set g_histsInMem($regKey) [concat $g_histsInMem($regKey) [list $val 1]]
        } else {
            set g_histsInMem2($regKey) [concat $g_histsInMem2($regKey) [list $val 1]]
        }
    } else {
        set index [expr $index * 2]; #the index in the histogram hist

        #increment g_histsInMem($regKey)[index+1] {lreplace list first last ?element element ...?}
        if {$isDev1 == 0} {
            set g_histsInMem($regKey) [lreplace $g_histsInMem($regKey) [expr $index + 1] [expr $index + 1]  [expr [lindex $g_histsInMem($regKey) [expr $index + 1]] + 1]]
        } else {
            set g_histsInMem2($regKey) [lreplace $g_histsInMem2($regKey) [expr $index + 1] [expr $index + 1]  [expr [lindex $g_histsInMem2($regKey) [expr $index + 1]] + 1]]
        }
    }
}

proc printProgressLineStallCount {i reasons j iterations} {
    printProgressLine $j $iterations $i $reasons  Iterations Reasons
}

proc printProgressLine {i iterations {j 0} {samples 1} {iterationsHeader "Iterations"} {samplesHeader "Samples"} } {
    global g_progressPrev
    if {![info exists g_progressPrev] || $g_progressPrev == 100} {
        set g_progressPrev 0
    }
    set progressBarSize 20
    set posProgChar "#"
    set negProgChar "-"
    incr j
    incr i
    set progressPercent [expr 100*(($i-1)*$samples + $j)/($iterations*$samples)]
    if {$progressPercent != $g_progressPrev} {
        set progressInt [expr $progressPercent/(100/$progressBarSize)]
        set progressBar [format "\[%s%s\]" [string repeat $posProgChar $progressInt] [string repeat $negProgChar [expr $progressBarSize-$progressInt] ] ]
        if {$samples == 1} {
            set printString [format "%s:%d/%d; Overall progress:%s(%3.0f%s)      " $iterationsHeader $i $iterations $progressBar $progressPercent "%"]
        } elseif {$iterations ==1} {
            set printString [format "%s:%d/%d; Overall progress:%s(%3.0f%s)      " $samplesHeader $j $samples $progressBar $progressPercent "%"]
        } else {
            set printString [format "%s:%d/%d; %s:%d/%d; Overall progress:%s(%3.0f%s)      " $iterationsHeader $i $iterations $samplesHeader $j $samples $progressBar $progressPercent "%"]
        }
        if { [isDiag] == 1 } {
            puts -nonewline "\r$printString"
        } else {
            puts "$printString"
        }
        incr g_progressPrev
        flush stdout
    }

}

#Root histogram function
proc hist {Samples Dump dumpFileId isDevices dev0 dev1 dumpFileId2 {currentIteration 0} {iterations 1}} {
    global g_rxHistRegs
    global g_txHistRegs
    global g_pxpHistRegs
    global g_extHistRegs
    global g_chipType
    global g_arrChipTypes

    set defaultInvalidVal 0x19999999

    if { $g_chipType == $g_arrChipTypes(E5)} {
        #Enable report of current value for XSEM_REG_DBG_QUEUE_* registers, instead of max value
        reg_write -n XSEM_REG_DBG_QUEUE_PEFORMANCE_MON_STAT  -v 1
    }

    set regsToHist [concat $g_rxHistRegs $g_txHistRegs $g_pxpHistRegs $g_extHistRegs]
    for {set i 0} {$i < $Samples} {incr i} {
        printProgressLine $currentIteration $iterations $i $Samples
        foreach reg $regsToHist {
            set val [getRegVal $reg]
            if {$val < $defaultInvalidVal} {
                addValToHist $val $reg 0
            } else {
                #read has failed on emulation. wait 100ms.
                after 100
            }
            if {$Dump == 1} {
               puts -nonewline $dumpFileId "$val,"
            }
            if {$isDevices == 1} {
                dev $dev1
                set val [getRegVal $reg]
                if {$val < $defaultInvalidVal} {
                    addValToHist $val $reg 1
                }
                dev $dev0
                if {$Dump == 1} {
                    puts -nonewline $dumpFileId2 "$val,"
                }
            }
        }
        if {$Dump == 1} {
            puts $dumpFileId ""
            if {$isDevices == 1} {
                puts $dumpFileId2 ""
            }
        }
    after 10
    }
}


#Go over each of the registers and print the histogram from the run. Divided to two sections - Rx and Tx registers.
proc printHistsToFile {outFile isDevices dev0 dev1} {
    global g_txHistRegs
    global g_rxHistRegs
    global g_pxpHistRegs
    global g_extHistRegs

    set isExt [expr {[llength $g_extHistRegs] != 0}]
    set fileId [open $outFile "w"]

    #print for engine 0
    printHistPerDevice $fileId $isDevices $dev0 0

    #print for engine 1
    if {$isDevices == 1} {
        printHistPerDevice $fileId $isDevices $dev1 1
    }
    close $fileId
}


proc printHistPerDevice {fileId isDevices dev isDev1} {
    global g_txHistRegs
    global g_rxHistRegs
    global g_pxpHistRegs
    global g_extHistRegs
    global g_chipType
    global g_arrChipTypes

    if {$isDevices == 1} {
        puts $fileId "\n\n--------------------Device $dev-----------------"
    }

    set isExt [expr {[llength $g_extHistRegs] != 0}]

    if {$g_chipType == $g_arrChipTypes(E5)} {
        puts $fileId "\n\n----------------------PCIe data-------------------"
    } else {
        puts $fileId "\n\n----------------------PXP data-------------------"
    }
    foreach reg $g_pxpHistRegs {
        printSingleHist $fileId $reg $isDev1
    }

    puts $fileId "\n\n----------------------RX data--------------------"
    foreach reg $g_rxHistRegs {
        printSingleHist $fileId $reg $isDev1
    }

    puts $fileId "\n\n----------------------TX data--------------------"
    foreach reg $g_txHistRegs {
        printSingleHist $fileId $reg $isDev1
    }

    if {$isExt} {
        puts $fileId "\n\n-----------------External(custom) data-----------"
        foreach reg $g_extHistRegs {
            printSingleHist $fileId $reg $isDev1
        }
    }

}

proc stripFormatFromRegName {reg} {
    regsub -all {@\S+@} $reg "" regNoFormat
    set regNoFormat
}


#prints to file a single register's histogram.
proc printSingleHist {fileId reg isDev1} {
    global g_histsInMem
    global g_histsInMem2
    global g_numberOfHistBuckets
    global g_regsText
    global g_decimalHistVals
    global g_regsFocCredit

    array unset histArray *
    #fetch relevant array or skip register
    if {$isDev1 == 0} {
        if {![info exists g_histsInMem($reg)]} {
            #skip registers with no data
            return
        }
        array set histArray $g_histsInMem($reg)
    } else {
        if {![info exists g_histsInMem2($reg)]} {
            #skip registers with no data
            return
        }
        array set histArray $g_histsInMem2($reg)
    }

    #Set format
    set hexOrDecFormat [getHistValueFormat $reg]
    #extract register name without format
    set regNoFormat  [stripFormatFromRegName $reg]

    #print title of current register
    set line ""
    if { ![info exists g_regsText($reg)]} {
        set line "Histogram on the values of register: $regNoFormat"
    } else {
        set line "Histogram on the values of register: $g_regsText($reg)"
    }

    set resetVal [reset_value -n [getRegNoField $regNoFormat]]
    if {![isEmptyString $resetVal]} {
       set line "$line (Reset value $resetVal)"
    }
    puts $fileId $line

    set orderedValues [lsort -integer [array names histArray]]

    if {[llength $orderedValues] > $g_numberOfHistBuckets} {
        if {$isDev1 == 0} {
            getBuckets $orderedValues $fileId $g_histsInMem($reg) $reg
        } else {
            getBuckets $orderedValues $fileId $g_histsInMem2($reg) $reg
        }
    } else {
        foreach value $orderedValues {
            puts $fileId "Number of samples of value [format $hexOrDecFormat $value]: $histArray($value)"
        }
    }
    focCreditThreshold $fileId $reg $isDev1
    puts $fileId "==============================================================================="

}

proc getHistValueFormat {reg} {
    global g_decimalHistVals
    #Set default format
    set hexOrDecFormat "0x%X"
    if {$g_decimalHistVals} {
        set hexOrDecFormat "%d"
    }

    #extract format from register name
    regsub -all {(@\S+@)*(.*)} $reg \\1 customFormat
    regsub -all {@(\S+)@} $customFormat \\1 customFormat

    #Set per-register format if available
    if {[string length $customFormat] != 0} {
        set hexOrDecFormat $customFormat
    }

    return $hexOrDecFormat
}

#Show outstanding instead of remaining
proc inverseRegisters {isDev1} {
    global g_inverseRegisters
    global g_regsText
    global g_histsInMem
    global g_histsInMem2

    foreach reg [array names g_inverseRegisters] {
        array unset currentArray *
        array unset diffArray *
        array unset defaultValHistArray *
        if {$isDev1 == 0} {
            if {![info exists g_histsInMem($reg)] || ![info exists g_histsInMem($g_inverseRegisters($reg))]} {
                continue
            }
            array set defaultValHistArray $g_histsInMem($g_inverseRegisters($reg))
            array set currentArray $g_histsInMem($reg)
            array unset g_histsInMem $reg
        } else {
            if {![info exists g_histsInMem2($reg)] || ![info exists g_histsInMem2($g_inverseRegisters($reg))]} {
                continue
            }
            array set defaultValHistArray $g_histsInMem2($g_inverseRegisters($reg))
            array set currentArray $g_histsInMem2($reg)
            array unset g_histsInMem2 $reg
        }
        set startVal [array names defaultValHistArray]
        if {![info exists startVal]} {
            if {$isDev1 == 0} {
                set g_histsInMem($reg) $currentArray
            } else {
                set g_histsInMem2($reg) $currentArray
            }
            continue
        }

        #add default val to histogram title
        if {[info exists g_regsText($reg)]} {
            set valFormat [getHistValueFormat $reg]
            regsub -all {#MAXVALSTRING#} $g_regsText($reg) [format $valFormat $startVal] newTitle
            set g_regsText($reg) $newTitle
        }

        foreach value [array names currentArray] {
            set newVal [expr $startVal - $value]
            set diffArray($newVal) $currentArray($value)
        }
        if {$isDev1 == 0} {
            set g_histsInMem($reg) [array get diffArray]
        } else {
            set g_histsInMem2($reg) [array get diffArray]
        }
    }
    removeCfgRegisters $isDev1
}

#remove configuration registers from histogram
proc removeCfgRegisters {isDev1} {
    global g_inverseRegisters
    global g_histsInMem
    global g_histsInMem2

    if {!$isDev1} {
        set suffix ""
    } else {
        set suffix "2"
    }

    foreach reg [array names g_inverseRegisters] {
        set reg $g_inverseRegisters($reg)
        if {[info exists g_histsInMem${suffix}($reg)]} {
            array unset g_histsInMem${suffix} $reg
        }
    }
}

proc focCreditThreshold {fileId reg isDev1} {
    global g_histsInMem
    global g_histsInMem2
    global g_regsFocCredit
    global g_decimalHistVals
    global g_regsText
    global g_chipType
    global g_arrChipTypes

    set reg [stripFormatFromRegName $reg]

    #return if not E5 or if the register is not a credit register
    if {$g_chipType != $g_arrChipTypes(E5) || ![info exists g_regsFocCredit($reg)]} {
        return
    }

    set minFocCreditReg ""
    #get the name of the appropriate 'threshold' register based on the current foc_credit register
    regsub -all {(.*)\((\d+)b\)} $g_regsFocCredit($reg) \\2 bitWidth
    if {$bitWidth eq $g_regsFocCredit($reg)} {
        set bitWidth ""
    }
    regsub -all {SEM_REG_FOC_CREDIT.*} $reg "SEM_REG_FOC_MIN_MESSAGE_CREDIT" minFocCreditReg

    #List of legal foc_min_msg_credit register names
    set focMinMsgCreditRegs_E5 [sl {
        XSEM_REG_FOC_MIN_MESSAGE_CREDIT
        YSEM_REG_FOC_MIN_MESSAGE_CREDIT
        PSEM_REG_FOC_MIN_MESSAGE_CREDIT
        TSEM_REG_FOC_MIN_MESSAGE_CREDIT
        MSEM_REG_FOC_MIN_MESSAGE_CREDIT
        USEM_REG_FOC_MIN_MESSAGE_CREDIT
    }]

    #make sure register name is valid
    if {![isInList $minFocCreditReg $focMinMsgCreditRegs_E5]} {
        puts "WARNING: '$minFocCreditReg' is not defined."
        return
    }

    #get threshold value from register
    set minFocCreditValMask 0xFF
    set minValue [expr [getRegVal $minFocCreditReg] & $minFocCreditValMask]

    #make sure a valid value was read
    if {![info exists minValue]} {
        return
    }

    array unset histArray *
    if {$isDev1 == 0} {
        array set histArray $g_histsInMem($reg)
    } else {
        array set histArray $g_histsInMem2($reg)
    }
    set orderedValues [lsort -integer [array names histArray]]

    #calculate how many samples are below (or equal to) threshold and how many are above
    set below 0
    set above 0
    #set FOC_CREDIT multiplier. 256b wide FOC interface should be counted twice.
    if {$bitWidth ne ""} {
        set mult [expr $bitWidth/128]
    } else {
        #deault is 128bit (unless specified otherwise)
        set mult 1
    }
    #count how many below and how many above
    foreach value $orderedValues {
        if {$value <= $minValue} {
            set below [expr $below + $mult*$histArray($value)]
        } else {
            set above [expr $above + $mult*$histArray($value)]
        }
    }

    set hexOrDecFormat [getHistValueFormat $reg]

    set total [expr $below+$above]
    if {$total == 0} {
        #handle case of 0 samples
        set total 1
    }
    set belowPercent [expr 100.0*$below/$total]
    set abovePercent [expr 100.0*$above/$total]
    #count length of value strings to print the below/above lines in alignment
    set msl [string length [lindex [lsort -real [list $below $above]] end]] ;#msl - max string length
    puts $fileId [format "FOC Credit threshold:  $hexOrDecFormat" $minValue]
    puts $fileId "[format "Samples <= $hexOrDecFormat: %$msl.0f(%3.0f" $minValue $below $belowPercent]%)"
    puts $fileId "[format "Samples >  $hexOrDecFormat: %$msl.0f(%3.0f" $minValue $above $abovePercent]%)"
}

#Divide the register's histogram into buckets (num of buckets can be set in cfg file) and print to file.

#wrapper
proc getBuckets {valuesList fileId regHist reg} {
    global g_bucketType
    #getBucketsSimple or getBucketsWeighted
    getBuckets$g_bucketType $valuesList $fileId $regHist $reg
}

#simple bucketing algorithm - split to even buckets based on min/max values
proc getBucketsSimple {valuesList fileId regHist reg} {
    global g_numberOfHistBuckets
    global g_decimalHistVals
    array unset histArray *
    array set histArray $regHist

    set minVal [lindex $valuesList 0]
    set maxVal [lindex $valuesList end]

    set bucketSize [expr [expr $maxVal - $minVal] / $g_numberOfHistBuckets]

    array unset buckets *
    array set buckets {}
    foreach value $valuesList {
        set bucketMin [expr (($value - $minVal) / ($bucketSize))*$bucketSize + $minVal]; #Remember, the integer division do round down. So we get the bucket of the value
            if { ![info exists buckets($bucketMin)]} {
                set buckets($bucketMin) $histArray($value)
            } else {
                set buckets($bucketMin) [expr $buckets($bucketMin) + $histArray($value)]
            }
    }

    set sortedBucketKeys [lsort -integer [array names buckets]]
    set hexOrDecFormat [getHistValueFormat $reg]
    foreach bucketKey $sortedBucketKeys {
        set bucketMax [expr $bucketKey + $bucketSize]
        set numOfVals $buckets($bucketKey)
        puts $fileId [format "Number of samples >= $hexOrDecFormat && <= $hexOrDecFormat: %d" $bucketKey $bucketMax $numOfVals]
    }
}

#Divide the register's histogram into buckets based on number of samples foreach value
proc getBucketsWeighted {valuesList fileId regHist reg} {
    #The calculation is <approximate bucket size> = <number of total samples>/<number of buckets>
    #Samples are being accumulated untill they reach a "cricital mass" (the approximate bucket size).
    #If a specific value was sampled more times than the approx bucket size, it will have it's own bucket.
    #In this case the sum accumulated thus far will be set in it's own separate bucket
    #The buckets are in-order based on the value measured, so each bucket holds the
    #values from the previous bucket (non inclusive) to the next bucket (inclusive).
    #Example:
    #Let's say that the values measured are as follows: <value>:<number of samples of value>
    #0x1:4; 0x2:5; 0x3:10; ;0x4:13 ;0x5:2 ;0x6:3 ;0x7:3 ;0x8:3 ;0x9:7
    #And g_numberOfHistBuckets == 5  with 50 samples (approximate bucket size is 50/5=10)
    #The buckets that will be received are:
    # Number of samples >= 0x0 && <= 0x2: 9  (4+5)
    # Number of samples >  0x2 && <= 0x3: 10 (10)
    # Number of samples >  0x3 && <= 0x4: 13 (13)
    # Number of samples >  0x4 && <= 0x8: 11 (2+3+3+3)
    # Number of samples >  0x8 && <= 0x9: 7  (7)
    #It is possible that the actual number of buckets will be less than g_numberOfHistBuckets
    global g_numberOfHistBuckets
    global g_decimalHistVals
    array unset histArray *
    array set histArray $regHist

    set numOfValsTotal 0
    foreach value $valuesList {
        set numOfValsTotal [expr $numOfValsTotal + $histArray($value)]
    }
    set bucketLimit [expr $numOfValsTotal/$g_numberOfHistBuckets]

    #buckets array:
    #key   : Bucket's highest value
    #val[0]: Number of samples in bucket
    #val[1]: Bucket average value
    #val[2]: Bucket's lowest value
    array unset buckets *
    array set   buckets {}
    set currentCount 0
    set prevVal 0
    set rangeBottom -1
    set currentAvg 0
    set mostCommon {-1 0}
    foreach value $valuesList {
        if {$histArray($value) >= [lindex $mostCommon 1]} {
            set mostCommon [list $value $histArray($value)]
        }
        if {$rangeBottom == -1 } {
            set rangeBottom $value
        }
        set prevAvg $currentAvg
        set currentAvg   [expr $currentAvg*$currentCount + $value*$histArray($value)]
        set currentCount [expr $currentCount + $histArray($value)]
        set currentAvg   [expr $currentAvg/$currentCount]
        if {$currentCount >= $bucketLimit} {
            if {$histArray($value) >= $bucketLimit} {
                #if current value is >= the limit, it should be in a bucket of its own
                #subtract the value count from the total count (so far)
                set currentCount [expr $currentCount - $histArray($value)]
                #add a new entry for this value
                set buckets($value) [list $histArray($value) $value $value]
                if {$prevVal != 0 && $currentCount != 0} {
                    set buckets($prevVal) [list $currentCount $prevAvg $rangeBottom]
                    set prevVal 0
                }
            } else {
                set buckets($value) [list $currentCount $currentAvg $rangeBottom]
                set prevVal $value
            }
            set currentCount 0
            set rangeBottom -1
        } elseif {$value == [lindex $valuesList end]} {
            #if last
            set buckets($value) [list $currentCount $currentAvg $rangeBottom]
            set rangeBottom -1
        } else {
            set prevVal $value
        }
    }

    set sortedBucketKeys [lsort -integer [array names buckets]]
    set hexOrDecFormat [getHistValueFormat $reg]

    foreach bucketKey $sortedBucketKeys {
        set numOfVals   [lindex $buckets($bucketKey) 0]
        set avgVal      [lindex $buckets($bucketKey) 1]
        set startVal    [lindex $buckets($bucketKey) 2]

        if {$startVal == $bucketKey} {
            #print each line with approximate alignment to the rest of the lines
            set x "-"
            set xString [string repeat $x [expr [string length [format $hexOrDecFormat $bucketKey]]*2 + 6 ] ]
            puts $fileId [format "Number of samples of value $xString>$hexOrDecFormat : %d" $bucketKey $numOfVals]
        } else {
            puts $fileId [format "Number of samples >= $hexOrDecFormat && <= $hexOrDecFormat (AVG $hexOrDecFormat): %d" $startVal $bucketKey $avgVal $numOfVals]
        }
    }
    puts $fileId [format "Most common value is $hexOrDecFormat with %d samples." [lindex $mostCommon 0] [lindex $mostCommon 1]]
}

# move the cursor $linesCount lines up and delete the lines on the way
proc deleteLastLines {linesCount} {

    # for every line until the destination
    for {set i 0} {$i < $linesCount} {incr i} {
        #delete the line
        set deleteLineString "\x1b\[2K"
        puts -nonewline $deleteLineString
        # move the cursor one line up
        set cursorUpString "\x1b\[1F"
        puts -nonewline $cursorUpString
    }

    #delete the last line
    set deleteLineString "\x1b\[2K"
    puts -nonewline $deleteLineString
}

#======= Functions for performing stall cycles count =======
# save storms' stall masks in order to restore these masks after running the stall count
proc saveStormsStallMask {} {
    global g_storms
    global g_stallMasks
    global g_sem_regs
    global g_coreAString
    global g_coreBString
    global g_noCoreString

    #remove previous entries if such exist
    if {[info exists g_stallMasks]} {
        unset g_stallMasks
    }

    foreach core [dict keys $g_sem_regs] {
        dict for {storm sem} $g_storms {
            # read the mask of the stall reason
            set cycMaskReg _REG_FAST_MEMORY_STALL_CYCLES_MASK
            set cycMaskAddr [reg_offset -n $sem$cycMaskReg]
            set mask [reg read $cycMaskAddr]

            # save this storm's mask in the dictionary
            dict set g_stallMasks $core $sem $mask
        }
    }
}

# restore storms' stall masks to the value they had before running the stall count
proc restoreStormsStallMask {} {
    global g_stallMasks
    global g_sem_regs
    global g_coreAString
    global g_coreBString

    foreach core [dict keys $g_sem_regs] {
        dict for {sem mask} [dict get $g_stallMasks $core] {

            # read the mask of the stall reason
            set cycMaskReg _REG_FAST_MEMORY_STALL_CYCLES_MASK
            set cycMaskAddr [reg_offset -n $sem$cycMaskReg]
            # write the current mask
            reg write $cycMaskAddr $mask

            # read the number of stall cycles - only for clearing the register
            set val [readStallCyc $sem $core]

        }
    }
}

# Read the storm's stall cycles count
proc readStallCyc {sem {core ""}} {
    global g_chipType
    global g_arrChipTypes
    global g_coreAString
    global g_coreBString

    if {$g_chipType == $g_arrChipTypes(E5)} {
        if {$core eq $g_coreAString} {
            set stallCycsReg _REG_FAST_MEMORY_STORM_STALL_CYCLES_A
        } elseif {$core eq $g_coreBString} {
            set stallCycsReg _REG_FAST_MEMORY_STORM_STALL_CYCLES_B
        }
    } else {
        set stallCycsReg _REG_FAST_MEMORY_STORM_STALL_CYCLES
    }
    set stallCycsAddr [reg_offset -n $sem$stallCycsReg]
    set val [reg read $stallCycsAddr]
    return $val
}

# write the stall mask to all of the storms
proc writeStallMaskToStorms {mask} {
    global g_storms
    global g_sem_regs

    foreach core [dict keys $g_sem_regs] {
        dict for {storm sem} $g_storms {
            # write the mask of the stall reason that we want to check
            set cycMaskReg _REG_FAST_MEMORY_STALL_CYCLES_MASK
            set cycMaskAddr [reg_offset -n $sem$cycMaskReg]

            # write the current mask
            reg write $cycMaskAddr $mask

            # read the number of stall cycles - only for clearing the register
            set val [readStallCyc $sem $core]
        }
    }
}

# For each storm - check how many cycles that storm is stalled because of the masked reasons in $duration ms
proc allStormsStallCount {duration mask maskName} {
    global g_storms
    global g_stormStallCountLists
    global g_stormStallCycles
    global g_frequency
    global g_sem_regs

    writeStallMaskToStorms $mask

    # sleep for $duration ms
    after $duration

    foreach core [dict keys $g_sem_regs] {
        dict for {storm sem} $g_storms {
            # read the number of stall cycles since we cleared the register and append it to the string that will be printed (if it's not 0)
            set val [readStallCyc $sem $core]
            if { $val != 0 } {
                # calculate duration in microseconds (instead of milliseconds)
                set microsecDuration [expr $duration * 1000]
                # calculating utilization
                set per [calcUtilization $val $microsecDuration $g_frequency]
                # add the line to this storm's lines that will be printed
                dict set g_stormStallCountLists $core $storm [format "%s\n%s%sSTORM %s: %.2f%s" [dict get $g_stormStallCountLists $core $storm] $core $storm $maskName $per "%"]

                # sum the stall cycles for this storm
                set curr [dict get $g_stormStallCycles $core $storm]
                dict set g_stormStallCycles $core $storm [expr $curr + $per]
            }
        }
    }
}

# For each storm and each stalling reason - check how many cycles that storm is stalled because of this reasons in $duration ms
proc stallCount {duration isFile fileId iter displayUtil keepIters totalIterations {numStormUtilLines 0}} {
    global g_chipType
    global g_arrChipTypes
    global g_stallCountNames
    global g_storms
    global g_stormStallCountLists
    global g_stormStallCycles
    global g_lineCount
    global g_sem_regs

    set mask 0
    set len [llength $g_stallCountNames]

    # dict for saving the string of the results for each storm
    clearDict g_stormStallCountLists

    # dict for saving the total number of stall cycles for each storm
    clearDict g_stormStallCycles

    # save storms' stall masks in order to restore these masks after running the stall count
    saveStormsStallMask

    # for each storm - initialize the string that will hold the stall count results
    foreach core [dict keys $g_sem_regs] {
        dict for {storm sem} $g_storms {
            dict set g_stormStallCountLists $core $storm ""
            dict set g_stormStallCycles $core $storm 0
        }
    }

    #for progressLine
    puts "\n";
    # go through every reason
    for {set i 0} {$i < $len} {incr i} {
        set mask [expr 1 << $i]
        # perform stall count
        printProgressLineStallCount $i $len $iter $totalIterations
        allStormsStallCount $duration $mask [lindex $g_stallCountNames $i]
    }

    #delete progress line
    deleteLastLines 1;

    # add cpu utilization lines to the lines counter
    if {$displayUtil == 1} {
        #If storm utilization display is on, increase number of lines to delete
        set g_lineCount [expr $numStormUtilLines + $g_lineCount +2]
    }

    if {$iter != 0 && $keepIters == 0} {
        # move the cursor $g_lineCount lines up
        deleteLastLines $g_lineCount
    } else {
        #first iteration
        puts ""
    }
    set g_lineCount 4

    # for each storm - print the stall count results (looping over the dict for keeping storms' order)
    foreach core [dict keys $g_sem_regs] {
        dict for {storm sem} $g_storms {
            set str [dict get $g_stormStallCountLists $core $storm]
            if {![isEmptyString $str]} {
                puts "$str"

                foreach line [split $str \n] {
                    incr g_lineCount
                }

                # print to file if needed
                if {$isFile} {
                    puts $fileId $str
                }
            }
        }
    }

    # prepare summary line (looping over the dict for keeping storms' order)
    puts "\nStall count summary: Storm %stall; ,iteration number"
    foreach core [dict keys $g_sem_regs] {
        incr g_lineCount
        set line $core
        dict for {storm sem} $g_storms {
            set per [dict get $g_stormStallCycles $core $storm]
            append line [format " %s %.2f|" $storm $per]
        }
        append line [format " ,%s" $iter]
        # display summary line
        puts $line
        if {$isFile} {
            puts $fileId $line
        }
    }

    # restore storms' stall masks to the value they had before running the stall count
    restoreStormsStallMask
}

#======= Functions for collecting pcie statistics =======
proc pcieStats {duration isFile fileId iter keepIters} {

    global g_pcieStatsRegs
    global g_regsText
    global g_lineCount

    # get control register address
    set ctlReg PGLCS_REG_PGL_CS_PCIE_STATIS_CTL
    set ctlRegAddr [reg_offset -n $ctlReg]

    # calculate duration in microseconds (instead of milliseconds)
    set microsecDuration [expr $duration * 1000]

    # write the number of microseconds to the control registers and set the enable bit
    # bit 0: enable bit, bits 1-7: reserved, bits 8-31: time (in us) to collect statistics
    set ctlVal [expr 1+(($microsecDuration)<<8)]
    reg write $ctlRegAddr $ctlVal

    # sleep while statistics are being collected
    after $duration

    # only read the registers after the enable bit is reset to '0'
    set counter 0
    set ctlRegVal [reg read $ctlRegAddr]
    while {[expr ($ctlRegVal&0x1)] == 1} {
        # after 3 iterations, raise an error
        if {$counter >= 3} {
            error "PCIe statistics enable bit is not reset to '0'."
        }

        # sleep again if bit wasn't reset
        after $duration

        # read bit again
        set ctlRegVal [reg read $ctlRegAddr]

        incr counter
    }

    if {$iter != 0 && $keepIters == 0} {
        # move the cursor $g_lineCount lines up
        deleteLastLines $g_lineCount
    } else {
        puts ""
    }

    puts [format "PCIe Core statistics, iteration #: [expr $iter + 1]\n"]
    set g_lineCount 2

    # read collected statistic
    foreach pcieReg $g_pcieStatsRegs {
        set val [getRegVal $pcieReg]

        # mask the value
        set val [expr ($val&0xffffffffff)]

        # calculate number per second
        set per [expr ($val)/(($duration)/1000)]

        # print the register's text with the value we got
        regsub -all {(@\S+@)*(.*)} $pcieReg \\1 customFormat
        regsub -all {@(\S+)@} $customFormat \\1 customFormat
        if {[string length $customFormat] == 0} {
            set customFormat "%s"
        }
        set line [format "%s per second: $customFormat" $g_regsText($pcieReg) $per]

        puts $line
        if {$isFile} {
            puts $fileId $line
        }

        set g_lineCount [expr 1+$g_lineCount]
    }
}

#======= Auxiliary functions =======
#print all colors for selection
proc cTest {} {
    for {set i 0} {$i < 16} {incr i} {
        for {set j 0} {$j < 16} {incr j} {
            set color [expr $i*$j]
            puts -nonewline "\u001b\[38\;5\;${color}m$color "
        }
        puts -nonewline "\n"
    }
    puts -nonewline "\033\[0m"
}

#print in color
proc cPuts {str {color none} {isBright 0}} {
    puts [cStr $str $color $isBright]
}

#print in color - 256 colors
proc c256 {str {color 0}} {
    global g_noColor
    if {[info exists g_noColor] && $g_noColor} {
        return $str
    }

    if {![isNumeric $color]} {
        return $str
    }
    return "\u001b\[38\;5\;${color}m$str\033\[0m"
}

#get colored string
proc cStr {str {color none} {isBright 0}} {
    global g_noColor
    set cc 0

    if {[info exists g_noColor] && $g_noColor} {
        return $str
    }
    if {[isNumeric $color]} {
        set cc $color
    } else {
        switch -- $color {
            "red"     {set cc 31}
            "green"   {set cc 32}
            "yellow"  {set cc 33}
            "blue"    {set cc 34}
            "magenta" {set cc 35}
            "cyan"    {set cc 36}
            "white"   {set cc 37}
            "none"    {set cc  0}
        }

        if {$isBright} {
            set cc [expr $cc + 60]
        }
    }
    return "\033\[${cc}m$str\033\[0m"
}

proc isNumeric {var} {
    return [ regexp {^([0-9]+)$} $var ]
}

proc wordToUpperChars {word} {
    split [string toupper $word] {}
}

proc isInList {word list} {
    expr {[lsearch -exact $list $word] != -1}
}

#checks if string is empty - spaces are considered empty as well
proc isEmptyString {str} {
    regexp {^\s*$} $str
}

#For input <REG>.<FIELD> returns <REG>
proc getRegNoField {reg} {
    set regField [split $reg "."]
    if {[llength regField] == 1} {
        set reg [lindex $regField 0]
    }
    return $reg
}

proc cursor {{state "normal"}} {
    switch -- $state {
        "normal" {set op "cnorm"}
        "off"    {set op "civis"}
        "on"     {set op "cvvis"}
    }
    exec sh -c "tput $op || true"
}

#returns hex string of input
proc hex {args} {
    format "0x%X" $args
}


#parse external input file
proc parseExternalInput {externalFileName {isExt 1}} {
    if {!$isExt} {
        return ""
    }
    set fileId [open $externalFileName "r"]
    set fileData [read $fileId]
    close $fileId
    return [sl $fileData]
}

# add help
profile_help


#register list init
proc initRegsForProccessing {} {

    global g_arrChipTypes
    global g_chipType

    #Add registers to save a histogram of either Rx, Tx or PXP.
    #To sample a register from an array of registers, use <regName>+<index>. For example: YSEM_REG_FOC_CREDIT is a struct with six registers, indexed 0 to 5,
    #and each register represents a client. So if you want the value of the register the represents the client PBF, since PBF is client number 5 you should
    #write: YSEM_REG_FOC_CREDIT+5 .
    #To sample an address, add the prefix REG_ADDR_ to the required address.
    #To sample a register which is divided into low 32 bits and hi 32 bits, add the prefix LoAndHi# to the required register, for example:
    #LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXTLP_STATIS will return the value of (PGLCS_REG_PGL_CS_PCIE_TXTLP_STATIS_LO + (PGLCS_REG_PGL_CS_PCIE_TXTLP_STATIS << 32))
    #to specify custom format for a register - start the register name string with '@<custom format>@'. Example: @%d@USEM_REG_NUM_OF_THREADS.NUM_OF_THREADS will always be printed in decimal
    global g_rxHistRegs
    set g_rxHistRegs [sl {
        BRB_REG_MAC0_TC_OCCUPANCY_0                   ;#
        BRB_REG_MAC1_TC_OCCUPANCY_0                   ;#
        PRS_REG_CCFC_SEARCH_CURRENT_CREDIT            ;#
        PRS_REG_CCFC_LOAD_CURRENT_CREDIT              ;#
        PRS_REG_TCFC_LOAD_CURRENT_CREDIT              ;#
        TMLD_REG_TID_REMAIN_CREDITS                   ;#
        TMLD_REG_FOC_REMAIN_CREDITS                   ;#
        MULD_REG_TID_REMAIN_CREDITS                   ;#
        MULD_REG_FOC_REMAIN_CREDITS                   ;#
        MULD_REG_BD_PENDING_MSG_CTR                   ;#
        MULD_REG_SGE_PENDING_MSG_CTR                  ;#
        TCM_REG_FIC_INIT_CRD                          ;#
        MCM_REG_FIC_INIT_CRD                          ;#
        UCM_REG_FIC_INIT_CRD                          ;#
        TCM_REG_XX_FREE_CNT                           ;#
        MCM_REG_XX_FREE_CNT                           ;#
        UCM_REG_XX_FREE_CNT                           ;#
        TCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        MCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        UCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        TSEM_REG_FOC_CREDIT+1                         ;#
        MSEM_REG_FOC_CREDIT+2                         ;#
        USEM_REG_FOC_CREDIT+4                         ;#
        PSWRQ2_REG_VQ9_ENTRY_CNT                      ;#
        MSEM_REG_FOC_CREDIT+5                         ;#
        TCM_REG_AGG_CON_FIC_BUF_CRD                   ;#
        TCM_REG_SM_CON_FIC_BUF_CRD                    ;#
        TCM_REG_AGG_CON_BUF_CRD_AGG                   ;#
        TCM_REG_AGG_CON_BUF_CRD_AGGST                 ;#
        TCM_REG_AGG_CON_CMD_BUF_CRD_DIR               ;#
        TCM_REG_SM_CON_CMD_BUF_CRD_DIR                ;#
        TCM_REG_IS_PBF_FILL_LVL                       ;#
        TSDM_REG_QUEUE_FULL.INP_QUE_GRC_MASTER        ;#
        MSDM_REG_QUEUE_FULL.INP_QUE_GRC_MASTER        ;#
        @%d@TSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
        @%d@MSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
        @%d@USEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
    }]

    global rxHistRegs_E4
    set rxHistRegs_E4 [sl {
        PRS_REG_TCM_CURRENT_CREDIT                    ;#
    }]

    global rxHistRegs_AH
    set rxHistRegs_AH [sl {
        MSDM_REG_RMT_XCM_FIFO_FULL                    ;#
    }]

    global rxHistRegs_E5
    set rxHistRegs_E5 [sl {
        MSDM_REG_RMT_XCM_FIFO_FULL                    ;#
        PTLD_REG_FOC_REMAIN_CREDITS                   ;#
    }]

    global g_txHistRegs
    set g_txHistRegs [sl {
        QM_REG_PQFILLLVLTX+0                          ;#
        QM_REG_VOQCRDLINE+0                           ;#
        QM_REG_VOQCRDBYTE+0                           ;#
        XCM_REG_FIC_INIT_CRD                          ;#
        YCM_REG_FIC_INIT_CRD                          ;#
        PCM_REG_FIC_INIT_CRD                          ;#
        XCM_REG_XX_FREE_CNT                           ;#
        YCM_REG_XX_FREE_CNT                           ;#
        PCM_REG_XX_FREE_CNT                           ;#
        XCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        YCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        PCM_REG_XX_LCID_CAM_FILL_LVL                  ;#
        PBF_REG_BTB_ALLOCATED_BLOCKS_SHARED           ;#
        XYLD_REG_TID_REMAIN_CREDITS                   ;#
        XYLD_REG_FOC_REMAIN_CREDITS                   ;#
        PBF_REG_BTB_ALLOCATED_BLOCKS_VOQ0             ;#
        PBF_REG_YCMD_QS_CMD_CNT_VOQ0                  ;#
        NIG_REG_RX_FLOWCTRL_STATUS                    ;#
        BTB_REG_BLOCK_OCCUPANCY                       ;#
        PSWRQ2_REG_VQ4_ENTRY_CNT                      ;#
        PSWRQ2_REG_SR_CNT_PER_VQ_4                    ;#
        PSWRQ2_REG_BLK_CNT_PER_VQ_4                   ;#
        PSWRQ2_REG_BLK_CNT_PER_VQ_6                   ;#
        PSWRQ2_REG_BLK_CNT                            ;#
        PSWRQ2_REG_BLK_NUM_CFG                        ;#
        PSWRQ2_REG_SR_CNT                             ;#
        PSWRQ2_REG_SR_NUM_CFG                         ;#
        YSEM_REG_FOC_CREDIT+5                         ;#
        XCM_REG_QM_ACT_ST_CNT+0                       ;#
        QM_REG_PQFILLLVLTX+2                          ;#
        QM_REG_PQFILLLVLTX+3                          ;#
        QM_REG_PQSTSTX+0                              ;#
        QM_REG_PQSTSTX+2                              ;#
        QM_REG_PQSTSTX+3                              ;#
        QM_REG_CMCRD_4                                ;#
        QM_REG_CMCRD_5                                ;#
        QM_REG_CMCRD_8                                ;#
        QM_REG_CMCRD_9                                ;#
        QM_REG_OUTLDREQCRDCONNTX                      ;#
        XCM_REG_IS_MSDM_FILL_LVL                      ;#
        PSEM_REG_FOC_CREDIT+0                         ;#
        PSEM_REG_FOC_CREDIT+1                         ;#
        XCM_REG_QM_ACT_ST_CNT+2                       ;#
        XCM_REG_QM_ACT_ST_CNT+3                       ;#
        PSDM_REG_QM_FULL                              ;#
        PSDM_REG_QUEUE_FULL                           ;#
        QM_REG_PXP_REQ_CRD                            ;#
        @%d@XSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
        @%d@YSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
        @%d@PSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS    ;#
    }]
    global txHistRegs_E4
    set txHistRegs_E4 [sl {}]

    global txHistRegs_E5
    set txHistRegs_E5 [sl {
        PBF_REG_BTB_OUTSTANDING_BLOCKS_COUNTER        ;#
        XSEM_REG_DBG_QUEUE_MAX_THREAD_VALUE           ;#
        XSEM_REG_DBG_QUEUE_MAX_SLEEP_VALUE            ;#
        XSEM_REG_DBG_QUEUE_FOC_MAX_VALUE              ;#
    }]

    if {$g_chipType == $g_arrChipTypes(E5)} {
        #E5
        set g_rxHistRegs [concat $g_rxHistRegs $rxHistRegs_E5]
        set g_txHistRegs [concat $g_txHistRegs $txHistRegs_E5]
    } else {
        #E4
        set g_rxHistRegs [concat $g_rxHistRegs $rxHistRegs_E4]
        set g_txHistRegs [concat $g_txHistRegs $txHistRegs_E4]
    }

    if {$g_chipType == $g_arrChipTypes(AH)} {
        #Arrowhead specific
        set g_rxHistRegs [concat $g_rxHistRegs $rxHistRegs_AH]
    }

    global g_pxpHistRegs
    if {$g_chipType == $g_arrChipTypes(AH)} {
        #arrowhead
        set g_pxpHistRegs [sl {
            PSWRQ2_REG_CREDIT_WR_STS
            PSWRQ2_REG_CREDIT_RD_STS
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_HEADER_FC_CREDIT
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_DATA_FC_CREDIT
        }]
    } elseif {$g_chipType == $g_arrChipTypes(BB)} {
        #bb
        set g_pxpHistRegs [sl {
            PSWRQ2_REG_CREDIT_WR_STS
            PSWRQ2_REG_CREDIT_RD_STS
            PGLCS_REG_PGL_CS_TL_HDR_FC_ST
            PGLCS_REG_PGL_CS_TL_DAT_FC_ST
        }]
    } else {
        #e5
        set g_pxpHistRegs  [sl {
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.P_DATA_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.N_DATA_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.C_DATA_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.P_CMD_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.N_CMD_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.C_CMD_VOLUME
            PGL2PEM_REG_PEM_NATIVE_PEM_CPL_LUT_VALID.TAG
        }]
    }

    #PCI registers'
    global g_pciRegs
    if {$g_chipType == $g_arrChipTypes(AH)} {
        set g_pciRegs [sl {
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_HEADER_FC_CREDIT
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_DATA_FC_CREDIT
        }]
    } elseif {$g_chipType == $g_arrChipTypes(E5)} {
        set g_pciRegs {}
    } else {
        set g_pciRegs [sl {
            PGLCS_REG_PGL_CS_TL_HDR_FC_ST
            PGLCS_REG_PGL_CS_TL_DAT_FC_ST
        }]
    }

    #If the needed value is not a full register, select the bitmask and offset for reading it:
    global g_offsetRegsArgs
    array unset g_offsetRegsArgs *
    array set g_offsetRegsArgs [sl {
        #PCI registers'
        QM_REG_VOQCRDLINE+0                    {0x7FFF 0}
        QM_REG_VOQCRDBYTE+0                    {0x7FFFFF 0}

        #SDM registers
        TSDM_REG_QUEUE_FULL.INP_QUE_GRC_MASTER {0x100 8}
        MSDM_REG_QUEUE_FULL.INP_QUE_GRC_MASTER {0x100 8}

        #SEMI registers
        XSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
        YSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
        PSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
        TSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
        MSEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
        USEM_REG_NUM_OF_THREADS.NUM_OF_THREADS {0x1F 0}
    }]

    #special registers that needs to be presented as <constant max value>-<measured value> instead of <measured value>
    #array 'key' is the register name, array 'val' is that register's accomodating 'constant max value' register to be subtracted from
    global g_inverseRegisters
    array unset g_inverseRegisters *
    array set g_inverseRegisters [sl {
        PSWRQ2_REG_BLK_CNT              PSWRQ2_REG_BLK_NUM_CFG ;#
        PSWRQ2_REG_BLK_CNT_PER_VQ_4     PSWRQ2_REG_BLK_NUM_CFG ;#
        PSWRQ2_REG_SR_CNT               PSWRQ2_REG_SR_NUM_CFG  ;#
        PSWRQ2_REG_SR_CNT_PER_VQ_4      PSWRQ2_REG_SR_NUM_CFG  ;#
    }]

    global g_offsetRegsArgs
    if {$g_chipType == $g_arrChipTypes(AH)} {
        array set g_offsetRegsArgs [sl {
            #PCI registers'
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_HEADER_FC_CREDIT {0xFF000 12}
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_DATA_FC_CREDIT {0xFFF 0}
        }]
    } else {
        array set g_offsetRegsArgs [sl {
            #PCI registers'
            PGLCS_REG_PGL_CS_TL_HDR_FC_ST {0xFF00 8}
            PGLCS_REG_PGL_CS_TL_DAT_FC_ST {0xFFF 0}
        }]
    }

    if {$g_chipType == $g_arrChipTypes(E5)} {
        array set g_offsetRegsArgs [sl {
            #pem_csr regs
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.P_DATA_VOLUME {0x00000000000000FF  0}
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.N_DATA_VOLUME {0x00000000000FF000 12}
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.C_DATA_VOLUME {0x00000000FF000000 24}
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.P_CMD_VOLUME  {0x000000FF00000000 32}
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.N_CMD_VOLUME  {0x000FF00000000000 44}
            PGL2PEM_REG_PEM_NATIVE_PEM_EBO_FIFO_STATUS.C_CMD_VOLUME  {0xFF00000000000000 56}

            PGL2PEM_REG_PEM_NATIVE_PEM_CPL_LUT_VALID.TAG {0x3FF  0}
        }]
    }

    #The following registers groups will return the sum of its members when calling proc getRegVal
    global g_regsGroups
    array unset g_regsGroups *
    array set g_regsGroups [sl {
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY {\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_0\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_1\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_2\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_3\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_4\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_5\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_6\
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY_7\
        }
    }]

    #The following Registers' will be printed to screen during the run.
    global g_regsToCount
    set g_regsToCount [sl {
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY
        NIG_REG_RX_FLOWCTRL_STATUS
    }];  # what is NIG_REG_RX_FLOWCTRL_STATUS ???


    #to specify FOC interface width, add '(<width in bits>b)' at the end of the appropriate array value.
    #For example:
    #MSEM_REG_FOC_CREDIT+2 "MSEM_REG_FOC_CREDIT->MULD (256b)"
    global g_regsFocCredit
    array set g_regsFocCredit [sl {
        YSEM_REG_FOC_CREDIT+5                   "(YSEM_REG_FOC_CREDIT+5) YSEM->PBF FOC CREDIT (256b)"
        TSEM_REG_FOC_CREDIT+1                   "(TSEM_REG_FOC_CREDIT+1) TSEM->RSS FOC CREDIT"
        PSEM_REG_FOC_CREDIT+0                   "(PSEM_REG_FOC_CREDIT+0) PSEM->PCM FOC CREDIT"
        PSEM_REG_FOC_CREDIT+1                   "(PSEM_REG_FOC_CREDIT+1) PSEM->PBF FOC CREDIT (256b)"
        MSEM_REG_FOC_CREDIT+2                   "(MSEM_REG_FOC_CREDIT+2) MSEM->MULD FOC CREDIT (256b)"
        MSEM_REG_FOC_CREDIT+5                   "(MSEM_REG_FOC_CREDIT+5) MSEM->PRM FOC CREDIT (256b)"
        USEM_REG_FOC_CREDIT+4                   "(USEM_REG_FOC_CREDIT+4) USEM->PRM FOC CREDIT (256b)"
    }]


    #To replace a register with a text string (only for printing!) supply the text here
    global g_regsText
    array unset g_regsText *
    array set g_regsText [sl {
        NIG_REG_STAT_RX_BRB_DISCARD_PRIORITY    "BRB drops"
        NIG_REG_RX_FLOWCTRL_STATUS              "RX pause"
        PSWRQ2_REG_BLK_CNT                      "PSWRQ2_REG_BLK_CNT (Outstanding; Max is #MAXVALSTRING#)"
        PSWRQ2_REG_BLK_CNT_PER_VQ_4             "PSWRQ2_REG_BLK_CNT_PER_VQ_4 (Outstanding; Max is #MAXVALSTRING#)"
        PSWRQ2_REG_SR_CNT                       "PSWRQ2_REG_SR_CNT (Outstanding; Max is #MAXVALSTRING#)"
        PSWRQ2_REG_SR_CNT_PER_VQ_4              "PSWRQ2_REG_SR_CNT_PER_VQ_4 (Outstanding; Max is #MAXVALSTRING#)"
    }]

    array set g_regsText [array get g_regsFocCredit]

    if {$g_chipType == $g_arrChipTypes(AH)} {
        array set g_regsText [sl {
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_HEADER_FC_CREDIT     "PCI header credit"
            PGLCS_REG_PGL_CS_TX_P_FC_CREDIT_STATUS_OFF.TX_P_DATA_FC_CREDIT       "PCI data credit"
        }]
    } else {
        array set g_regsText [sl {
            PGLCS_REG_PGL_CS_TL_HDR_FC_ST                "PCI header credit"
            PGLCS_REG_PGL_CS_TL_DAT_FC_ST                "PCI data credit"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXTLP_STATIS   "sent TLP bytes"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXDLLP_STATIS  "sent DLLP bytes"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXOS_STATIS    "sent ordered set bytes"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXTLP_STATIS   "received TLP bytes"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXDLLP_STATIS  "received DLLP bytes"
            LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXOS_STATIS    "received ordered set bytes"
            PGLCS_REG_PGL_CS_PCIE_PLRXERR_STATIS         "errors detected by Physical Layer Receiver"
            PGLCS_REG_PGL_CS_PCIE_RXDLLPERR_STATIS       "DLLP CRC errors detected by Data Link Layer"
            PGLCS_REG_PGL_CS_PCIE_RXTLPERR_STATIS        "TLP LCRC and sequence number errors detected by Data Link Layer"
        }]
    }

    # ======= stall count configuration =======
    global g_stallCountNames_E5
    set g_stallCountNames_E5 [sl {
        MUX_COMMON_COL_STALL        ;#0
        CAM_REQ_STALL               ;#1
        CAM_RSP_STALL               ;#2
        EXT_ST_STALL                ;#3
        EXT_LD_STALL                ;#4
        COMMON_MUX_VFC_STORE_STALL  ;#5
        COMMON_MUX_VFC_LOAD_STALL   ;#6
        DEBUG_STALL                 ;#7
        MISC_LOCAL_MUX_SEL_STALL_EN ;#8
        MISC_LOCAL_MUX_EXT_STALL    ;#9
        MISC_LOCAL_MUX_INT_STALL    ;#10
        PRAM_MUX_BKPT_STALL         ;#11
        RBC_MUX_STALL_COMMON        ;#12
        FIN_STALL                   ;#13
        MUX_COMMON_MUX_IRAM_STALL   ;#14
        PRAM_MUX_PRTY_STALL         ;#15
        STORM_ATTN_STALL            ;#16
        RBC_MUX_STALL_STORM         ;#17
        PRAM_MUX_PIPE_STALL         ;#18
        MUX_LOCK_STALL              ;#19
        RAM_MUX_BKPT_STALL          ;#10
        MISC_LOCAL_MUX_OTHER_STALL  ;#21
    }]

    global g_stallCountNames
    set g_stallCountNames [sl {
        ECC_RMW                     ;#0
        CAM_STORE                   ;#1
        CAM_LOAD                    ;#2
        EXTERNAL_STORE              ;#3
        EXTERNAL_LOAD               ;#4
        VFC_STORE                   ;#5
        VFC_LOAD                    ;#6
        FAST_DEBUG                  ;#7
        SELECTIVE                   ;#8
        EXTERNAL_STALL              ;#9
        INTERNAL_STALL              ;#10
        BREAKPOINT                  ;#11
        RBC0                        ;#12
        FIN                         ;#13
        ECC_ERROR                   ;#14
        PRAM_PARITY                 ;#15
        ATTENTION                   ;#16
        RBC1                        ;#17
        RBC2                        ;#18
    }]
    if {$g_chipType == $g_arrChipTypes(E5)} {
        set g_stallCountNames $g_stallCountNames_E5
    }

    #======= Pcie statistics configuration =======
    global g_pcieStatsRegs
    set g_pcieStatsRegs [sl {
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXTLP_STATIS
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXDLLP_STATIS
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_TXOS_STATIS
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXTLP_STATIS
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXDLLP_STATIS
        LoAndHi#PGLCS_REG_PGL_CS_PCIE_RXOS_STATIS
        PGLCS_REG_PGL_CS_PCIE_PLRXERR_STATIS
        PGLCS_REG_PGL_CS_PCIE_RXDLLPERR_STATIS
        PGLCS_REG_PGL_CS_PCIE_RXTLPERR_STATIS
    }]

}
