###########################################################################
#
# Copyright(c) 2014-2017 Cavium, Inc., all rights reserved
# Proprietary and Confidential Information.
#
# Module Description:
# This script tests access to the EVEREST registers, based on an RDF file
# which contains a list of [name, offset, access_type, count, reset_value]
# for each register.
#
#This script uses the regsUtils.tcl script to parse the RDF into
# global arrays: regsOffset regsType regsSize and regsResetVal                                              
#
###########################################################################

# Specify any dependency on other packages
package require Tcl

# Specify the package this module provides
package provide _regtest 1.0

namespace eval _regtest {
    
    # Export the package
    namespace export regtest   
}

# Give a short-name for the regression test.
interp alias {} regtest {}  "_regtest::regtest_main"
interp alias {} regtest_all {}  "_regtest::regtest_all"

#Globals:

#Blocks table is an array of block name & block offset
global blocksTable

#Blocks to skip
global skipBlocks { PXP2 MCP }

#Registers information in 4 arrays
global regsName
global regsOffset
global regsSize
global regsType
global regsResetVal
global regsProtectRead
global regsProtectWrite
global BB_A0_ExceptionList
global blocksTable



# -----------------------------------------------------------------
# This is the main procedure of the regtest.
# The proc parses the input options, builds the blocks offset table and
# Calls the regtest_test procedure which parses the RDF file and
# performs the reset_value and access_type tests for each register.
# -----------------------------------------------------------------
proc _regtest::regtest_main {{args}} {

    #use the global variables
    
    global skipBlocks
    global regsName
    global regsOffset
    global regsSize
    global regsType
    global regsResetVal
    global regsProtectRead
    global regsProtectWrite
    global BB_A0_ExceptionList
    global blocksTable

    variable isEmul
    variable isFpga

    # skip this blocks from testing.
    set skipBlocks    { PXP2 MCP }
    set testBlock     ""
    set endBlock      ""
    set returnVal     0
    set skipCommonReg 0
    set extended      0
    
    # Parse the options.
    for {set i 0} {$i < [llength $args]} {incr i} {
        switch -- [lindex $args $i] {
            "-f"         { incr i; set rdfFile [lindex $args $i] }
            "-skip"      { incr i; lappend skipBlocks [lindex $args $i] }
            "-b"         { incr i; set testBlock [lindex $args $i] }
            "-e"         { incr i; set endBlock [lindex $args $i] }
            "-skipCommon" { set skipCommonReg 1 }    
            "-extended"  { set extended 1 }            
            "-help"      { rt_help; return 1 }
            default      { rt_help; return 1 }
        }
    }

    #Build excption list (if not already built)
    if {![info exists BB_A0_ExceptionList]} {
        puts "Build exception list"
        build_exception_list 
    }
 
    # parse the rdf file into regsOffset array    
    if {[parse_rdf -skip $skipBlocks] == -1} {
        return -1
    }
    
    if {$extended} {
        # reset the chip and halt the MCP before testing
        # only in extended mode
        prep_for_test
    }

    #Clear regFifo results before test
    puts "dbgRegFifo"
    dbgRegFifo      
    
    # run the regtest_test on all the registers in regsOffset
    puts "Start Testing ..."
    
    set returnVal [regtest_test $testBlock $skipCommonReg $extended]

    # Verify there are no GRC errors
    puts "dbgRegFifo"
    set regFifoResults [dbgRegFifo]
    if {![string match "*fifo contained 0 elements" $regFifoResults]} {
        # mark returnVal as failed
        if {$returnVal == "passed"} {set returnVal "failed"}
        lappend returnVal "ERROR: There are grc errors !\n $regFifoResults"
        puts "ERROR: There are grc errors !\n $regFifoResults"
    }

    #Verify there are no PARITY error
    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS"
    append mask_reg "_REG_PRTY_MASK"

    set offset [reg_offset -n $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n  $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal "ERROR: $prty_reg returned $val (mask $mask)!"
            puts "ERROR: $prty_reg returned $val (mask $mask)!"
        }
    }

    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS_H_0"
    append mask_reg "_REG_PRTY_MASK_H_0"
    set offset [reg_offset -n  $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n  $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal "ERROR: $prty_reg returned $val (mask $mask) !"
            puts "ERROR: $prty_reg returned $val (mask $mask) !"
        }
    }

    set prty_reg $testBlock
    set mask_reg $testBlock
    append prty_reg "_REG_PRTY_STS_H_1"
    append mask_reg "_REG_PRTY_MASK_H_1"
    set offset [reg_offset -n $prty_reg]
    if {$offset != -1} {
        set mask [reg_read -n $mask_reg]
        set val [reg read $offset]
        if {($val & ~$mask) != 0} {
            # mark returnVal as failed
            if {$returnVal == "passed"} {set returnVal "failed"}
            lappend returnVal  "ERROR: $prty_reg returned $val (mask $mask)!"
            puts "ERROR: $prty_reg returned $val (mask $mask)!"
        }
    }
           
    if {$returnVal == "passed"} {
        puts "***********************************************************"
        puts "**             $testBlock REG TEST PASSED                  "
        puts "***********************************************************"
    } else {
        puts "***********************************************************"
        puts "**             $testBlock REG TEST FAILED !!!!             "
        puts "***********************************************************"
    }
    
    if {$extended} {
        # Restore reset values and run MCP
        # only in extended mode
        cleanup_after_test
    }
        
    return $returnVal
 
}

# -----------------------------------------------------------------
# This proc checks if register should be skipped from reset value check
# -----------------------------------------------------------------
proc _regtest::skip_reset_val_check { {name} } {
    # Skip these register test of reset values
    if {[string match "*SEM_REG_FAST_MEMORY*" $name] ||
        [string match "NIG_REG_LB_ARB_PRIORITY_CLIENT" $name] ||
        [string match "NIG_REG_TX_ARB_PRIORITY_CLIENT" $name] ||
        [string match "NIG_REG_LLH_PF_CLS_FILTERS_MAP" $name] || 
        [string match "NIG_REG_TX_DRAIN_EN" $name] ||
        [string match "PRS_REG_FIFO_EMPTY_FLAGS" $name] ||
        [string match "NIG_REG_TSGEN_FREECNT_*" $name] ||
        [string match "NIG_REG_TSGEN_SYNC_TIME_*" $name] ||
        [string match "DORQ_REG_RTC_CUR_VAL" $name] ||
        [string match "PCIE_CTL_REG_PCIE_STATUS" $name] ||
        [string match "PCIE_CTL_REG_PCIE_PHY_STATUS*" $name]} {
	    return 1
    } else {
	    return 0
    }    
}

# -----------------------------------------------------------------
# This proc parses each line in the rdf file and performs for each line
# the test_reset_value() of the register and test_access_type()
# once by writing all 0 to the register, then writing all 1 to register.
# -----------------------------------------------------------------
proc _regtest::regtest_test { {testBlock} {skipCommonReg} {extended} } {        
        global regsName
        global regsOffset
        global regsSize
        global regsType
        global regsResetVal
        global regsSplitVal
        global regsPhase
        global regsProtectRead
        global regsProtectWrite

        variable isEmul
        variable isFpga

        # Initialize variables
        set line         0
        set testResult   ""
        set offset       0
        set dataWidth    32 
        set curFunc      $::current(FUNC)
        
        # Perform reset value check, only in extended mode
        if {$extended} {
            puts "Start reset value check .."

            # Go through each register in the regsOffset array
            # And check the reset value of the register
            for {set idx 0 } {$idx < [llength $regsName]} {incr idx} {
                set b_run_test  0
                set name [lindex $regsName $idx]
                
                # Skip registers which are not in the tested block
                set tmpStr [format "%s_REG_*" $testBlock]
                if { ($testBlock != "") && (![string match "$tmpStr" $name]) } {
                    continue
                }
		
		# Skip blocks that are not reset by reset_chip
                if {[string match "PSW*" $testBlock] ||
		    [string match "GRC" $testBlock] ||
		    [string match "MISC" $testBlock] ||
		    [string match "CNIG" $testBlock] ||
		    [string match "MISCS" $testBlock] ||
                    [string match "IGU" $testBlock] ||
                    [string match "PTU" $testBlock] ||
                    [string match "NWM" $testBlock] ||
                    [string match "PGLUE_B" $testBlock] ||
                    [string match "PXPREQBUS" $testBlock] ||
                    [string match "MCP2" $testBlock]} {
                    continue
                }
               
                set offset [lindex $regsOffset $idx]
                set dataWidth [lindex $regsSize $idx]
                set accessType [lindex $regsType $idx]
                set resetVal [lindex $regsResetVal $idx]
                set splitVal [lindex $regsSplitVal $idx]

                # For WB regs, calc the reg reserved width (the upper 2^x)
                if {$dataWidth > 32} {
                    set power [expr (log($dataWidth)/log(2))]
                    #round up the power value
                    if { $power != [expr (int($power))]} {
                        set power [expr (int($power))]
                        incr power
                    }
                    set width [expr (int(pow(2,$power)))]
                    set noWords [expr ($width/32)]
                 } else  {
                    set noWords 1
                 }
                
                # Check if the offset field is a range (startOffset - endOffset)
                # If so, parse the start offset and end offset.
                if { ([string first "-" $offset] != -1) } {
                    set startOffset [get_start_offset $offset]
                    set endOffset   [get_end_offset $offset $startOffset]
                } else  {
                    # Or a single register
                    set startOffset $offset
                    set endOffset   $offset
                }            
               
                # If reset value is not define, skip testing it
                if { $resetVal == "X" } {
                   continue
                }
                                                                          
                # Check if need to skip register cause its in the exception list            
                if { [skip_register $name]} {
                    continue
                }
		
		if {[skip_reset_val_check $name]} {
		    continue
		}
                
                # Check if need to run reset_test on this register            
                set b_run_test [check_split_reg $skipCommonReg $splitVal $curFunc]            
                
                # Perform reset test to the register
                if {$b_run_test} {
                    puts "DEBUG: Check reset value of register $name ($startOffset-$endOffset)"
                    # Test the registers range (or the single register)
                    for {set testingOffset $startOffset} \
                        {$testingOffset <= $endOffset} {incr testingOffset [expr ($noWords * 4)]  } {                  
                        
                        set resetResult [test_reset_value $name $testingOffset $resetVal $dataWidth $accessType]
                        
                        if {$resetResult != "passed"} {
                            # if test result is empty, set it to the reset test result
                            if {$testResult == "" } {
                                set testResult $resetResult
                            } else {
                                #Otherwise, append the reset test result to the existing result
                                lappend testResult "$resetResult \n"
                            }
                        }
                    }
                }
            }
        }
              
        puts "Start access check .."
        set noRegs 0
        set mode [nvm secure_mode]
        
        # Go through each register in the regsOffset array
        # And check test the access type
        for {set idx 0 } {$idx < [llength $regsName]} {incr idx} {
            set b_run_test 0
            set name [lindex $regsName $idx]

            # Skip registers which are not in the tested block
            set tmpStr [format "%s_REG_*" $testBlock]
            if { ($testBlock != "") && (![string match "$tmpStr" $name]) } {
                continue
            }
            
            set offset [lindex $regsOffset $idx]
            set dataWidth [lindex $regsSize $idx]
            set accessType [lindex $regsType $idx]
            set resetVal [lindex $regsResetVal $idx]
            set splitVal [lindex $regsSplitVal $idx]
            set protectRead [lindex $regsProtectRead $idx]
            set protectWrite [lindex $regsProtectWrite $idx]
          
            # Check if need to skip register cause its in the exception list            
            if { [skip_register $name]} {                             
                continue
            }

            # Check if need to run access_test on this register            
            set b_run_test [check_split_reg $skipCommonReg $splitVal $curFunc]

            # Don't test the WR (write-only) registers
            if { (($accessType == "WR") || ($accessType == "W"))} {
                set b_run_test 0
            }

            # Don't test registers with protect read UA, HV
            if {(($mode == "Secure Mode") || ($mode == "Fully Secure Mode")) && ($protectRead != "PDA")} {
                #puts "Skip register $name cause of protect read."
                set b_run_test 0
            }            
            
            if {$b_run_test} {

                incr noRegs
                
                #puts "Testing access of register $name"
                
                # For WB regs, calc the reg reserved width (the upper 2^x)
                if {$dataWidth > 32} {
                    set power [expr (log($dataWidth)/log(2))]
                    #round up the power value
                    if { $power != [expr (int($power))]} {
                        set power [expr (int($power))]
                        incr power
                    }
                    set width [expr (int(pow(2,$power)))]
                    set noWords [expr ($width/32)]
                } else  {
                    set noWords 1
                }

                # Check if the offset field is a range (startOffset - endOffset)
                # If so, parse the start offset and end offset.
                if { ([string first "-" $offset] != -1) } {
                    set startOffset [get_start_offset $offset]
                    set endOffset   [get_end_offset $offset $startOffset]                        
                } else  {
                    # Or a single register
                    set startOffset $offset
                    set endOffset   $offset
                }            
                    
                # Test the registers range (or the single register)
                for {set testingOffset $startOffset} \
                    {$testingOffset <= $endOffset} {incr testingOffset [expr ($noWords * 4)]  } { 
                        
                        # Test access type by writing all 0
                        set access0Result [test_access_type $name $testingOffset \
                                            $accessType 0x0 $dataWidth $protectRead $protectWrite $mode]
                        
                        # Test access type by writing all 1
                        set access1Result [test_access_type $name $testingOffset \
                                           $accessType 0xFFFFFFFF $dataWidth $protectRead $protectWrite $mode]
                        
                        if { $access0Result != "passed"} {
                            # if test result is empty, set it to the access test result
                            if {$testResult == "" } {
                                set testResult $access0Result
                            } else {
                                #Otherwise, append the access test result to the existing result
                                lappend testResult $access0Result
                            }
                        }
                        if { $access1Result != "passed"} {
                            # if test result is empty, set it to the access test result
                            if {$testResult == "" } {
                                set testResult $access1Result
                            } else {
                                #Otherwise, append the access test result to the existing result
                                lappend testResult $access1Result
                            }
                        }
                }            
            }
            
            #By default (not extended mode), test only one R/W register per block
            if {!$extended && ($accessType == "RW") && ($noRegs >= 1)} {
                break;
            }
        }        
        
        if { $noRegs != 0 } {
                puts "Tested $noRegs registers \n"
        } else {
          puts "ERROR: no registers were tested\n"
          set testResult "passed"             
        }
        
        #no errors were found, so return "passed"
        if {$testResult == ""} {
            set testResult "passed"
        }

        return $testResult
    }



# -----------------------------------------------------------------
# This proc prints out the usage information
# -----------------------------------------------------------------
proc _regtest::rt_help {}  {
    
    puts stdout "Usage:   regtest -b <BLOCK>              "
    puts stdout "Optional:                                "
    puts stdout "          -extended: runs extended tests "
    puts stdout "          -f     <rdf file name>         "
    puts stdout "          -skip  <skip blocks>           "
    puts stdout "          -e     <end block>             "
    puts stdout "          -skipCommon                    "
    puts stdout "If no rdf file is specified, the test    "
    puts stdout "will use the rdf under current directory."
}

# -----------------------------------------------------------------
# This proc resets the chip and makes the initializations required.
# -----------------------------------------------------------------
proc _regtest::prep_for_test {}  {  
    variable isEmul
    variable isFpga

    puts "Run _regtest::prep_for_test"
    set isEmul [is_emul]
    set isFpga [is_fpga]

    if {!$isEmul && !$isFpga} {
       mcp halt
    } 

    # Reset the chip
    reset_chip 1

    if {$isEmul} {
        delay ms 3000	
    } 

    # This initialization is done in dev_if_init
    #reg_write -n PGLUE_B_REG_START_INIT_PTT_GTT -v 1
    #reg_write -n PGLUE_B_REG_START_INIT_INB_INT_MEM -v 1
    
    #Dummy read required for emulation
    if {$isEmul} {
	delay ms 10
        reg read 0x8000	
    }
	
}

# -----------------------------------------------------------------
# This proc cleans up after tests
# -----------------------------------------------------------------
proc _regtest::cleanup_after_test {}  {  
  variable isEmul
  variable isFpga

  puts "Run _regtest::cleanup_after_test"
  if {$isEmul} {
      reset_chip 1
      delay ms 3000
  } elseif {$isFpga} {
  	reset_chip 1
  } else {
      mcp run
      reset_chip
  }
}


# -----------------------------------------------------------------
# This proc tests the access type of the register by writing a value
# to the register and checking the read value in different cases:
# incase of RO - Just read
# incase of RW - that the register value has changed
# incase of RC, ST - that the second read cleared the register.
# -----------------------------------------------------------------
proc _regtest::test_access_type {{name} {offset} {access} {value} {dataWidth} {protectRead} {protectWrite} {mode}}  {
    
    # Save the number of words (32 bit) of the register
    set noWords [expr ($dataWidth/32)]
    if { [expr ($dataWidth % 32) != 0] } {
        incr noWords
    }
       
    # Save the register's original value and mask in arrays
    # incase of WB (wide bus) each part (32bit) of the register's original
    # value is saved in an array orgValue.
    # Otherwise (dataWidth<32bit) original value is in 1'st element of the array.
    for {set i 0} {$i < $noWords} {incr i} {
         set wordOffset [expr ($offset + $i*4)]
         set orgValue($i,value) [reg read $wordOffset]
                     
         # Create the mask according to the register's dataWidth
         if {$i < [expr ($noWords-1)]} {
            set mask($i,value) 0xFFFFFFFF            
         } else  {
             # The last part of the register's mask is according to remaining
             # data width e.g 100 bit register, mask is for remaining 4 bits (0x1111)
             set mask($i,value) [get_mask $dataWidth]             
         }         
    } 
    
    
    # Don't attempt to write to registers with protectWrite UA, HV.
    if {(($mode == "Secure Mode") || ($mode == "Fully Secure Mode")) && ($protectWrite != "PDA")} {
        #puts "skip protectWrite register $name"
        return "passed"
    }
       
    # Write to the register words the value
    if {($access != "ST") && ($access != "RC") && ($access != "R") && ($access != "WB_R")} {
        puts "DEBUG: Write to register $name access $access value $value"
        for {set i 0} {$i < $noWords} {incr i} {
            set wordOffset [expr ($offset + $i*4)]
            if {$i < [expr ($noWords-1)]} {
                set writtenValue($i,value) $value
                reg write $wordOffset $value            
            } else  {
                # write to the remaining part of the register
                # only to the valid bits (according to mask)
                set writtenWord [expr {$value & $mask($i,value)}]            
                set writtenValue($i,value) $writtenWord
                reg write $wordOffset $writtenWord            
            }
        }
    }
    
    set accessTestResult ""

    # Read the register's value
    # Incase of WB go through all its words (32bit each)
    puts "DEBUG: Read register $name access $access"
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr {$offset + $i*4}]
        set readValue($i,value) [reg read $wordOffset]
        
        # save in temporary variables the current register's word (32bit)
        # readValue, orgValue and writtenValue
        set readWord     [expr {$readValue($i,value) & $mask($i,value)}]
        set orgWord      $orgValue($i,value)
        
        # Verify that the read value is as expected
        switch $access {
            
            "RW" {
                set writeMask [write_mask -n $name]
                set writtenWord  [expr {$writtenValue($i,value) & $writeMask}]
               
                # read-write register read value should be as written value
                # Ignore the reseved bits according to register's write_mask
                if {$readWord != $writtenWord} {
                    set res "\n$name \t[format "0\x%x" $wordOffset] \t$access \t$dataWidth \twrite_mask $writeMask  \t ERROR: Testing RW register failed (wrote [format "0x%x" $writtenWord] read [format "0x%x" $readWord])"                    
                    puts $res
                    lappend accessTestResult $res
                }          
            }
            "ST" -
            "RC" {
                set secondRead [reg read $wordOffset]
                # Read-Clear/Statistic (RC,ST) register second read should return all 0
                if  { $secondRead != 0 } {
                    set res "\n$name \t[format "0\x%x" $wordOffset] \t$access \t$dataWidth \t ERROR: Testing RC, ST register failed (2'nd read not zero [format "0x%x" $secondRead])"                                        
                    puts $res
                    lappend accessTestResult $res                    
                }        
            }
            "R" -
            "WB_R" {
                # Nothing to check
            }
            "WB" {
                set writtenWord  $writtenValue($i,value)
               
                # handle as a read-write register 
                # Ignore the reseved bits according to register's write_mask
                if {$readWord != [expr {$writtenWord & [write_mask -n $name]}]} {
                    set res "\n$name \t[format "0\x%x" $wordOffset] \t$access \t$dataWidth \t ERROR: Testing WB register failed (wrote [format "0x%x" $writtenWord] read [format "0x%x" $readWord])"
                    puts $res
                    lappend accessTestResult $res
                }         
            }
            
            default {
                set res "\n$name \t[format "0\x%x" $wordOffset] \t$access \t$dataWidth \t ERROR: Unknown access type"
                puts $res
                lappend accessTestResult $res
            }
        }
        
    }
   
    # Restore the original value of the register's words
    if {($access != "ST") && ($access != "RC")&& ($access != "R") && ($access != "WB_R")} {
        for {set i 0} {$i < $noWords} {incr i} {
            set wordOffset [expr ($offset + $i*4)]        
            reg write  $wordOffset $orgValue($i,value)
        }
    }

    if {$accessTestResult == ""} {
        set accessTestResult "passed"
    } 

    return $accessTestResult
}

# -----------------------------------------------------------------
# This proc tests the reset value of the register
# -----------------------------------------------------------------
proc _regtest::test_reset_value {{name} {offset} {resetValue} {dataWidth} {access}}  {
    
    set resetTestResult "passed"

    # Do not check WR (write-only) registers
    if { $access=="WR" } {
        return 0
    }
    
    # Save the number of words (32 bit) of the register
    set noWords [expr ($dataWidth/32)]
    if { [expr ($dataWidth % 32) != 0] } {
        incr noWords
    }

    # Create the mask according to the register's dataWidth
    for {set i 0} {$i < $noWords} {incr i} {                  
        if {$i < [expr ($noWords-1)]} {
            set mask($i,value) 0xFFFFFFFF
        } else  {
            # The last part of the register's mask is according to remaining
            # data width e.g 100 bit register, mask is for remaining 4 bits (0x1111)
            set mask($i,value) [get_mask $dataWidth]
        }
    }
    
    # Read the register's value
    # Incase of WB go through all its words (32bit each)
    for {set i 0} {$i < $noWords} {incr i} {
        set wordOffset [expr ($offset + $i*4)]
        set readValue($i,value) [reg read $wordOffset]
    
        # Mask the invalid bits
        set readWord   [expr {$readValue($i,value) & $mask($i,value)}]
        set resetValue [expr {$resetValue & $mask($i,value)}]
          
        # Verify that the read value is as expected (resetValue)
        if {$readWord != $resetValue} {
            set resetTestResult "\n$name \t[format "0\x%x" $wordOffset] \t$access \t$dataWidth \t ERROR: Reset value is different than expected (expected [format "0\x%x" $resetValue] read [format "0\x%x" $readWord])"
            puts $resetTestResult
            return $resetTestResult
        }
    }
    return $resetTestResult
}


# -----------------------------------------------------------------
# This procedure uses regtest_main procedure on all blocks
# -----------------------------------------------------------------
proc _regtest::regtest_all {{args}} {
        global blocksTable
        
        
        for {set i 0} {$i < [llength $args]} {incr i} {
            switch -- [lindex $args $i] {
                "-f"         { incr i; set rdfFile [lindex $args $i] }            
            }
        }      
        
        parse_rdf
        
        foreach blockName [lsort [array names blocksTable]] {
                # Skip these blocks from testing
                set skipBlocks { AEU MCP PXP2 PXPCS EMAC0 EMAC1}
                if { [lsearch -exact $skipBlocks $blockName] != -1 } {
                        continue
                }                        
                puts ""
                puts "***********************************************************"
                puts "******          Testing block $blockName                ***"
                puts "***********************************************************"
                log open $blockName.txt
                set testResult [regtest -b $blockName -f $rdfFile]
                log close

                # Update the results log
                log open "reg_res.txt"
                if {$testResult == 0} {
                    puts "$blockName PASSED"
                } else if {$testResult == 1} {
                    puts "$blockName FAILED"
                } else {
                    puts "$blockName N/A"
                }
                log close
        }
}

