###########################################################################
#
# Copyright(c) 2014-2017 Cavium, Inc., all rights reserved
# Proprietary and Confidential Information.
#
# Module Description:
# This script includes impotant utilities such as:
# parse_rdf, reg_read, reg_offset, reg_size, reg_type, reg_info
# read_regs, get_block_registers
##########################################################################


#Blocks table is an array of block name & block offset
global blocksTable

#Blocks to skip
global skipBlocks

#Registeers to skip
global exceptionList

# Regs information in 4 arrays
global list regsName {}
global list regsOffset {}
global list regsSize {}
global list regsCount {}
global list regsType {}
global list regsResetVal {}
global list regsSplitVal {}
global list regsProtectRead {}
global list regsProtectWrite {}
global list regsWriteMask {}
global emacOrgVal
global chip_num
global org_ump_en

# Global register file
global rdfFile

# -----------------------------------------------------------------
# Help info on "parse_rdf" command
# -----------------------------------------------------------------
proc parse_rdf_help {args} {
   set help_info     "Parse the registers file into global arrays e.g. regsOffset\n"
   append help_info  "Usage: parse_rdf -f <RDF file> -b <Block name>\n"
   append help_info  "optional:        -skip <Block to skip> -e < ending block>"
   add_help parse_rdf $help_info
}


#----------------------------------------------------------------------------
# Get RDF file according This utility is used for stopping the STORMs
#----------------------------------------------------------------------------
proc get_rdf_file {} {
    global rdfFile
    global g_chipType
    global g_arrChipTypes
    if {[info exists ::current(RDF_PATH)]} {
        set rdf_path $::current(RDF_PATH)
    } elseif {$::sys(IS_UEFI) && [info exists ::env(HOME)]} {
      set rdf_path "$::env(HOME)\\rdfs"
    } else {
        set rdf_path "./rdfs"
    }

    if {![info exists g_chipType]} {
        get_chip_type
    }

    if {$::sys(IS_UEFI)} {
      if {$g_chipType == $g_arrChipTypes(AH)} {
          set rdfFile "$rdf_path\\k2.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(BB) && [is_A0]} {
          set rdfFile "$rdf_path\\bb_a0.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(BB) && [is_B0]} {
          set rdfFile "$rdf_path\\bb.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(E5)} {
          set rdfFile "$rdf_path\\e5.rdf"
      } else {
          puts "Failed to get RDF file. Unknown device !!"
      }
    } else {
       if {$g_chipType == $g_arrChipTypes(AH)} {
          set rdfFile "$rdf_path/k2.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(BB) && [is_A0]} {
          set rdfFile "$rdf_path/bb_a0.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(BB) && [is_B0]} {
          set rdfFile "$rdf_path/bb.rdf"
      } elseif {$g_chipType == $g_arrChipTypes(E5)} {
          set rdfFile "$rdf_path/e5.rdf"
      } else {
          puts "Failed to get RDF file. Unknown device !!"
      }      
    }
}

proc get_chip_mode {} {
    global g_chipMode
    global g_chipModeAsic
    global g_chipModeEmul
    global g_chipModeFpga
    global g_chipModeName

    #asic/emul/fpga
    set g_chipModeAsic 0
    set g_chipModeEmul 1
    set g_chipModeFpga 2

    #determine chipMode
    set g_chipMode 0
    set g_chipMode [expr $g_chipModeAsic + $g_chipModeEmul*[is_emul] + $g_chipModeFpga*[is_fpga]]
    switch $g_chipMode {
        0       { set g_chipModeName ASIC                 }
        1       { set g_chipModeName Emulation            }
        2       { set g_chipModeName FPGA                 }
        default { error "Chip mode unknown:'$g_chipMode'."}
    }

    return g_chipMode
}



#----------------------------------------------------------------------------
# Get the chip type and set g_chipType
#----------------------------------------------------------------------------
proc get_chip_type {} {
    global g_chipType
    global g_arrChipTypes
    global g_arrChipTypesReverse

    # chip types
    array unset g_arrChipTypes *
    array set g_arrChipTypes [sl {
        AH 0 ;#E4 ArrowHead
        BB 1 ;#E4 BigBear
        E5 2 ;#E5
    }]

    array unset g_arrChipTypesReverse *
    foreach chipType [array names g_arrChipTypes] {
        array set g_arrChipTypesReverse [sl {$g_arrChipTypes($chipType) $chipType}]
    }

    #skip if chipType was already defined
    if {[info exists g_chipType]} { return g_chipType}

    # set chip type
    if {[is_e4_BigBear] == 1} {
        set g_chipType $g_arrChipTypes(BB)
    } elseif {[is_e4_K2] == 1} {
        set g_chipType $g_arrChipTypes(AH)
    } elseif {[is_e5] == 1} {
        set g_chipType $g_arrChipTypes(E5)
    } else {
        error "Chip type is not supported! This script only supports Arrowhead, BigBear and E5. If the type is indeed one of these types, please set the type using one of the flags: \[-ah | -bb | e5\]"
    }

    return g_chipType
}

# -----------------------------------------------------------------
# The proc parses the EVEREST RDF file which contains a list of
# [name, offset, access_type, count, reset_value, split_value, protect_read, protect_write] for each register.
# Saves the registers offset, dataWidth in regsInfo global array.
# -----------------------------------------------------------------
proc parse_rdf {args} {
   global regsName
   global regsOffset
   global regsSize
   global regsCount
   global regsType
   global regsResetVal
   global regsSplitVal
   global regsProtectRead
   global regsProtectWrite
   global regsWriteMask
   global skipBlocks
   global chip_num
   global rdfFile

   set skipBlocks { }
   set testBlock ""
   set doParsing 0
   set clrFlag   0


   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-f"         { incr i; set rdfFile [lindex $args $i] }
         "-skip"      { incr i; lappend skipBlocks [lindex $args $i] }
         "-b"         { incr i; set testBlock [lindex $args $i] }
         "-e"         { incr i; set endBlock [lindex $args $i] }
         "-clr"       { set clrFlag 1 }
         "-help"      { parse_rdf_usage; return -1 }
         default      { parse_rdf_usage; return -1 }
      }
   }

   # If clear flag was set, then clear the regsOffset array
   if { $clrFlag } {
      un_parse_rdf
   } elseif { [info exists regsName] && ([llength $regsName] > 0)} {
         #if RDF file allready parsed, return
         return 0
   }

   if {![info exists rdfFile]} {
      get_rdf_file
   }

   # Verify that the rdf file exists
   if {[file size $rdfFile] == 0} {
      puts "The RDF $rdfFile does not exists"
      return -1
   }

   # split the skipBlocks list into seperate blocks
   set skipBlocks [split $skipBlocks]

   # build the blocks table (block name, block offset)
   # from the rdf file.
   build_blocks_table

   # Open the rdf file for reading
   set infile [open $rdfFile r]

   # Go through each line in the rdf file
   while { [eof $infile] != 1 } {
      set protectR ""
      set protectW ""

      # parse the read line into fields
      set line [gets $infile]
      scan $line %s%s%s%s%s%s%s%s%s%s regName offset accessType dataWidth count resetVal splitVal protectR protectW writeMask

      #skip remark lines and empty lines
      if { [skip_line $regName]} {
         continue
      }

      # Check if line includes only the block name,
      if { [string first ":" $regName] != -1 } {

         # Get the next block name and offset
         set blockName [lindex [split $regName ":"] 0]
         set blockOffset [block_offset -b $blockName]
         if {$blockOffset == -1} {
            puts "ERROR: block $blockName offset was not found !!"
            break
         }

         # If the -b option is set, check if the block name equals the
         # requested block to test.
         if { ($testBlock != "") && ($testBlock == $blockName)} {
            set doParsing 1
         # If the -b option is not set, check if we do not have to skip
         # this block cause it is in the skipBlocks list.
         } elseif { ($testBlock == "") && ([lsearch -exact $skipBlocks $blockName] == -1)} {
            set doParsing 1
         } else {
            set doParsing 0
         }

         # Skip the block name line.
         continue
      }

      # If we need to skip this block, continue to next block
      if {$doParsing == 0} {
         continue
      }

      # Check if the offset field is a range (startOffset - endOffset)
      # If so, parse the start offset and end offset.
      if { ([string first "-" $offset] != -1) } {
         set startOffset [lindex [split $offset "-"] 0]
         set endOffset   [lindex [split $offset "-"] 1]

      # Or several registers (count > 1)
      }  elseif { $count > 1 } {
         # For WB regs, calc the reg reserved width (the upper 2^x)
         if {$dataWidth > 32} {
            set power [expr (log($dataWidth)/log(2))]
            # round up the power
            if {$power != int($power)} {
               set power [expr (int($power))]
               incr power
            }
            set width [expr (int(pow(2,$power)))]
            set noWords [expr ($width/32)]
         } else  {
            set noWords 1
         }
         set startOffset $offset
         set endOffset   [expr ($startOffset + (($count-1) * 4))]
      } else  {
         # Or a single register
         set startOffset $offset
         set endOffset   $startOffset
      }

      # Add to offsets the block's offset
      set startOffset [format "0\x%08x" [incr startOffset $blockOffset]]
      set endOffset   [format "0\x%08x" [incr endOffset $blockOffset]]

      # Save the register's offset and dataWidth in regsInfo array
      if {$endOffset == $startOffset} {
         lappend regsOffset $startOffset
      } else  {
         # incase of a range of offsets
         lappend regsOffset "$startOffset-$endOffset"
      }

      lappend regsName $regName
      lappend regsSize $dataWidth
      lappend regsType $accessType
      lappend regsSplitVal $splitVal
      lappend regsResetVal $resetVal
      lappend regsCount $count
      lappend regsProtectRead $protectR
      lappend regsProtectWrite $protectW
      lappend regsWriteMask $writeMask
   }
   close $infile
   return 0
}

# -----------------------------------------------------------------
# This proc prints out the deletes the lists used by parse_rdf
# -----------------------------------------------------------------
proc un_parse_rdf {}  {
   global regsName
   global regsOffset
   global regsSize
   global regsCount
   global regsType
   global regsResetVal
   global regsSplitVal
   global regsProtectRead
   global regsProtectWrite
   global regsWriteMask

   set regsName {}
   set regsOffset {}
   set regsSize {}
   set regsType {}
   set regsSplitVal {}
   set regsResetVal {}
   set regsCount {}
   set regsProtectRead {}
   set regsProtectWrite {}
   set regsWriteMask {}
}

# -----------------------------------------------------------------
# This proc prints out the parse_rdf proc usage information
# -----------------------------------------------------------------
proc parse_rdf_usage {}  {

   puts stdout "Usage:   parse_rdf -f     <rdf file name>"
   puts stdout "                   -skip  <skip blocks>  "
   puts stdout "                   -b     <test only this block>    "
   puts stdout "                   -e     <end block>    "
   puts stdout "                   -clr                  "

}

# -----------------------------------------------------------------
# This proc builds the blocks array: block name, block offset
# -----------------------------------------------------------------
proc build_blocks_table {}  {

   #use the global array blocksTable
   global blocksTable
   global rdfFile

   get_rdf_file

   # Open the rdf file for reading
   set infile [open $rdfFile r]

   # Go through the rdf file, until the first block name:
   while { [eof $infile] != 1 } {
      # Parse the first line into fields
      set line [gets $infile]
      scan $line "%s%s" blockName blockOffset

      #skip the remark lines and empty lines
      if { [skip_line $blockName] } {
         continue
      }

      # When reaching the first block name: exit
      if {[string first ":" $blockName] != -1} {
         close $infile
         return
      }

      # Save the blockName and blockOffset in the blocksTable array
      set blocksTable($blockName) $blockOffset
   }
   close $infile
}

# -----------------------------------------------------------------
# Help info on "block_offset" command
# -----------------------------------------------------------------
proc block_offset_help {args} {

   set help_info     "Returns the block's offset from BAR by its name\n"
   append help_info  "Usage: block_offset -b <blockname>\n"
   add_help block_offset $help_info
}

# -----------------------------------------------------------------
# This proc returns the block's offset
# -----------------------------------------------------------------
proc block_offset {args}  {
   #use the global array blocksTable
   global blocksTable

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: block_offset -b <blockName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-b"      { incr i; set blockName [lindex $args $i] }
         "-help"   { puts "Usage: block_offset -b <blockName>"; return -1 }
         default   { puts "Usage: block_offset -b <blockName>"; return -1 }
      }
   }

   # Search for the matching name in blocksTable
   foreach name [array names blocksTable] {
      # found the matching block in blocksTable, return its offset
      if {[string compare $name $blockName] == 0} {
         return $blocksTable($name)
      }
   }

   #if we got here, it means we did not find a matching block
   puts "The block $blockName does not exist in blocks table"
   return -1
}

# -----------------------------------------------------------------
# This proc parses the start offset out of an offset field
# (startOffset - endOffset)
# -----------------------------------------------------------------
proc get_start_offset {{offset}}  {

   # Parse out the starting offset
   return  [lindex [split $offset "-"] 0]
}

# -----------------------------------------------------------------
# This proc parses the start offset out of an offset field
# (startOffset - endOffset)
# -----------------------------------------------------------------
proc get_end_offset {{offset} {start}}  {

   # Parse out the ending offset
   set ending  [lindex [split $offset "-"] 1]

   # Since we do not want to spend time testing a whole RAM
   if { $ending > [expr ($start + 0x30)] } {
      set ending [expr ($start +0x30)]
   }
   return $ending
}

# -----------------------------------------------------------------
# This proc checks if we need to skip this line cause its a remark
# (starts with #) or an empty line
# -----------------------------------------------------------------
proc skip_line {regName} {
   if { [string equal -length 1 $regName "#"] == 1 } {
      return 1;
   }
   if { [string length $regName] == 0 } {
      return 1;
   }

   return 0;
}

# -----------------------------------------------------------------
# Help info on "reset_chip" command
# -----------------------------------------------------------------
proc reset_chip_help {args} {
   set help_info     "Reset the chip using the MISC block\n"
   append help_info  "Usage: reset_chip\n"
   add_help reset_chip $help_info
}

# ----------------------------------------------------------------------
# The proc resets all the blocks in the chip using a MISC reset register
# ----------------------------------------------------------------------
proc reset_chip { {noMcpRun 0}} {
   global regsOffset

   if {($::current(DRV_STATE) == "RUNNING") ||
       ($::current(DRV_STATE) == "LINK_UP") ||
       ($::sys(DRV_DBG) == 1)} {
       puts "reset_chip can't be used when driver is loaded !!"
       return
   }
   puts "Reset chip"

   # Parse the appropriate RDF file
   parse_rdf

   # Reset all blocks in MISC_REG_RESET_PL_UA except for: MISC, GRC, RBCN
   #[0] rst_misc_core; [1] rst_grc; [2] rst_rbcn; [3] rst_rbcz;
   set offset [reg_offset -n "MISC_REG_RESET_PL_UA"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0x8

   # Reset all blocks in MISC_REG_RESET_PL_HV except for:
   # PSWHST, PSWRQ, PSWRD, PSWWR, ATC
   #[0] rst_pswhst; [1] rst_pswrq; [2] rst_pswrd; [3] rst_pswwr; [4] rst_atc;
   set offset [reg_offset -n "MISC_REG_RESET_PL_HV"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0x0

   # Reset all blocks in MISC_REG_RESET_REG_PL_PDA_VMAIN_1,2
   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_1"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0xF7FFFFFF

   # Reset all blocks in MISC_REG_RESET_PL_PDA_VMAIN_2
   #[0] rst_rbcf; [1] rst_rbcx; [2] rst_rbcs; [3] rst_mcm; [4] rst_pcm; [5] rst_ycm; [6] rst_msdm;
   #[7] rst_psdm; [8] rst_ysdm;  [9] rst_msem; [10] rst_psem; [11] rst_ysem; [12] rst_xyld; [13] rst_tmld;
   #[14] rst_muld; [15] rst_rbcp; [16] rst_rdif; [17] rst_tdif; [18] rst_rss; [19] rst_cau; [20] rst_ptu;
   #[21] rst_prm; [22] rst_rbcy; [23] rst_rbcq; [24] rst_rbcm; [25] Reserved; [26] rst_rbcv; [27] rst_ypld;
   #[28] rst_ptld; [29] rst_rgfs; [30] rst_tgfs; [31] rst_clk_pb;.
   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_2"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0xFDFFFFFF

   # Reset RGFC, TGFC blocks in E5
   if {[is_e5]} {
        set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_3"]
        set startOffset [lindex [split $offset "-"] 0]
        reg write [expr ($startOffset+8)] 0x3
   }

   # Reset all blocks in MISC_REG_RESET_REG_PL_PDA_VMUX except for: XMAC
   #[0] rst_nig; [1] rst_rbch; [2] rst_nig_hard; [3] rst_dbg; [4] rst_xmac;
   #[5] rst_xmac_soft; [6] rst_mstat_nw;
   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VAUX"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0x38F

   # Reset all blocks in MISCS_REG_RESET_PL_UA except for: GRC, MCP, MISC, DBU, CRBCN
   #[0] rst_grc; [1] rst_mcp_n_reset_reg_hard_core; [2] rst_mcp_n_hard_core_rst_b;
   #[3] rst_mcp_n_reset_cmn_cpu; [4] rst_mcp_n_reset_cmn_core; [5] rst_misc_core;
   #[6] rst_dbue (UART); [7] rst_bmb; [8] rst_ipc; [9] rst_crbcn; [31-10] reserved.
   set offset [reg_offset -n "MISCS_REG_RESET_PL_UA"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0x80

   # Reset all blocks in MISCS_REG_RESET_PL_HV except for: CNIG, PGLC, PXPV, CRBCH, CPMU
   #[0] rst_cnig; [1] rst_pglc; [2] rst_pxpv; [3] rst_crbch; [4] rst_opte;
   #[5] rst_ncsi; [6] rst_umac; [7] rst_mstat; [8] rst_cpmu;
   #[9] rst_rbcb; [10] rst_rbcw; [11] rst_opcs; [12] rst_nws; [13] rst_ms;  [14] rst_led
   #[16] pcie_ctl; [31:17] reserved;
   # NOTE: Do not reset CPMU, since after reset CLOCK SLOWDOWN is enabled, there are bugs
   # in this feature, which cause intermitted interupt lost on second PF.
   # NOTE: Do not reset OPTE block (unless 100G) it causes many attentions.
   # Add reset to NWM, RBCW blocks
   set offset [reg_offset -n "MISCS_REG_RESET_PL_HV"]
   set startOffset [lindex [split $offset "-"] 0]
   reg write [expr ($startOffset+8)] 0x146E0

   # Remove chip from reset
   set offset [reg_offset -n "MISC_REG_RESET_PL_UA"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0xF

   set offset [reg_offset -n "MISC_REG_RESET_PL_HV"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0x1F

   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_1"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0xFFFFFFFF

   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_2"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0xFFFFFFFF

    if {[is_e5]} {
        set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_3"]
        set startOffset [lindex [split $offset "-"] 0]
        reg write [expr ($startOffset+4)] 0x3
   }

   set offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VAUX"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0x38F

   set offset [reg_offset -n "MISCS_REG_RESET_PL_UA"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0x3FF

   # Do not reset OPTE block (unless 100G) it causes many attentions
   # Get NWM, RBCW blocks out of reset
   set offset [reg_offset -n "MISCS_REG_RESET_PL_HV"]
   set startOffset [get_start_offset $offset]
   reg write [expr ($startOffset+4)] 0x1F7EF

   # MCP should be reset after resetting the chip
   if {!$noMcpRun} {
      puts "Reset MCP .."
      mcp reset
   }

   # Delay is required so that blocks will get out of reset
   delay ms 1000
}

# -----------------------------------------------------------------
# Help info on "reg_read" command
# -----------------------------------------------------------------
proc reg_read_help {args} {
   set help_info     "Reads a register's value by its name\n"
   append help_info  "Usage: reg_read -n <regName> \[-d for decimal value\]\n"
   add_help reg_read $help_info
}

# -----------------------------------------------------------------
# This proc returns the register's read value (by its name)
# -----------------------------------------------------------------
proc reg_read {args} {

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_read -n <regName> \[-d for decimal value\]"
      return -1
   }

   set dec_mode 0
    
   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-d"      { set dec_mode 1}
         "-help"   { puts "Usage: reg_read -n <regName> \[-d for decimal value\]"; return -1 }
         default   { puts "Usage: reg_read -n <regName> \[-d for decimal value\]"; return -1 }
      }
   }

   set offset [reg_offset -n $regName]
   if {$offset == -1} {
      return -1;
   }

   if {[string first "-" $offset] != -1} {
        set start_offset [get_start_offset $offset]
   	set end_offset [lindex [split $offset "-"] 1]
   } else {
        set start_offset $offset
        set end_offset $offset
   }
   set val_list {}
   set addr $start_offset 
   for {set i 0} {$addr <= $end_offset} {incr i} {
      set val [reg read $addr]
      if {$dec_mode} {
         set val [expr $val]
      }
      lappend val_list $val
      incr addr 4
   }

   return $val_list
}

# Added by VitalyO
# -----------------------------------------------------------------
# Help info on "read_mem" command
# -----------------------------------------------------------------
proc read_mem_help {args} {    
   set help_info     "Reads a memory by its name\n"
   append help_info  "Usage: read_mem -n <regName> \[-s <size>\] \[-f <output file handler>\]\n"
   append help_info  "if size provided (in regs) reads up to specified size. Otherwise reads entire specified memory content\n"
   add_help read_mem $help_info
}

# -----------------------------------------------------------------
# This proc returns the memory content value (by its name)
# -----------------------------------------------------------------
proc read_mem {args} {
   set usage "Usage: read_mem -n <regName> \[-s <size>\] \[-f <output file handler>\]"

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts $usage
      return -1
   }
   set size 0
   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName	[lindex $args $i] }
         "-s"      { incr i; set size		[lindex $args $i] }
         "-f"      { incr i; set outfile	[lindex $args $i] }
         "-help"   { puts $usage; return -1 }
         default   { puts $usage; return -1 }
      }
   }
    
   #get the register's offset
   set startOffset [format "0x%x" [get_start_offset [reg_offset -n $regName]]]
   if {$size} {
       set endOffset [expr $startOffset + $size - 4]
   } else {
       set endOffset [lindex [split [reg_offset -n $regName] "-"] 1]
   }

   if {![info exists outfile]} {
       set outfile stdout
    }

   if {$startOffset != -1} {
       puts $outfile "Printing memory $regName"
       for {set i $startOffset} {$i <= $endOffset} {set i [format "0x%x" [expr $i + 0x4]]} {
           set val [reg read $i]
           puts $outfile "$i: $val"
       }
   } else {
      return -1
   }
}

# -----------------------------------------------------------------
# Help info on "rd" command
# -----------------------------------------------------------------
proc rd_help {args} {
   set help_info     "\tReads a register's value by its name\n"
   append help_info  "Usage: rd <regName> \[1- for decimal value\]\n"
   add_help rd $help_info
}

# -----------------------------------------------------------------
# returns the register's read value (by its name)
# -----------------------------------------------------------------
proc rd {regName {dec 0}} {
    if ($dec) {        
        return [reg_read -n $regName -d]
    } else {
        return [reg_read -n $regName]
    }
}


#-----------------------------------------------------------------
# This proc writes value to the register (by its name)
# -----------------------------------------------------------------
proc reg_write {args} {
   set usageString "Usage: reg_write -n <regName> -v <value> \[-s\]\n\t\t\t\[-s 0/1\] suppress vfId validation (default 0 => no suppression)" 

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts $usageString
      return -1
   }
  
   set skipVfValidation 0

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-v"      { incr i; set value [lindex $args $i] }
         "-s"    { incr i; set skipVfValidation [lindex $args $i] }
         "-help"   { puts $usageString; return -1 }
         default   { puts $usageString; return -1 }
      }
   }

   #get the register's offset
   set offset [reg_offset -n $regName]
   if {$offset == -1} {
      return -1;
   }

   set start_offset [get_start_offset $offset]

   return [reg write $start_offset $value $skipVfValidation]
}

# -----------------------------------------------------------------
# Help info on "wr" command
# -----------------------------------------------------------------
proc wr_help {args} {
   set help_info     "\tWrite a register's value by its name\n"
   append help_info  "Usage: wr <regName> <value>\n"
   add_help wr $help_info
}

# -----------------------------------------------------------------
# write the specified value to the specified register (by its name)
# -----------------------------------------------------------------
proc wr {reg_name val} {
    reg_write -n $reg_name -v $val
}

# -----------------------------------------------------------------
# Help info on "reg_offset" command
# -----------------------------------------------------------------
proc reg_offset_help {args} {
   set help_info     "Get the register's offset address from BAR by its name\n"
   append help_info  "Usage: reg_offset -n <regName>\n"
   add_help reg_offset $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers offset (by its name)
# -----------------------------------------------------------------
proc reg_offset {args} {
   #use the global array regsOffset
   global regsOffset
   global regsName
   global rdfFile

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_offset -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reg_offset -n <regName>"; return -1 }
         default   { puts "Usage: reg_offset -n <regName>"; return -1 }
      }
   }

   # If parse_rdf wasn't done yet
   if {![info exists regsName]} {
       parse_rdf
   }

   set idx [lsearch -exact $regsName $regName]
   if {$idx < 0} {

      puts "$regName was not found in the RDF file ($rdfFile)"
      return -1
   } else {
      return [lindex $regsOffset $idx]
   }
}

# -----------------------------------------------------------------
# Help info on "reg_size" command
# -----------------------------------------------------------------
proc reg_size_help {args} {
   set help_info     "Get the register's size (in bits) by its name\n"
   append help_info  "Usage: reg_size -n <regName>\n"
   add_help reg_size $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers size (by its name)
# -----------------------------------------------------------------
proc reg_size {args} {
   #use the global array regsSize
   global regsSize
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_size -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reg_size -n <regName>"; return -1 }
         default   { puts "Usage: reg_size -n <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsSize $idx]
}

# -----------------------------------------------------------------
# Help info on "reg_size" command
# -----------------------------------------------------------------
proc reg_count_help {args} {
   set help_info     "Get the register's count by its name (2^count is the total register's array size)\n"
   append help_info  "Usage: reg_count -n <regName>\n"
   add_help reg_count $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers size (by its name)
# -----------------------------------------------------------------
proc reg_count {args} {
   #use the global array regsSize
   global regsCount
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_count -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reg_count -n <regName>"; return -1 }
         default   { puts "Usage: reg_count -n <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsCount $idx]
}

# -----------------------------------------------------------------
# Help info on "reg_type" command
# -----------------------------------------------------------------
proc reg_type_help {args} {
   set help_info     "Get the register's access type (RW/R/ST/WB) by its name\n"
   append help_info  "Usage: reg_type -n <regName>\n"
   append help_info  "RW - read write register\n"
   append help_info  "R  - read only register\n"
   append help_info  "ST - statistics register, clears on second read\n"
   append help_info  "RC - read clears register, clears on second readn"
   append help_info  "WB - wide bus register (> 32 bit)\n"
   add_help reg_type $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers type (by its name)
# -----------------------------------------------------------------
proc reg_type {args} {
   #use the global array regsType
   global regsType
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_type -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reg_type <regName>"; return -1 }
         default   { puts "Usage: reg_type <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsType $idx]
}

# -----------------------------------------------------------------
# Help info on "reset_value" command
# -----------------------------------------------------------------
proc reset_value_help {args} {
   set help_info     "Get the register's reset value by its name\n"
   append help_info  "Usage: reset_value -n <regName>\n"
   add_help reset_value $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers reset value (by its name)
# -----------------------------------------------------------------
proc reset_value {args} {
   global regsResetVal
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reset_value -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reset_value -n <regName>"; return -1 }
         default   { puts "Usage: reset_value -n <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsResetVal $idx]
}

# -----------------------------------------------------------------
# Help info on "split_value" command
# -----------------------------------------------------------------
proc split_value_help {args} {
   set help_info     "Get the register's split value by its name\n"
   append help_info  "Usage: split_value -n <regName>\n"
   add_help split_value $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers split value (by its name)
# -----------------------------------------------------------------
proc split_value {args} {

   global regsSplitVal
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: split_value -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: split_value <regName>"; return -1 }
         default   { puts "Usage: split_value <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsSplitVal $idx]
}

# -----------------------------------------------------------------
# Help info on "protect_value" command
# -----------------------------------------------------------------
proc protect_value_help {args} {
   set help_info     "Get the register's protect value by its name\n"
   append help_info  "Usage: protect_value <read/write> -n <regName>\n"
   add_help protect_value $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers protect value (by its name)
# -----------------------------------------------------------------
proc protect_value {args} {
   global regsProtectRead
   global regsProtectWrite
   global regsName

   set usageStr "Usage: protect_value <read/write> -n <regName>"

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts $usageStr
      return -1
   }

   set rw [lindex $args 0]

   # Parse the options.
   for {set i 1} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts $usageStr; return -1 }
         default   { puts $usageStr; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   if {$rw == "read"} {
      return [lindex $regsProtectRead $idx]
   } elseif {$rw == "write"} {
      return [lindex $regsProtectWrite $idx]
   } else {
      puts $usageStr
      return -1
   }
}

# -----------------------------------------------------------------
# Help info on "write_mask" command
# -----------------------------------------------------------------
proc write_mask_help {args} {
   set help_info     "Get the register's write mask by its name\n"
   append help_info  "Usage: write_mask -n <regName>\n"
   add_help write_mask $help_info
}

# -----------------------------------------------------------------
# This proc returns the registers write mask (by its name)
# -----------------------------------------------------------------
proc write_mask {args} {

   global regsWriteMask
   global regsName

   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: write_mask -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: write_mask <regName>"; return -1 }
         default   { puts "Usage: write_mask <regName>"; return -1 }
      }
   }

   set idx [lsearch -exact $regsName $regName]
   return [lindex $regsWriteMask $idx]
}


# -----------------------------------------------------------------
# Help info on "reg_info" command
# -----------------------------------------------------------------
proc reg_info_help {args} {
   set help_info     "Get the register's information (offset, size, type) by its name\n"
   append help_info  "Usage: reg_info -n <regName>\n"
   add_help reg_info $help_info
}

# -----------------------------------------------------------------
# This proc returns all the registers information (by its name)
# -----------------------------------------------------------------
proc reg_info {args} {
   # if no arguments, print the usage
   if {[llength $args] == 0} {
      puts "Usage: reg_info -n <regName>"
      return -1
   }

   # Parse the options.
   for {set i 0} {$i < [llength $args]} {incr i} {
      switch -- [lindex $args $i] {
         "-n"      { incr i; set regName [lindex $args $i] }
         "-help"   { puts "Usage: reg_info -n <regName>"; return -1 }
         default   { puts "Usage: reg_info -n <regName>"; return -1 }
      }
   }

   puts "offset: [reg_offset -n $regName] size: [reg_size -n $regName] type: [reg_type -n $regName]"
}



# -----------------------------------------------------------------
# Help info on "read_regs" command
# -----------------------------------------------------------------
proc read_regs_help {args} {
   set help_info     "Read all the registers values and print them into a file\n"
   append help_info  "Usage: read_regs -out <outfile>"
   append help_info  "optional:        -all -b <blockName> -noPrint\n\n"
   append help_info  "Notes: \n"
   append help_info  "-all reads all registers\n"
   append help_info  "-b <blockname> reads only this block registers\n"
   append help_info  "-noPrint does not print any extra print-outs\n"
   add_help read_regs $help_info
}


# -----------------------------------------------------------------
# This proc checks if block is in reset
# -----------------------------------------------------------------
proc is_block_in_reset {blockName} {
   # Read MISC/MISCS Reset registers
   set miscUa [reg read [get_start_offset [reg_offset -n "MISC_REG_RESET_PL_UA"]]]
   set miscHv [reg read [get_start_offset [reg_offset -n "MISC_REG_RESET_PL_HV"]]]
   set miscPdaVmain1 [reg read [get_start_offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_1"]]]
   set miscPdaVmain2 [reg read [get_start_offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VMAIN_2"]]]
   set miscPdaVaux [reg read [get_start_offset [reg_offset -n "MISC_REG_RESET_PL_PDA_VAUX"]]]
   set miscsUa [reg read [get_start_offset [reg_offset -n "MISCS_REG_RESET_PL_UA"]]]
   set miscsHv [reg read [get_start_offset [reg_offset -n "MISCS_REG_RESET_PL_HV"]]]

   switch $blockName {
      "RBCN"        { if {[expr ($miscUa & 0x4)] != 0x4} {return 1} }
      "RBCZ"        { if {[expr ($miscUa & 0x8)] != 0x8} {return 1} }
      "PSWHST"        { if {[expr ($miscHv & 0x1)] != 0x1} {return 1} }
      "PSWRQ"        { if {[expr ($miscHv & 0x2)] != 0x2} {return 1} }
      "PSWRD"        { if {[expr ($miscHv & 0x4)] != 0x4} {return 1} }
      "PSWWR"        { if {[expr ($miscHv & 0x8)] != 0x8} {return 1} }
      "ATC"        { if {[expr ($miscHv & 0x10)] != 0x10} {return 1} }
      "BRB"        { if {[expr ($miscPdaVmain1 & 0x1)] != 0x1} {return 1} }
      "PRS"        { if {[expr ($miscPdaVmain1 & 0x2)] != 0x2} {return 1} }
      "SRC"        { if {[expr ($miscPdaVmain1 & 0x4)] != 0x4} {return 1} }
      "TSDM"        { if {[expr ($miscPdaVmain1 & 0x8)] != 0x8} {return 1} }
      "TSEM"        { if {[expr ($miscPdaVmain1 & 0x10)] != 0x10} {return 1} }
      "TCM"        { if {[expr ($miscPdaVmain1 & 0x20)] != 0x20} {return 1} }
      "RBCR"        { if {[expr ($miscPdaVmain1 & 0x40)] != 0x40} {return 1} }
      "USDM"        { if {[expr ($miscPdaVmain1 & 0x80)] != 0x80} {return 1} }
      "UCM"        { if {[expr ($miscPdaVmain1 & 0x100)] != 0x100} {return 1} }
      "USEM"        { if {[expr ($miscPdaVmain1 & 0x200)] != 0x200} {return 1} }
      "BTB"        { if {[expr ($miscPdaVmain1 & 0x400)] != 0x400} {return 1} }
      "PBF_PB1"        { if {[expr ($miscPdaVmain1 & 0x800)] != 0x800} {return 1} }
      "PBF_PB2"        { if {[expr ($miscPdaVmain1 & 0x1000)] != 0x1000} {return 1} }
      "RPB"        { if {[expr ($miscPdaVmain1 & 0x2000)] != 0x2000} {return 1} }
      "RBCU"        { if {[expr ($miscPdaVmain1 & 0x4000)] != 0x4000} {return 1} }
      "PBF"        { if {[expr ($miscPdaVmain1 & 0x8000)] != 0x8000} {return 1} }
      "QM"        { if {[expr ($miscPdaVmain1 & 0x10000)] != 0x10000} {return 1} }
      "TM"        { if {[expr ($miscPdaVmain1 & 0x20000)] != 0x20000} {return 1} }
      "DORQ"        { if {[expr ($miscPdaVmain1 & 0x40000)] != 0x40000} {return 1} }
      "XCM"        { if {[expr ($miscPdaVmain1 & 0x80000)] != 0x80000} {return 1} }
      "XSDM"        { if {[expr ($miscPdaVmain1 & 0x100000)] != 0x100000} {return 1} }
      "XSEM"        { if {[expr ($miscPdaVmain1 & 0x200000)] != 0x200000} {return 1} }
      "RBCT"        { if {[expr ($miscPdaVmain1 & 0x400000)] != 0x400000} {return 1} }
      "CDU"        { if {[expr ($miscPdaVmain1 & 0x800000)] != 0x800000} {return 1} }
      "CCFC"        { if {[expr ($miscPdaVmain1 & 0x1000000)] != 0x1000000} {return 1} }
      "TCFC"        { if {[expr ($miscPdaVmain1 & 0x2000000)] != 0x2000000} {return 1} }
      "RBCP"        { if {[expr ($miscPdaVmain1 & 0x4000000)] != 0x4000000} {return 1} }
      "IGU"        { if {[expr ($miscPdaVmain1 & 0x8000000)] != 0x8000000} {return 1} }
      "DMAE"        { if {[expr ($miscPdaVmain1 & 0x10000000)] != 0x10000000} {return 1} }
      "RBCF"        { if {[expr ($miscPdaVmain2 & 0x1)] != 0x1} {return 1} }
      "RBCX"        { if {[expr ($miscPdaVmain2 & 0x2)] != 0x2} {return 1} }
      "RBCS"        { if {[expr ($miscPdaVmain2 & 0x4)] != 0x4} {return 1} }
      "MCM"        { if {[expr ($miscPdaVmain2 & 0x8)] != 0x8} {return 1} }
      "PCM"        { if {[expr ($miscPdaVmain2 & 0x10)] != 0x10} {return 1} }
      "YCM"        { if {[expr ($miscPdaVmain2 & 0x20)] != 0x20} {return 1} }
      "MSDM"        { if {[expr ($miscPdaVmain2 & 0x40)] != 0x40} {return 1} }
      "YSDM"        { if {[expr ($miscPdaVmain2 & 0x80)] != 0x80} {return 1} }
      "PSDM"        { if {[expr ($miscPdaVmain2 & 0x100)] != 0x100} {return 1} }
      "MSEM"        { if {[expr ($miscPdaVmain2 & 0x200)] != 0x200} {return 1} }
      "YSEM"        { if {[expr ($miscPdaVmain2 & 0x400)] != 0x400} {return 1} }
      "PSEM"        { if {[expr ($miscPdaVmain2 & 0x800)] != 0x800} {return 1} }
      "XYLD"        { if {[expr ($miscPdaVmain2 & 0x1000)] != 0x1000} {return 1} }
      "TMLD"        { if {[expr ($miscPdaVmain2 & 0x2000)] != 0x2000} {return 1} }
      "MULD"        { if {[expr ($miscPdaVmain2 & 0x4000)] != 0x4000} {return 1} }
      "YULD"        { if {[expr ($miscPdaVmain2 & 0x8000)] != 0x8000} {return 1} }
      "RDIF"        { if {[expr ($miscPdaVmain2 & 0x10000)] != 0x10000} {return 1} }
      "TDIF"        { if {[expr ($miscPdaVmain2 & 0x20000)] != 0x20000} {return 1} }
      "RSS"        { if {[expr ($miscPdaVmain2 & 0x40000)] != 0x40000} {return 1} }
      "CAU"        { if {[expr ($miscPdaVmain2 & 0x80000)] != 0x80000} {return 1} }
      "PTU"        { if {[expr ($miscPdaVmain2 & 0x100000)] != 0x100000} {return 1} }
      "PRM"        { if {[expr ($miscPdaVmain2 & 0x200000)] != 0x200000} {return 1} }
      "YPLD"        { if {[expr ($miscPdaVmain2 & 0x8000000)] != 0x8000000} {return 1} }
      "PTLD"        { if {[expr ($miscPdaVmain2 & 0x10000000)] != 0x10000000} {return 1} }
      "RGFS"        { if {[expr ($miscPdaVmain2 & 0x20000000)] != 0x20000000} {return 1} }
      "TGFS"        { if {[expr ($miscPdaVmain2 & 0x40000000)] != 0x40000000} {return 1} }
      "NIG"        { if {[expr ($miscPdaVaux & 0x1)] != 0x1} {return 1} }
      "RBCH"        { if {[expr ($miscPdaVaux & 0x2)] != 0x2} {return 1} }
      "DBG"        { if {[expr ($miscPdaVaux & 0x8)] != 0x8} {return 1} }
      "XMAC"        { if {[expr ($miscPdaVaux & 0x10)] != 0x10} {return 1} }
      "DBUE"        { if {[expr ($miscsUa & 0x40)] != 0x40} {return 1} }
      "BMB"        { if {[expr ($miscsUa & 0x80)] != 0x80} {return 1} }
      "IPC"        { if {[expr ($miscsUa & 0x100)] != 0x100} {return 1} }
      "CRBCN"        { if {[expr ($miscsUa & 0x200)] != 0x200} {return 1} }
      "CNIG"        { if {[expr ($miscsHv & 0x1)] != 0x1} {return 1} }
      "PGLC"        { if {[expr ($miscsHv & 0x2)] != 0x2} {return 1} }
      "PXPV"        { if {[expr ($miscsHv & 0x4)] != 0x4} {return 1} }
      "CRBCH"        { if {[expr ($miscsHv & 0x8)] != 0x8} {return 1} }
      "OPTE"        { if {[expr ($miscsHv & 0x10)] != 0x10} {return 1} }
      "NCSI"        { if {[expr ($miscsHv & 0x20)] != 0x20} {return 1} }
      "UMAC"        { if {[expr ($miscsHv & 0x40)] != 0x40} {return 1} }
      "MSTAT"        { if {[expr ($miscsHv & 0x80)] != 0x80} {return 1} }
      "CPMU"        { if {[expr ($miscsHv & 0x100)] != 0x100} {return 1} }
   }

   return 0
}

# -----------------------------------------------------------------
# This proc reads all the registers of the RDF file and prints
# the register's name, offset, size, type and read value in an output file.
# The outfput file name is the same as the RDF file with the extention
# ".out" instead of ".rdf"
# -----------------------------------------------------------------
proc read_regs {args} {

    set outfileName ""
    set testBlock ""
    set noPrint 0
    set did 0
    set splitRegs 0

    # use globals
    global regsOffset
    global regsName

    set exceptionReadList [list MCP_REG_UCINT_AVS_DATA]


    # if no arguments, print the usage
    if {[llength $args] == 0} {
        read_regs_usage
        return -1
    }



    # Parse the options.
    for {set i 0} {$i < [llength $args]} {incr i} {
        switch -- [lindex $args $i] {
                "-b"         { incr i; set testBlock [lindex $args $i]}
                "-out"       { incr i; set outfileName [lindex $args $i]}
                "-silent"    { set noPrint 1}
                "-split_regs" { set splitRegs 1}
                "-help"      { read_regs_usage; return -1 }
                default      { read_regs_usage; return -1 }
         }
    }


    # Open the output file
    if {$outfileName != ""} {
        set outfile [open "$outfileName" a]
    }

    # Parse the appropriate RDF file
    parse_rdf

    # If block in reset print error and return 1
    if { $testBlock != "" } {
        if {[is_block_in_reset $testBlock] == 1} {
            puts "ERROR: Can not read from block $testBlock. Block is in reset"
            # Close the output file
            if {$outfileName != ""} {
                close $outfile
            }
            return 1
        }
    }

    # Go through each register in the regsOffset array
    for {set idx 0} {$idx < [llength $regsName]} {incr idx} {

        set name [lindex $regsName $idx]

        # Set the "testBlock_REG" temporary string
        set tmpStr $testBlock
        append tmpStr "_REG*"


        # If the -b option is set, skip the register if its not in the tested block
        if { ($testBlock != "") && ![string match $tmpStr $name]} {
            continue
        }

        # skip registers that are in the exceptionReadList
        if { [lsearch -exact $exceptionReadList $name] != -1 } {
            puts "Skipping $name"
            continue;
        }

     # skip special registers
        if { [ string match "*UNUSED*" $name ]} {
            continue;
        }
        if { [ string match "*INT_STS*" $name ]} {
            continue;
        }
        if { [ string match "*INT_STS_CLR*" $name ]} {
            continue;
        }
        if { [ string match "*INT_STS_WR*" $name ]} {
            continue;
        }
        if { [ string match "*PRTY_STS*" $name ]} {
            continue;
        }
        if { [ string match "*PRTY_MASK*" $name ]} {
            continue;
        }
        if { [ string match "*INT_MASK*" $name ]} {
            continue;
        }

        set offset [reg_offset -n $name]
        # if offset includes a range, read only the startOffset
        if { ([string first "-" $offset] != -1) } {
            set offset [lindex [split $offset "-"] 0]
        }

        # Get the register's information (size, type)
        set size [reg_size -n $name]
        set type [reg_type -n $name]
        set dataWidth  [reg_size -n $name]
        set splitVal [split_value -n $name]

        # Incase user requests to show only split registers
        # skip this register if it's splitVal is 0
        if {$splitRegs && ($splitVal==0)} {
            continue
        }

        # For WB regs, calc the reg reserved width (the upper 2^x)
        if {$dataWidth > 32} {
            set power [expr (log($dataWidth)/log(2))]
            #round up the power value
            if { $power != [expr (int($power))]} {
                set power [expr (int($power))]
                incr power
            }
            set width [expr (int(pow(2,$power)))]
            set noWords [expr ($width/32)]
        } else  {
            set noWords 1
        }

        # Incase of WB go through all its words (32bit each)
        for {set i 0} {$i < $noWords} {incr i} {
            set wordOffset [expr ($offset + $i*4)]
            set value [reg read $wordOffset]

            #add delay between words in WB register
            if {$noWords > 1} {after 100}

            # print the registers name, information and read value to output file
            if {$outfileName != ""} {
                puts $outfile "$name \t[format "0x%x" $wordOffset] \t$size \t$type \t$value"
            }

            # print information to screen also if required
            if {$noPrint != 1} {
                puts "$name \t[format "0x%x" $wordOffset] \t$size \t$type \t$value"
            }
        }
    }

    if {$outfileName != ""} {
        close $outfile
        return 0
    }
}

# -----------------------------------------------------------------
# This proc prints out the read_regs proc usage information
# -----------------------------------------------------------------
proc read_regs_usage {}  {

    puts stdout "Usage:   read_regs "
    puts stdout "                   -b <blockName>      :read only this block registers"
    puts stdout "                   -silent             :dont print to screen"
    puts stdout "                   -out <filename>     :output to this file"

}

# -----------------------------------------------------------------
# This procedure performs read_regs on all blocks
# -----------------------------------------------------------------
proc read_regs_all {{args}} {
        global blocksTable
        global regsOffset

        set silent 0
        set did 0
        set outfileName ""
        set splitRegs 0

        if {[info exists ::sys(NO_PCI)]} {
            set read_via_no_pci $::sys(NO_PCI)
        } else {
            set read_via_no_pci [regexp {NO_PCI} $::tcl_prompt1]
        }

        # Parse the options.
        for {set i 0} {$i < [llength $args]} {incr i} {
            switch -- [lindex $args $i] {
                "-out"       { incr i; set outfileName [lindex $args $i]}
                "-silent"    { set silent 1}
                "-split_regs"    { set splitRegs 1}
                "-help"      { read_regs_all_usage; return -1 }
                default      { read_regs_all_usage; return -1 }
            }
        }

        # Parse the appropriate RDF file
        parse_rdf

        foreach blockName [lsort [array names blocksTable]] {
                # Skip these blocks from testing
                set skipBlocks { }

                if { [lsearch -exact $skipBlocks $blockName] != -1 } {
                        continue
                }
                if {$silent} {
                     if {$splitRegs} {
                        read_regs -b $blockName -out $outfileName -silent -split_regs
                     } else {
                        read_regs -b $blockName -out $outfileName -silent
                     }
                } else {
                     if {$splitRegs} {
                        read_regs -b $blockName -out $outfileName -split_regs
                     } else {
                        read_regs -b $blockName -out $outfileName
                     }
                }

        }
}

# -----------------------------------------------------------------
# This proc prints out the read_regs proc usage information
# -----------------------------------------------------------------
proc read_regs_all_usage {}  {

    puts stdout "Usage:   read_regs_all                 :read all registers"
    puts stdout "                   -silent             :dont print to screen"
    puts stdout "                   -split_regs         :read only split registers"
    puts stdout "                   -out <filename>     :output to this file"

}

# -----------------------------------------------------------------
# Checks if running on FPGA or emulation
# -----------------------------------------------------------------
proc is_sim {} {
   set isSim 0

   #Read register MISCS_REG_CHIP_REV 0x9770 by offset and not by name
   #Since we don't know yet which RDF to parse
   set chipRev [expr {[reg read 0x9770] & 0xf}]

   if {$chipRev > 5} {
      set isSim 1
   }

   return $isSim
}

# -----------------------------------------------------------------
# Checks if running on FPGA
# -----------------------------------------------------------------
proc is_fpga {} {
   set isFpga 0

   # Parse the appropriate RDF file
   parse_rdf

   set chipRev [reg_read -n "MISCS_REG_CHIP_REV"]

   if {[is_sim] && (($chipRev & 0x1) == 0x1)} {
      set isFpga 1
   }

   return $isFpga
}

# -----------------------------------------------------------------
# Checks if running on Emulation
# -----------------------------------------------------------------
proc is_emul {} {
   set isEmul 0

   # Parse the appropriate RDF file
   parse_rdf

   set chipRev [reg_read -n "MISCS_REG_CHIP_REV"]

   # if simulation and bit 0 is 0
   if {([is_sim] && (($chipRev & 0x1) == 0x0))} {
      set isEmul 1
   }

   return $isEmul
}



# -----------------------------------------------------------------
# Checks if running on FPGA
# -----------------------------------------------------------------
proc chip_rev {} {
   global regsOffset
   # Parse the appropriate RDF file
   parse_rdf

   set chipRev [expr {[reg_read -n "MISCS_REG_CHIP_REV"] & 0xf}]

   return $chipRev
}

# -----------------------------------------------------------------
# Returns the chip revision Id (0=A0, 1=B0, ...)
# -----------------------------------------------------------------
proc chip_rev_id {} {
   #Read register MISCS_REG_CHIP_REV 0x9770 by offset and not by name
   #Since we don't know yet which RDF to parse
   set rev_reg_val [reg read 0x9770]
   set rev_id 0

   if {![is_sim]} {
      set rev_id [expr {$rev_reg_val & 0xf}]
   } else {
      set rev_id [expr {7 - (($rev_reg_val >> 1) & 0xf)}]
   }

   return $rev_id
}

# -----------------------------------------------------------------
# Checks if chip revision is A0
# -----------------------------------------------------------------
proc is_A0 {} {
   set isA0 0

   if {[chip_rev_id] == 0} {
      set isA0 1
   }

   return $isA0
}

# -----------------------------------------------------------------
# Checks if chip revision is B0
# -----------------------------------------------------------------
proc is_B0 {} {
   set isB0 0

   if {[chip_rev_id] == 1} {
      set isB0 1
   }

   return $isB0
}

# -----------------------------------------------------------------
# This proc gets the mask (number of 1's) according to the reg dataWidth
# e.g. if dataWidth=2 then mask is 0x11, is dataWidth=8 then mask is 0x11111111
# -----------------------------------------------------------------
proc get_mask {{dataWidth}}  {
   set mask 1
   set multiple 1

   # incase of a register > 32 bit, mask if of the remainder
   if {$dataWidth > 32} {
      set size [expr ($dataWidth % 32)]
   } else  {
      set size $dataWidth
   }

   # If register size is a multiple of 32 (%32 equals 0),
   # then last word mask is 0xFFFFFFFF
   if {$size == 0} {
      return 0xFFFFFFFF
   }

   for {set i 0} {$i < [expr ($size - 1)]} {incr i} {
      set multiple [expr $multiple * 2]
      set mask [expr ($mask + $multiple)]
   }
   set mask [format "0\x%08x" $mask]
   return $mask
}


# -----------------------------------------------------------------
# This proc checks if we need to test this register
# returns 1 - incase we need to test this register
# returns 0 - incase there is no need to test this register
# -----------------------------------------------------------------
proc check_split_reg {{skipCommonReg} {splitVal} {curFunc}} {


    switch $splitVal {
            "1" {
               # Incase of common register (splitVal==1), run test only on Function 0
               # Testing common register on other functions, if skipCommonReg
               # flag is not set.
               if {($curFunc==0) || !$skipCommonReg } {
                  return 1
               } else {
                  return 0
               }
            }

            default {
               #In other cases, check register on each function
               return 1
            }
    }
    #puts "DEBUG: Skip register from testing (splitVal $splitVal port $port curFunc $curFunc)"
    return 0
}

#----------------------------------------------------------------------------
# Check if we are in E2 4 port mode
#----------------------------------------------------------------------------
proc is_4port_mode {} {
   parse_rdf

   # ovwr_flag is determined by MISC_REG_PORT4MODE_EN_OVWR bit 0
   set ovwr_flag [reg_read -n MISC_REG_PORT4MODE_EN_OVWR]

   if {$ovwr_flag == 0} {
      # if overwrite flag is not set, read MISC_REG_PORT4MODE_EN
      if {[reg_read -n MISC_REG_PORT4MODE_EN] == 0x1} {
         return 1
      } else {
         return 0
      }
   } else {
      # if overwrite flag is set, read MISC_REG_PORT4MODE_EN_OVWR bit 1
      return [expr {([reg_read -n MISC_REG_PORT4MODE_EN_OVWR] >> 1) & 1}]
   }
}

#----------------------------------------------------------------------------
# Check if chip is E4
#----------------------------------------------------------------------------
proc is_e4 {} {
   if {[is_e4_K2] || [is_e4_BigBear]} {
      return 1
   } else {
      return 0
   }
}

#----------------------------------------------------------------------------
# Check if chip is E4 K2
#----------------------------------------------------------------------------
proc is_e4_K2 {} {

   if {($::sys(NO_PCI) == 0) && [info exist ::current(CHIP_NUM)]} {
      if {$::current(CHIP_NUM) == "57940"} {
         return 1
      } else {
         return 0
      }
   } else {
      #incase of NO_PCI mode, read MISCS_REG_CHIP_NUM
      set chip_num [reg read 0x976c]
      if {($chip_num == 0x8070) || ($chip_num == 0x8071) ||
          ($chip_num == 0x8072) || ($chip_num == 0x8073) ||
          ($chip_num == 0x8080) || ($chip_num == 0x8081 ) ||
          ($chip_num == 0x8082) || ($chip_num == 0x8083 ) ||
          ($chip_num == 0x8084) || ($chip_num == 0x8085 ) ||
          ($chip_num == 0x8086) || ($chip_num == 0x8087 ) ||
          ($chip_num == 0x8088) || ($chip_num == 0x808c) ||
          ($chip_num == 0x8090) || ($chip_num == 0x8004) ||
          ($chip_num == 0x8006) || ($chip_num == 0x8016) ||
          ($chip_num == 0x8005) || ($chip_num == 0x800A) ||
          ($chip_num == 0x801A) || ($chip_num == 0x8009) ||
          ($chip_num == 0x80F0) || ($chip_num == 0x809A) ||
          ($chip_num == 0x808a) || ($chip_num == 0x80D0)} {
         return 1
      } else {
         return 0
      }
   }
}

#----------------------------------------------------------------------------
# Check if chip is E4 Big Bear
#----------------------------------------------------------------------------
proc is_e4_BigBear {} {

   if {($::sys(NO_PCI) == 0) && [info exist ::current(CHIP_NUM)]} {
      if {$::current(CHIP_NUM) == "57980"} {
         return 1
      } else {
         return 0
      }
   } else {
      #incase of NO_PCI mode, read MISCS_REG_CHIP_NUM
      set chip_num [reg read 0x976c]
      if {($chip_num == 0x1634) || ($chip_num == 0x1635) ||
          ($chip_num == 0x1644) || ($chip_num == 0x1654) ||
          ($chip_num == 0x1656) || ($chip_num == 0x1664) ||
          ($chip_num == 0x165c) || ($chip_num == 0x165e) ||
          ($chip_num == 0x165f) || ($chip_num == 0x1666) ||
          ($chip_num == 0x1657) || ($chip_num == 0x1629) ||
          ($chip_num == 0x163f) || ($chip_num == 0x163d) ||
          ($chip_num == 0x163e) || ($chip_num == 0x163c)} {
         return 1
      } else {
         return 0
      }
   }
}

#----------------------------------------------------------------------------
# Check if chip is E5
#----------------------------------------------------------------------------
proc is_e5 {} {

    if {($::sys(NO_PCI) == 0) && [info exist ::current(CHIP_NUM)]} {
        if {$::current(CHIP_NUM) == "55000" || $::current(CHIP_NUM) == "CN5500"} {
            return 1
        } else {
            return 0
        }
    } else {
        #incase of NO_PCI mode, read MISCS_REG_CHIP_NUM
        set chip_num [reg read 0x976c]
        set e5_chip_nums [sl {
            0x8101
            0x8102
            0x8104
            0x8108
            0x810C
            0x8170
            0x8180
            0x8184
            0x818C
            0x8190
        }]
        #return 1 if chip_num is part of array, 0 if not
        return [expr [lsearch -integer $e5_chip_nums $chip_num] != -1]
   }
}


#----------------------------------------------------------------------------
# Set XMAC internal loopback
#----------------------------------------------------------------------------
proc set_xmac_lb {} {
   parse_rdf

   # Set XMAC loopback
   reg_write -n XMAC_REG_CTRL -v 0xB
}

# -----------------------------------------------------------------
# Help info on "set_xmac_lb" command
# -----------------------------------------------------------------
proc set_xmac_lb_help {args} {
   set help_info     "Set internal XMAC loopback\n"
   append help_info  "Usage: set_xmac_lb\n"
   add_help set_xmac_lb $help_info
}

#----------------------------------------------------------------------------
# Recover board via serial
#----------------------------------------------------------------------------
proc recover {} {
   reg write 0x3130 0x0800D0B6
   reg write 0x3134 0x800045C0
   reg write 0x3130 0x0800D124
   reg write 0x3134 0x8000280D
   reg write 0x3134 0x8000A80D
}

#----------------------------------------------------------------------------
# Recover board via serial
#----------------------------------------------------------------------------
proc pcie_recover {} {
      for {set i 0} {$i < 100000} {incr i} {
         set val [reg read 0x2000]
         if {($val == 0x163014e4) || ($val == 0x163414e4)} {
            puts "device detected i=$i"
            break
         }
      }

      for {set i 0} {$i < 10000} {incr i} {
         reg write 0x24dc 0x00315F02
      }
}


#----------------------------------------------------------------------------
# OTP programming via serial
#----------------------------------------------------------------------------
proc read_otp {addr} {
     set otp_ctrl [expr (0<<1) | ($addr << 6) | (1<<18)]
    #puts "otp_ctrl [format \"0x%x\" $otp_ctrl]"
    reg write 0x20204 [expr $otp_ctrl]
    reg write 0x20204 [expr $otp_ctrl | 1]
    after 10
    #puts "[reg read 0x20208]"
    set val [reg read 0x20210]
    return $val
}
proc dump_otp {} {
 for {set row 0 } {$row < 64} {incr row} {
   set val [read_otp $row]
   puts "Row $row (bits [expr $row*32]-[expr ($row+1)*32]): $val"
 }
}

proc otp {{device_id 0x4}} {
   # Read device ID
   set val [reg read 0x2000]

   # Wrong reading from serial port.
   # workaround: close the port and re-open it
   if {$val == 0x63414e4} {
      serport close
      serport open 0 115200
      serport redirect duart
   }

   # Program Enable
   reg write 0x2020c 0xf
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0x4
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0x8
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0xd
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   set val [reg read 0x20208]
   if {$val != 0x200e} {
      puts "Read val from 0x20208 is $val and not as expected 0x200e"
      return
   }

   # Program Word
   reg write 0x2020c $device_id
   reg write 0x20204 0x40214
   reg write 0x20204 0x40215
   after 10

   # Reload FOUT
   reg write 0x20204 0x4001a
   reg write 0x20204 0x4001b
   after 10

   # Read device ID
   set val [reg read 0x976c]
   if {$val == [expr 0x1630 | $device_id]} {
      puts "OTP programming completed successfully."
   } else {
      puts "OTP programing failed. Device ID $val not as expected 0x163414e4."
   }
}

proc write_otp {addr value} {
   # Read device ID
   set val [reg read 0x2000]

   # Wrong reading from serial port.
   # workaround: close the port and re-open it
   if {$val == 0x63414e4} {
      serport close
      serport open 0 115200
      serport redirect duart
   }

   # Program Enable
   reg write 0x2020c 0xf
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0x4
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0x8
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   reg write 0x2020c 0xd
   reg write 0x20204 0x40004
   reg write 0x20204 0x40005
   after 10

   set val [reg read 0x20208]
   if {$val != 0x200e} {
      puts "Read val from 0x20208 is $val and not as expected 0x200e"
      return
   }

   # Program Word
   reg write 0x2020c $value
   reg write 0x20204 [expr 0x40014 | ($addr << 6)]
   reg write 0x20204 [expr 0x40015 | ($addr << 6)]
   after 10

   # Reload FOUT
   reg write 0x20204 0x4001a
   reg write 0x20204 0x4001b
   after 10
}


proc readRamOffset {stormCapLetter {offset 0}} {

    append stormName $stormCapLetter stormRam
    append fastMemRef $stormCapLetter SEM_REG_FAST_MEMORY_INT_RAM

    set ramStartOffset [lindex [split [reg_offset -n $fastMemRef] -] 0]
    set fullOffset [expr ($ramStartOffset + $offset)]

    set value [reg read $fullOffset]

    return $value
}


# Add help information on all importatnt commands in this script
parse_rdf_help
block_offset_help
reg_read_help
reg_offset_help
reg_size_help
reg_type_help
reset_chip_help
reset_value_help
read_regs_help
rd_help
wr_help
read_mem_help
set_xmac_lb_help

