################################################################################
## procs for proccessing lists as scripts (allows usage of comments for example)
# One way to look at a Tcl script is as a sequence of words that are grouped
# into commands. The sl command presented below extracts those words into a
# single list while obeying the semantics of a Tcl script. Representing a list
# in this way, as a Tcl script, allows for elements on multiple lines without
# having to escape the newline, and also allows for comments between words in
# the list.
proc sl script {
    lsearch -all -inline -not -exact  [concat {*}[lmap part [commands $script] {
        if {[string match #* $part]} continue
        uplevel 1 list $part
    }]] {}
}

#Iterate over all elements in one or more lists and collect results
proc lmap {_var list body} {
    upvar 1 $_var var
    set res {}
    foreach var $list {lappend res [uplevel 1 $body]}
    set res
}

#returns a list of the commands in a script
proc commands script {
    set commands {}
    set chunk {}
    foreach line [split $script \n] {
        append chunk $line
        if {[info complete $chunk\n]} {
            # $chunk ends in a complete Tcl command, and none of the
            # newlines within it end a complete Tcl command.  If there
            # are multiple Tcl commands in $chunk, they must be
            # separated by semi-colons.
            set cmd {}
            foreach part [split $chunk \;] {
                append cmd $part
                if {[info complete $cmd\n]} {
                    set cmd [string trimleft $cmd[set cmd {}] "\f\n\r\t\v "]
                    if {[string match #* $cmd]} {
                        #the semi-colon was part of a comment.  Add it back
                        append cmd \;
                        continue
                    }
                    #drop empty commands
                    if {$cmd eq {}} {
                        continue
                    }
                    lappend commands $cmd
                    set cmd {}
                } else {
                    # No complete command yet.
                    # Replace semicolon and continue
                    append cmd \;
                }
            }
            # Handle comments, removing synthetic semicolon at the end
            if {$cmd ne {}} {
                lappend commands [string replace $cmd[set cmd {}] end end]
            }
            set chunk {}
        } else {
            # No end of command yet.  Put the newline back and continue
            append chunk \n
        }
    }
    if {![string match {} [string trimright $chunk]]} {
        return -code error "Can't parse script into a\
                sequence of commands.\n\tIncomplete\
                command:\n-----\n$chunk\n-----"
    }
    return $commands
}